@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2001
+ * Copyright (c) 1996-2003
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -43,7 +43,7 @@
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: bt_put.c,v 11.59 2001/07/02 01:05:35 bostic Exp $";
+static const char revid[] = "$Id: bt_put.c,v 11.78 2003/10/31 15:07:40 bostic Exp $";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -53,15 +53,18 @@ static const char revid[] = "$Id: bt_put.c,v 11.59 2001/07/02 01:05:35 bostic Ex
 #endif
 
 #include "db_int.h"
-#include "db_page.h"
-#include "btree.h"
+#include "dbinc/db_page.h"
+#include "dbinc/db_shash.h"
+#include "dbinc/btree.h"
+#include "dbinc/mp.h"
 
 static int __bam_build
 	       __P((DBC *, u_int32_t, DBT *, PAGE *, u_int32_t, u_int32_t));
 static int __bam_dup_convert __P((DBC *, PAGE *, u_int32_t));
 static int __bam_ovput
 	       __P((DBC *, u_int32_t, db_pgno_t, PAGE *, u_int32_t, DBT *));
-static u_int32_t __bam_partsize __P((u_int32_t, DBT *, PAGE *, u_int32_t));
+static u_int32_t
+	   __bam_partsize __P((DB *, u_int32_t, DBT *, PAGE *, u_int32_t));
 
 /*
  * __bam_iitem --
@@ -75,11 +78,13 @@ __bam_iitem(dbc, key, data, op, flags)
 	DBT *key, *data;
 	u_int32_t op, flags;
 {
+	DB_ENV *dbenv;
 	BKEYDATA *bk, bk_tmp;
 	BTREE *t;
 	BTREE_CURSOR *cp;
 	DB *dbp;
 	DBT bk_hdr, tdbt;
+	DB_MPOOLFILE *mpf;
 	PAGE *h;
 	db_indx_t indx;
 	u_int32_t data_size, have_bytes, need_bytes, needed;
@@ -88,6 +93,8 @@ __bam_iitem(dbc, key, data, op, flags)
 	COMPQUIET(bk, NULL);
 
 	dbp = dbc->dbp;
+	dbenv = dbp->dbenv;
+	mpf = dbp->mpf;
 	cp = (BTREE_CURSOR *)dbc->internal;
 	t = dbp->bt_internal;
 	h = cp->page;
@@ -98,11 +105,9 @@ __bam_iitem(dbc, key, data, op, flags)
 	 * Fixed-length records with partial puts: it's an error to specify
 	 * anything other simple overwrite.
 	 */
-	if (F_ISSET(dbp, DB_RE_FIXEDLEN) &&
-	    F_ISSET(data, DB_DBT_PARTIAL) && data->dlen != data->size) {
-		data_size = data->size;
-		goto len_err;
-	}
+	if (F_ISSET(dbp, DB_AM_FIXEDLEN) &&
+	    F_ISSET(data, DB_DBT_PARTIAL) && data->size != data->dlen)
+		return (__db_rec_repl(dbenv, data->size, data->dlen));
 
 	/*
 	 * Figure out how much space the data will take, including if it's a
@@ -113,15 +118,11 @@ __bam_iitem(dbc, key, data, op, flags)
 	 * the fixed-length record size.
 	 */
 	data_size = F_ISSET(data, DB_DBT_PARTIAL) ?
-	    __bam_partsize(op, data, h, indx) : data->size;
+	    __bam_partsize(dbp, op, data, h, indx) : data->size;
 	padrec = 0;
-	if (F_ISSET(dbp, DB_RE_FIXEDLEN)) {
-		if (data_size > t->re_len) {
-len_err:		__db_err(dbp->dbenv,
-			    "Length improper for fixed length record %lu",
-			    (u_long)data_size);
-			return (EINVAL);
-		}
+	if (F_ISSET(dbp, DB_AM_FIXEDLEN)) {
+		if (data_size > t->re_len)
+			return (__db_rec_toobig(dbenv, data_size, t->re_len));
 
 		/* Records that are deleted anyway needn't be padded out. */
 		if (!LF_ISSET(BI_DELETED) && data_size < t->re_len) {
@@ -151,12 +152,12 @@ len_err:		__db_err(dbp->dbenv,
 	 */
 	if (op == DB_CURRENT && dbp->dup_compare != NULL) {
 		if ((ret = __bam_cmp(dbp, data, h,
-		     indx + (TYPE(h) == P_LBTREE ? O_INDX : 0),
-		     dbp->dup_compare, &cmp)) != 0)
+		    indx + (TYPE(h) == P_LBTREE ? O_INDX : 0),
+		    dbp->dup_compare, &cmp)) != 0)
 			return (ret);
 		if (cmp != 0) {
-			__db_err(dbp->dbenv,
-			    "Current data differs from put data");
+			__db_err(dbenv,
+		"Existing data sorts differently from put data");
 			return (EINVAL);
 		}
 	}
@@ -195,7 +196,7 @@ len_err:		__db_err(dbp->dbenv,
 		 */
 		bigkey = 0;
 		if (op == DB_CURRENT) {
-			bk = GET_BKEYDATA(h,
+			bk = GET_BKEYDATA(dbp, h,
 			    indx + (TYPE(h) == P_LBTREE ? O_INDX : 0));
 			if (B_TYPE(bk->type) == B_KEYDATA)
 				have_bytes = BKEYDATA_PSIZE(bk->len);
@@ -215,19 +216,14 @@ len_err:		__db_err(dbp->dbenv,
 			needed += need_bytes - have_bytes;
 		break;
 	default:
-		return (__db_unknown_flag(dbp->dbenv, "__bam_iitem", op));
+		return (__db_unknown_flag(dbenv, "DB->put", op));
 	}
 
 	/*
 	 * If there's not enough room, or the user has put a ceiling on the
 	 * number of keys permitted in the page, split the page.
-	 *
-	 * XXX
-	 * The t->bt_maxkey test here may be insufficient -- do we have to
-	 * check in the btree split code, so we don't undo it there!?!?
 	 */
-	if (P_FREESPACE(h) < needed ||
-	    (t->bt_maxkey != 0 && NUM_ENT(h) > t->bt_maxkey))
+	if (P_FREESPACE(dbp, h) < needed)
 		return (DB_NEEDSPLIT);
 
 	/*
@@ -291,24 +287,25 @@ len_err:		__db_err(dbp->dbenv,
 		  * we deadlock or fail while deleting the overflow item or
 		  * replacing the non-overflow item, a subsequent cursor close
 		  * will try and remove the item because the cursor's delete
-		  * flag is set
+		  * flag is set.
 		  */
 		(void)__bam_ca_delete(dbp, PGNO(h), indx, 0);
 
 		if (TYPE(h) == P_LBTREE) {
 			++indx;
 			dupadjust = 1;
-
-			/*
-			 * In a Btree deleted records aren't counted (deleted
-			 * records are counted in a Recno because all accesses
-			 * are based on record number).  If it's a Btree and
-			 * it's a DB_CURRENT operation overwriting a previously
-			 * deleted record, increment the record count.
-			 */
-			was_deleted = B_DISSET(bk->type);
 		}
 
+		/*
+		 * In a Btree deleted records aren't counted (deleted records
+		 * are counted in a Recno because all accesses are based on
+		 * record number).  If it's a Btree and it's a DB_CURRENT
+		 * operation overwriting a previously deleted record, increment
+		 * the record count.
+		 */
+		if (TYPE(h) == P_LBTREE || TYPE(h) == P_LDUP)
+			was_deleted = B_DISSET(bk->type);
+
 		/*
 		 * 4. Delete and re-add the data item.
 		 *
@@ -328,7 +325,7 @@ len_err:		__db_err(dbp->dbenv,
 		replace = 1;
 		break;
 	default:
-		return (__db_unknown_flag(dbp->dbenv, "__bam_iitem", op));
+		return (__db_unknown_flag(dbenv, "DB->put", op));
 	}
 
 	/* Add the data. */
@@ -357,7 +354,7 @@ len_err:		__db_err(dbp->dbenv,
 		if (ret != 0)
 			return (ret);
 	}
-	if ((ret = memp_fset(dbp->mpf, h, DB_MPOOL_DIRTY)) != 0)
+	if ((ret = __memp_fset(mpf, h, DB_MPOOL_DIRTY)) != 0)
 		return (ret);
 
 	/*
@@ -385,7 +382,7 @@ len_err:		__db_err(dbp->dbenv,
 	 * up at least 25% of the space on the page.  If it does, move it onto
 	 * its own page.
 	 */
-	if (dupadjust && P_FREESPACE(h) <= dbp->pgsize / 2) {
+	if (dupadjust && P_FREESPACE(dbp, h) <= dbp->pgsize / 2) {
 		if ((ret = __bam_dup_convert(dbc, h, indx - O_INDX)) != 0)
 			return (ret);
 	}
@@ -402,7 +399,8 @@ len_err:		__db_err(dbp->dbenv,
  *	Figure out how much space a partial data item is in total.
  */
 static u_int32_t
-__bam_partsize(op, data, h, indx)
+__bam_partsize(dbp, op, data, h, indx)
+	DB *dbp;
 	u_int32_t op, indx;
 	DBT *data;
 	PAGE *h;
@@ -421,7 +419,7 @@ __bam_partsize(op, data, h, indx)
 	 * Otherwise, it's the data provided plus any already existing data
 	 * that we're not replacing.
 	 */
-	bk = GET_BKEYDATA(h, indx + (TYPE(h) == P_LBTREE ? O_INDX : 0));
+	bk = GET_BKEYDATA(dbp, h, indx + (TYPE(h) == P_LBTREE ? O_INDX : 0));
 	nbytes =
 	    B_TYPE(bk->type) == B_OVERFLOW ? ((BOVERFLOW *)bk)->tlen : bk->len;
 
@@ -470,7 +468,7 @@ __bam_build(dbc, op, dbt, h, indx, nbytes)
 	 * specified; get it over with.
 	 */
 	memset(rdata->data,
-	   F_ISSET(dbp, DB_RE_FIXEDLEN) ? t->re_pad : 0, nbytes);
+	   F_ISSET(dbp, DB_AM_FIXEDLEN) ? t->re_pad : 0, nbytes);
 
 	/*
 	 * In the next clauses, we need to do three things: a) set p to point
@@ -489,7 +487,8 @@ __bam_build(dbc, op, dbt, h, indx, nbytes)
 
 	/* Find the current record. */
 	if (indx < NUM_ENT(h)) {
-		bk = GET_BKEYDATA(h, indx + (TYPE(h) == P_LBTREE ? O_INDX : 0));
+		bk = GET_BKEYDATA(dbp, h, indx + (TYPE(h) == P_LBTREE ?
+		    O_INDX : 0));
 		bo = (BOVERFLOW *)bk;
 	} else {
 		bk = &tbk;
@@ -551,7 +550,7 @@ user_copy:
 	tlen += dbt->size;
 
 	/* Set the DBT to reference our new record. */
-	rdata->size = F_ISSET(dbp, DB_RE_FIXEDLEN) ? t->re_len : tlen;
+	rdata->size = F_ISSET(dbp, DB_AM_FIXEDLEN) ? t->re_len : tlen;
 	rdata->dlen = 0;
 	rdata->doff = 0;
 	rdata->flags = 0;
@@ -578,6 +577,7 @@ __bam_ritem(dbc, h, indx, data)
 	db_indx_t cnt, lo, ln, min, off, prefix, suffix;
 	int32_t nbytes;
 	int ret;
+	db_indx_t *inp;
 	u_int8_t *p, *t;
 
 	dbp = dbc->dbp;
@@ -587,10 +587,10 @@ __bam_ritem(dbc, h, indx, data)
 	 * to insert and whether it fits is handled in the caller.  All we do
 	 * here is manage the page shuffling.
 	 */
-	bk = GET_BKEYDATA(h, indx);
+	bk = GET_BKEYDATA(dbp, h, indx);
 
 	/* Log the change. */
-	if (DB_LOGGING(dbc)) {
+	if (DBC_LOGGING(dbc)) {
 		/*
 		 * We might as well check to see if the two data items share
 		 * a common prefix and suffix -- it can save us a lot of log
@@ -614,9 +614,8 @@ __bam_ritem(dbc, h, indx, data)
 		orig.size = bk->len - (prefix + suffix);
 		repl.data = (u_int8_t *)data->data + prefix;
 		repl.size = data->size - (prefix + suffix);
-		if ((ret = __bam_repl_log(dbp->dbenv, dbc->txn,
-		    &LSN(h), 0, dbp->log_fileid, PGNO(h), &LSN(h),
-		    (u_int32_t)indx, (u_int32_t)B_DISSET(bk->type),
+		if ((ret = __bam_repl_log(dbp, dbc->txn, &LSN(h), 0, PGNO(h),
+		    &LSN(h), (u_int32_t)indx, (u_int32_t)B_DISSET(bk->type),
 		    &orig, &repl, (u_int32_t)prefix, (u_int32_t)suffix)) != 0)
 			return (ret);
 	} else
@@ -626,6 +625,7 @@ __bam_ritem(dbc, h, indx, data)
 	 * Set references to the first in-use byte on the page and the
 	 * first byte of the item being replaced.
 	 */
+	inp = P_INP(dbp, h);
 	p = (u_int8_t *)h + HOFFSET(h);
 	t = (u_int8_t *)bk;
 
@@ -636,19 +636,19 @@ __bam_ritem(dbc, h, indx, data)
 	 * the regions overlap.
 	 */
 	lo = BKEYDATA_SIZE(bk->len);
-	ln = BKEYDATA_SIZE(data->size);
+	ln = (db_indx_t)BKEYDATA_SIZE(data->size);
 	if (lo != ln) {
 		nbytes = lo - ln;		/* Signed difference. */
 		if (p == t)			/* First index is fast. */
-			h->inp[indx] += nbytes;
+			inp[indx] += nbytes;
 		else {				/* Else, shift the page. */
-			memmove(p + nbytes, p, t - p);
+			memmove(p + nbytes, p, (size_t)(t - p));
 
 			/* Adjust the indices' offsets. */
-			off = h->inp[indx];
+			off = inp[indx];
 			for (cnt = 0; cnt < NUM_ENT(h); ++cnt)
-				if (h->inp[cnt] <= off)
-					h->inp[cnt] += nbytes;
+				if (inp[cnt] <= off)
+					inp[cnt] += nbytes;
 		}
 
 		/* Clean up the page and adjust the item's reference. */
@@ -679,25 +679,32 @@ __bam_dup_convert(dbc, h, indx)
 	BKEYDATA *bk;
 	DB *dbp;
 	DBT hdr;
+	DB_MPOOLFILE *mpf;
 	PAGE *dp;
-	db_indx_t cnt, cpindx, dindx, first, sz;
+	db_indx_t cnt, cpindx, dindx, first, *inp, sz;
 	int ret;
 
 	dbp = dbc->dbp;
+	mpf = dbp->mpf;
+	inp = P_INP(dbp, h);
 
 	/*
 	 * Count the duplicate records and calculate how much room they're
 	 * using on the page.
 	 */
-	while (indx > 0 && h->inp[indx] == h->inp[indx - P_INDX])
+	while (indx > 0 && inp[indx] == inp[indx - P_INDX])
 		indx -= P_INDX;
-	for (cnt = 0, sz = 0, first = indx;; ++cnt, indx += P_INDX) {
-		if (indx >= NUM_ENT(h) || h->inp[first] != h->inp[indx])
-			break;
-		bk = GET_BKEYDATA(h, indx);
-		sz += B_TYPE(bk->type) == B_KEYDATA ?
-		    BKEYDATA_PSIZE(bk->len) : BOVERFLOW_PSIZE;
-		bk = GET_BKEYDATA(h, indx + O_INDX);
+
+	/* Count the key once. */
+	bk = GET_BKEYDATA(dbp, h, indx);
+	sz = B_TYPE(bk->type) == B_KEYDATA ?
+	    BKEYDATA_PSIZE(bk->len) : BOVERFLOW_PSIZE;
+
+	/* Sum up all the data items. */
+	for (cnt = 0, first = indx;
+	    inp[first] == inp[indx] && indx < NUM_ENT(h);
+	    ++cnt, indx += P_INDX) {
+		bk = GET_BKEYDATA(dbp, h, indx + O_INDX);
 		sz += B_TYPE(bk->type) == B_KEYDATA ?
 		    BKEYDATA_PSIZE(bk->len) : BOVERFLOW_PSIZE;
 	}
@@ -752,7 +759,7 @@ __bam_dup_convert(dbc, h, indx)
 		 * deleted entries are discarded (if the deleted entry is
 		 * overflow, then free up those pages).
 		 */
-		bk = GET_BKEYDATA(h, dindx + 1);
+		bk = GET_BKEYDATA(dbp, h, dindx + 1);
 		hdr.data = bk;
 		hdr.size = B_TYPE(bk->type) == B_KEYDATA ?
 		    BKEYDATA_SIZE(bk->len) : BOVERFLOW_SIZE;
@@ -764,7 +771,7 @@ __bam_dup_convert(dbc, h, indx)
 			 */
 			if (B_TYPE(bk->type) == B_OVERFLOW &&
 			    (ret = __db_doff(dbc,
-			    (GET_BOVERFLOW(h, dindx + 1))->pgno)) != 0)
+			    (GET_BOVERFLOW(dbp, h, dindx + 1))->pgno)) != 0)
 				goto err;
 		} else {
 			if ((ret = __db_pitem(
@@ -788,17 +795,17 @@ __bam_dup_convert(dbc, h, indx)
 
 	/* Put in a new data item that points to the duplicates page. */
 	if ((ret = __bam_ovput(dbc,
-	     B_DUPLICATE, dp->pgno, h, first + 1, NULL)) != 0)
+	    B_DUPLICATE, dp->pgno, h, first + 1, NULL)) != 0)
 		goto err;
 
-	/* Adjust cursors for all the above movments. */
+	/* Adjust cursors for all the above movements. */
 	if ((ret = __bam_ca_di(dbc,
-	    PGNO(h), first + P_INDX, first + P_INDX - indx)) != 0)
+	    PGNO(h), first + P_INDX, (int)(first + P_INDX - indx))) != 0)
 		goto err;
 
-	return (memp_fput(dbp->mpf, dp, DB_MPOOL_DIRTY));
+	return (__memp_fput(mpf, dp, DB_MPOOL_DIRTY));
 
-err:	(void)memp_fput(dbp->mpf, dp, 0);
+err:	(void)__memp_fput(mpf, dp, 0);
 	return (ret);
 }
 
