@@ -1,14 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2002
+ * Copyright (c) 1996-2003
  *	Sleepycat Software.  All rights reserved.
  */
 
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: snprintf.c,v 11.10 2002/01/11 15:51:28 bostic Exp $";
+static const char revid[] = "$Id: snprintf.c,v 11.12 2003/05/02 16:10:41 bostic Exp $";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -29,7 +29,7 @@ static const char revid[] = "$Id: snprintf.c,v 11.10 2002/01/11 15:51:28 bostic
  */
 #ifndef HAVE_SNPRINTF
 int
-#ifdef __STDC__
+#ifdef STDC_HEADERS
 snprintf(char *str, size_t n, const char *fmt, ...)
 #else
 snprintf(str, n, fmt, va_alist)
@@ -62,7 +62,7 @@ snprintf(str, n, fmt, va_alist)
 		    sprintf(buf, "1234") != 4;
 	}
 
-#ifdef __STDC__
+#ifdef STDC_HEADERS
 	va_start(ap, fmt);
 #else
 	va_start(ap);
