@@ -1,14 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2002
+ * Copyright (c) 1996-2003
  *	Sleepycat Software.  All rights reserved.
  */
 
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: db_getlong.c,v 11.18 2002/03/28 20:13:33 bostic Exp $";
+static const char revid[] = "$Id: db_getlong.c,v 11.20 2003/01/08 04:07:46 bostic Exp $";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -26,11 +26,11 @@ static const char revid[] = "$Id: db_getlong.c,v 11.18 2002/03/28 20:13:33 bosti
  *	Return a long value inside of basic parameters.
  *
  * PUBLIC: int __db_getlong
- * PUBLIC:     __P((DB *, const char *, char *, long, long, long *));
+ * PUBLIC:     __P((DB_ENV *, const char *, char *, long, long, long *));
  */
 int
-__db_getlong(dbp, progname, p, min, max, storep)
-	DB *dbp;
+__db_getlong(dbenv, progname, p, min, max, storep)
+	DB_ENV *dbenv;
 	const char *progname;
 	char *p;
 	long min, max, *storep;
@@ -42,38 +42,38 @@ __db_getlong(dbp, progname, p, min, max, storep)
 	val = strtol(p, &end, 10);
 	if ((val == LONG_MIN || val == LONG_MAX) &&
 	    __os_get_errno() == ERANGE) {
-		if (dbp == NULL)
+		if (dbenv == NULL)
 			fprintf(stderr,
 			    "%s: %s: %s\n", progname, p, strerror(ERANGE));
 		else
-			dbp->err(dbp, ERANGE, "%s", p);
+			dbenv->err(dbenv, ERANGE, "%s", p);
 		return (1);
 	}
 	if (p[0] == '\0' || (end[0] != '\0' && end[0] != '\n')) {
-		if (dbp == NULL)
+		if (dbenv == NULL)
 			fprintf(stderr,
 			    "%s: %s: Invalid numeric argument\n", progname, p);
 		else
-			dbp->errx(dbp, "%s: Invalid numeric argument", p);
+			dbenv->errx(dbenv, "%s: Invalid numeric argument", p);
 		return (1);
 	}
 	if (val < min) {
-		if (dbp == NULL)
+		if (dbenv == NULL)
 			fprintf(stderr,
 			    "%s: %s: Less than minimum value (%ld)\n",
 			    progname, p, min);
 		else
-			dbp->errx(dbp,
+			dbenv->errx(dbenv,
 			    "%s: Less than minimum value (%ld)", p, min);
 		return (1);
 	}
 	if (val > max) {
-		if (dbp == NULL)
+		if (dbenv == NULL)
 			fprintf(stderr,
 			    "%s: %s: Greater than maximum value (%ld)\n",
 			    progname, p, max);
 		else
-			dbp->errx(dbp,
+			dbenv->errx(dbenv,
 			    "%s: Greater than maximum value (%ld)", p, max);
 		return (1);
 	}
@@ -86,11 +86,11 @@ __db_getlong(dbp, progname, p, min, max, storep)
  *	Return an unsigned long value inside of basic parameters.
  *
  * PUBLIC: int __db_getulong
- * PUBLIC:     __P((DB *, const char *, char *, u_long, u_long, u_long *));
+ * PUBLIC:     __P((DB_ENV *, const char *, char *, u_long, u_long, u_long *));
  */
 int
-__db_getulong(dbp, progname, p, min, max, storep)
-	DB *dbp;
+__db_getulong(dbenv, progname, p, min, max, storep)
+	DB_ENV *dbenv;
 	const char *progname;
 	char *p;
 	u_long min, max, *storep;
@@ -98,7 +98,7 @@ __db_getulong(dbp, progname, p, min, max, storep)
 #if !defined(HAVE_STRTOUL)
 	COMPQUIET(min, 0);
 
-	return (__db_getlong(dbp, progname, p, 0, max, (long *)storep));
+	return (__db_getlong(dbenv, progname, p, 0, max, (long *)storep));
 #else
 	u_long val;
 	char *end;
@@ -106,28 +106,28 @@ __db_getulong(dbp, progname, p, min, max, storep)
 	__os_set_errno(0);
 	val = strtoul(p, &end, 10);
 	if (val == ULONG_MAX && __os_get_errno() == ERANGE) {
-		if (dbp == NULL)
+		if (dbenv == NULL)
 			fprintf(stderr,
 			    "%s: %s: %s\n", progname, p, strerror(ERANGE));
 		else
-			dbp->err(dbp, ERANGE, "%s", p);
+			dbenv->err(dbenv, ERANGE, "%s", p);
 		return (1);
 	}
 	if (p[0] == '\0' || (end[0] != '\0' && end[0] != '\n')) {
-		if (dbp == NULL)
+		if (dbenv == NULL)
 			fprintf(stderr,
 			    "%s: %s: Invalid numeric argument\n", progname, p);
 		else
-			dbp->errx(dbp, "%s: Invalid numeric argument", p);
+			dbenv->errx(dbenv, "%s: Invalid numeric argument", p);
 		return (1);
 	}
 	if (val < min) {
-		if (dbp == NULL)
+		if (dbenv == NULL)
 			fprintf(stderr,
 			    "%s: %s: Less than minimum value (%lu)\n",
 			    progname, p, min);
 		else
-			dbp->errx(dbp,
+			dbenv->errx(dbenv,
 			    "%s: Less than minimum value (%lu)", p, min);
 		return (1);
 	}
@@ -139,12 +139,12 @@ __db_getulong(dbp, progname, p, min, max, storep)
 	 * may not exist on all platforms.
 	 */
 	if (max != 0 && val > max) {
-		if (dbp == NULL)
+		if (dbenv == NULL)
 			fprintf(stderr,
 			    "%s: %s: Greater than maximum value (%lu)\n",
 			    progname, p, max);
 		else
-			dbp->errx(dbp,
+			dbenv->errx(dbenv,
 			    "%s: Greater than maximum value (%lu)", p, max);
 		return (1);
 	}
