@@ -1,14 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2000-2001
+ * Copyright (c) 2000-2003
  *	Sleepycat Software.  All rights reserved.
  */
 
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: util_log.c,v 1.8 2001/01/25 18:22:38 bostic Exp $";
+static const char revid[] = "$Id: util_log.c,v 1.13 2003/05/05 19:54:57 bostic Exp $";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -27,11 +27,9 @@ static const char revid[] = "$Id: util_log.c,v 1.8 2001/01/25 18:22:38 bostic Ex
 #endif
 
 #include <string.h>
-#include <unistd.h>
 #endif
 
 #include "db_int.h"
-#include "common_ext.h"
 
 /*
  * __db_util_logset --
@@ -46,12 +44,14 @@ __db_util_logset(progname, fname)
 {
 	FILE *fp;
 	time_t now;
+	u_int32_t id;
 
 	if ((fp = fopen(fname, "w")) == NULL)
 		goto err;
 
 	(void)time(&now);
-	fprintf(fp, "%s: %lu %s", progname, (u_long)getpid(), ctime(&now));
+	__os_id(&id);
+	fprintf(fp, "%s: %lu %s", progname, (u_long)id, ctime(&now));
 
 	if (fclose(fp) == EOF)
 		goto err;
