@@ -1,14 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2002
+ * Copyright (c) 1997-2003
  *	Sleepycat Software.  All rights reserved.
  */
 
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: cxx_mpool.cpp,v 11.20 2002/07/03 21:03:53 bostic Exp $";
+static const char revid[] = "$Id: cxx_mpool.cpp,v 11.27 2003/06/30 17:19:41 bostic Exp $";
 #endif /* not lint */
 
 #include <errno.h>
@@ -35,7 +35,8 @@ int DbMpoolFile::_name _argspec						\
 	else								\
 		ret = mpf->_name _arglist;				\
 	if (!_retok(ret))						\
-		DB_ERROR("DbMpoolFile::"#_name, ret, ON_ERROR_UNKNOWN);	\
+		DB_ERROR(DbEnv::get_DbEnv(mpf->dbenv), \
+			"DbMpoolFile::"#_name, ret, ON_ERROR_UNKNOWN);	\
 	return (ret);							\
 }
 
@@ -66,6 +67,7 @@ int DbMpoolFile::close(u_int32_t flags)
 {
 	DB_MPOOLFILE *mpf = unwrap(this);
 	int ret;
+	DbEnv *dbenv = DbEnv::get_DbEnv(mpf->dbenv);
 
 	if (mpf == NULL)
 		ret = EINVAL;
@@ -79,32 +81,51 @@ int DbMpoolFile::close(u_int32_t flags)
 	delete this;
 
 	if (!DB_RETOK_STD(ret))
-		DB_ERROR("DbMpoolFile::close", ret, ON_ERROR_UNKNOWN);
+		DB_ERROR(dbenv, "DbMpoolFile::close", ret, ON_ERROR_UNKNOWN);
 
 	return (ret);
 }
 
 DB_MPOOLFILE_METHOD(get, (db_pgno_t *pgnoaddr, u_int32_t flags, void *pagep),
     (mpf, pgnoaddr, flags, pagep), DB_RETOK_MPGET)
-DB_MPOOLFILE_METHOD_VOID(last_pgno, (db_pgno_t *pgnoaddr), (mpf, pgnoaddr))
 DB_MPOOLFILE_METHOD(open,
     (const char *file, u_int32_t flags, int mode, size_t pagesize),
     (mpf, file, flags, mode, pagesize), DB_RETOK_STD)
 DB_MPOOLFILE_METHOD(put, (void *pgaddr, u_int32_t flags),
     (mpf, pgaddr, flags), DB_RETOK_STD)
-DB_MPOOLFILE_METHOD_VOID(refcnt, (db_pgno_t *pgnoaddr), (mpf, pgnoaddr))
 DB_MPOOLFILE_METHOD(set, (void *pgaddr, u_int32_t flags),
     (mpf, pgaddr, flags), DB_RETOK_STD)
+DB_MPOOLFILE_METHOD(get_clear_len, (u_int32_t *lenp),
+    (mpf, lenp), DB_RETOK_STD)
 DB_MPOOLFILE_METHOD(set_clear_len, (u_int32_t len),
     (mpf, len), DB_RETOK_STD)
+DB_MPOOLFILE_METHOD(get_fileid, (u_int8_t *fileid),
+    (mpf, fileid), DB_RETOK_STD)
 DB_MPOOLFILE_METHOD(set_fileid, (u_int8_t *fileid),
     (mpf, fileid), DB_RETOK_STD)
+DB_MPOOLFILE_METHOD(get_flags, (u_int32_t *flagsp),
+    (mpf, flagsp), DB_RETOK_STD)
+DB_MPOOLFILE_METHOD(set_flags, (u_int32_t flags, int onoff),
+    (mpf, flags, onoff), DB_RETOK_STD)
+DB_MPOOLFILE_METHOD(get_ftype, (int *ftypep),
+    (mpf, ftypep), DB_RETOK_STD)
 DB_MPOOLFILE_METHOD(set_ftype, (int ftype),
     (mpf, ftype), DB_RETOK_STD)
+DB_MPOOLFILE_METHOD(get_lsn_offset, (int32_t *offsetp),
+    (mpf, offsetp), DB_RETOK_STD)
 DB_MPOOLFILE_METHOD(set_lsn_offset, (int32_t offset),
     (mpf, offset), DB_RETOK_STD)
+DB_MPOOLFILE_METHOD(get_maxsize, (u_int32_t *gbytesp, u_int32_t *bytesp),
+    (mpf, gbytesp, bytesp), DB_RETOK_STD)
+DB_MPOOLFILE_METHOD(set_maxsize, (u_int32_t gbytes, u_int32_t bytes),
+    (mpf, gbytes, bytes), DB_RETOK_STD)
+DB_MPOOLFILE_METHOD(get_pgcookie, (DBT *dbt),
+    (mpf, dbt), DB_RETOK_STD)
 DB_MPOOLFILE_METHOD(set_pgcookie, (DBT *dbt),
     (mpf, dbt), DB_RETOK_STD)
-DB_MPOOLFILE_METHOD_VOID(set_unlink, (int ul), (mpf, ul))
+DB_MPOOLFILE_METHOD(get_priority, (DB_CACHE_PRIORITY *priorityp),
+    (mpf, priorityp), DB_RETOK_STD)
+DB_MPOOLFILE_METHOD(set_priority, (DB_CACHE_PRIORITY priority),
+    (mpf, priority), DB_RETOK_STD)
 DB_MPOOLFILE_METHOD(sync, (),
     (mpf), DB_RETOK_STD)
