@@ -1,14 +1,14 @@
 /*
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1998-2002
+ * Copyright (c) 1998-2003
  *	Sleepycat Software.  All rights reserved.
  */
 
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: db_join.c,v 11.55 2002/08/08 03:57:47 bostic Exp $";
+static const char revid[] = "$Id: db_join.c,v 11.65 2003/10/07 18:55:39 mjc Exp $";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -23,10 +23,11 @@ static const char revid[] = "$Id: db_join.c,v 11.55 2002/08/08 03:57:47 bostic E
 #include "dbinc/db_join.h"
 #include "dbinc/btree.h"
 
-static int __db_join_close __P((DBC *));
+static int __db_join_close_pp __P((DBC *));
 static int __db_join_cmp __P((const void *, const void *));
 static int __db_join_del __P((DBC *, u_int32_t));
 static int __db_join_get __P((DBC *, DBT *, DBT *, u_int32_t));
+static int __db_join_get_pp __P((DBC *, DBT *, DBT *, u_int32_t));
 static int __db_join_getnext __P((DBC *, DBT *, DBT *, u_int32_t, u_int32_t));
 static int __db_join_primget __P((DB *,
     DB_TXN *, u_int32_t, DBT *, DBT *, u_int32_t));
@@ -84,26 +85,18 @@ __db_join(primary, curslist, dbcp, flags)
 	DB_ENV *dbenv;
 	DBC *dbc;
 	JOIN_CURSOR *jc;
-	int ret;
-	u_int32_t i;
 	size_t ncurs, nslots;
+	u_int32_t i;
+	int ret;
 
-	COMPQUIET(nslots, 0);
-
-	PANIC_CHECK(primary->dbenv);
-
-	if ((ret = __db_joinchk(primary, curslist, flags)) != 0)
-		return (ret);
-
+	dbenv = primary->dbenv;
 	dbc = NULL;
 	jc = NULL;
-	dbenv = primary->dbenv;
 
 	if ((ret = __os_calloc(dbenv, 1, sizeof(DBC), &dbc)) != 0)
 		goto err;
 
-	if ((ret = __os_calloc(dbenv,
-	    1, sizeof(JOIN_CURSOR), &jc)) != 0)
+	if ((ret = __os_calloc(dbenv, 1, sizeof(JOIN_CURSOR), &jc)) != 0)
 		goto err;
 
 	if ((ret = __os_malloc(dbenv, 256, &jc->j_key.data)) != 0)
@@ -122,7 +115,7 @@ __db_join(primary, curslist, dbcp, flags)
 	 * the number of cursors involved in the join, because the
 	 * list is NULL-terminated.
 	 */
-	ncurs = jc->j_curslist - curslist;
+	ncurs = (size_t)(jc->j_curslist - curslist);
 	nslots = ncurs + 1;
 
 	/*
@@ -208,18 +201,21 @@ __db_join(primary, curslist, dbcp, flags)
 	 * because this is the last thing that can fail.  Modifier of this
 	 * function beware!
 	 */
-	if ((ret = jc->j_curslist[0]->c_dup(jc->j_curslist[0], jc->j_workcurs,
-	    DB_POSITIONI)) != 0)
+	if ((ret =
+	    __db_c_dup(jc->j_curslist[0], jc->j_workcurs, DB_POSITION)) != 0)
 		goto err;
 
-	dbc->c_close = __db_join_close;
+	dbc->c_close = __db_join_close_pp;
 	dbc->c_del = __db_join_del;
-	dbc->c_get = __db_join_get;
+	dbc->c_get = __db_join_get_pp;
 	dbc->c_put = __db_join_put;
-	dbc->internal = (DBC_INTERNAL *) jc;
+	dbc->internal = (DBC_INTERNAL *)jc;
 	dbc->dbp = primary;
 	jc->j_primary = primary;
 
+	/* Stash the first cursor's transaction here for easy access. */
+	dbc->txn = curslist[0]->txn;
+
 	*dbcp = dbc;
 
 	MUTEX_THREAD_LOCK(dbenv, primary->mutexp);
@@ -247,6 +243,36 @@ err:	if (jc != NULL) {
 	return (ret);
 }
 
+/*
+ * __db_join_close_pp --
+ *	DBC->c_close pre/post processing for join cursors.
+ */
+static int
+__db_join_close_pp(dbc)
+	DBC *dbc;
+{
+	DB_ENV *dbenv;
+	DB *dbp;
+	int handle_check, ret;
+
+	dbp = dbc->dbp;
+	dbenv = dbp->dbenv;
+
+	PANIC_CHECK(dbenv);
+
+	handle_check = IS_REPLICATED(dbenv, dbp);
+	if (handle_check &&
+	    (ret = __db_rep_enter(dbp, 0, dbc->txn != NULL)) != 0)
+		return (ret);
+
+	ret = __db_join_close(dbc);
+
+	if (handle_check)
+		__env_rep_exit(dbenv);
+
+	return (ret);
+}
+
 static int
 __db_join_put(dbc, key, data, flags)
 	DBC *dbc;
@@ -273,6 +299,76 @@ __db_join_del(dbc, flags)
 	return (EINVAL);
 }
 
+/*
+ * __db_join_get_pp --
+ *	DBjoin->get pre/post processing.
+ */
+static int
+__db_join_get_pp(dbc, key, data, flags)
+	DBC *dbc;
+	DBT *key, *data;
+	u_int32_t flags;
+{
+	DB *dbp;
+	DB_ENV *dbenv;
+	u_int32_t handle_check, ret, save_flags;
+
+	dbp = dbc->dbp;
+	dbenv = dbp->dbenv;
+
+	/* Save the original flags value. */
+	save_flags = flags;
+
+	PANIC_CHECK(dbenv);
+
+	if (LF_ISSET(DB_DIRTY_READ | DB_RMW)) {
+		if (!LOCKING_ON(dbp->dbenv))
+			return (__db_fnl(dbp->dbenv, "DBcursor->c_get"));
+
+		LF_CLR(DB_DIRTY_READ | DB_RMW);
+	}
+
+	switch (flags) {
+	case 0:
+	case DB_JOIN_ITEM:
+		break;
+	default:
+		return (__db_ferr(dbp->dbenv, "DBcursor->c_get", 0));
+	}
+
+	/*
+	 * A partial get of the key of a join cursor don't make much sense;
+	 * the entire key is necessary to query the primary database
+	 * and find the datum, and so regardless of the size of the key
+	 * it would not be a performance improvement.  Since it would require
+	 * special handling, we simply disallow it.
+	 *
+	 * A partial get of the data, however, potentially makes sense (if
+	 * all possible data are a predictable large structure, for instance)
+	 * and causes us no headaches, so we permit it.
+	 */
+	if (F_ISSET(key, DB_DBT_PARTIAL)) {
+		__db_err(dbp->dbenv,
+		    "DB_DBT_PARTIAL may not be set on key during join_get");
+		return (EINVAL);
+	}
+
+	handle_check = IS_REPLICATED(dbp->dbenv, dbp);
+	if (handle_check &&
+	    (ret = __db_rep_enter(dbp, 1, dbc->txn != NULL)) != 0)
+		return (ret);
+
+	/* Restore the original flags value. */
+	flags = save_flags;
+
+	ret = __db_join_get(dbc, key, data, flags);
+
+	if (handle_check)
+		__db_rep_exit(dbenv);
+
+	return (ret);
+}
+
 static int
 __db_join_get(dbc, key_arg, data_arg, flags)
 	DBC *dbc;
@@ -289,8 +385,6 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 	dbp = dbc->dbp;
 	jc = (JOIN_CURSOR *)dbc->internal;
 
-	PANIC_CHECK(dbp->dbenv);
-
 	operation = LF_ISSET(DB_OPFLAGS_MASK);
 
 	/* !!!
@@ -299,9 +393,6 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 	 */
 	opmods = LF_ISSET(DB_RMW | DB_DIRTY_READ);
 
-	if ((ret = __db_joingetchk(dbp, key_arg, flags)) != 0)
-		return (ret);
-
 	/*
 	 * Since we are fetching the key as a datum in the secondary indices,
 	 * we must be careful of caller-specified DB_DBT_* memory
@@ -329,8 +420,7 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 		goto samekey;
 	F_CLR(jc, JOIN_RETRY);
 
-retry:	ret = jc->j_workcurs[0]->c_real_get(jc->j_workcurs[0],
-	    &jc->j_key, key_n,
+retry:	ret = __db_c_get(jc->j_workcurs[0], &jc->j_key, key_n,
 	    opmods | (jc->j_exhausted[0] ? DB_NEXT_DUP : DB_CURRENT));
 
 	if (ret == ENOMEM) {
@@ -358,7 +448,7 @@ retry:	ret = jc->j_workcurs[0]->c_real_get(jc->j_workcurs[0],
 	 */
 	for (i = 1; i < jc->j_ncurs; i++) {
 		if (jc->j_fdupcurs[i] != NULL &&
-		    (ret = jc->j_fdupcurs[i]->c_close(jc->j_fdupcurs[i])) != 0)
+		    (ret = __db_c_close(jc->j_fdupcurs[i])) != 0)
 			goto err;
 		jc->j_fdupcurs[i] = NULL;
 	}
@@ -382,9 +472,8 @@ retry:	ret = jc->j_workcurs[0]->c_real_get(jc->j_workcurs[0],
 		DB_ASSERT(jc->j_curslist[i] != NULL);
 		if (jc->j_workcurs[i] == NULL)
 			/* If this is NULL, we need to dup curslist into it. */
-			if ((ret = jc->j_curslist[i]->c_dup(
-			    jc->j_curslist[i], jc->j_workcurs + i,
-			    DB_POSITIONI)) != 0)
+			if ((ret = __db_c_dup(jc->j_curslist[i],
+			    jc->j_workcurs + i, DB_POSITION)) != 0)
 				goto err;
 
 retry2:		cp = jc->j_workcurs[i];
@@ -435,7 +524,7 @@ retry2:		cp = jc->j_workcurs[i];
 					 * and let strange things happen--we
 					 * can't make rope childproof.
 					 */
-					if ((ret = jc->j_workcurs[j]->c_close(
+					if ((ret = __db_c_close(
 					    jc->j_workcurs[j])) != 0)
 						goto err;
 					if (!SORTED_SET(jc, 0) ||
@@ -448,10 +537,10 @@ retry2:		cp = jc->j_workcurs[i];
 						jc->j_workcurs[j] = NULL;
 					else
 						/* Partial reset suffices. */
-						if ((jc->j_fdupcurs[j]->c_dup(
+						if ((__db_c_dup(
 						    jc->j_fdupcurs[j],
 						    &jc->j_workcurs[j],
-						    DB_POSITIONI)) != 0)
+						    DB_POSITION)) != 0)
 							goto err;
 					jc->j_exhausted[j] = 0;
 				}
@@ -467,14 +556,13 @@ retry2:		cp = jc->j_workcurs[i];
 			for (j = i + 1;
 			    jc->j_workcurs[j] != NULL;
 			    j++) {
-				if ((ret = jc->j_workcurs[j]->c_close(
-				    jc->j_workcurs[j])) != 0)
+				if ((ret =
+				    __db_c_close(jc->j_workcurs[j])) != 0)
 					goto err;
 				jc->j_exhausted[j] = 0;
 				if (jc->j_fdupcurs[j] != NULL &&
-				    (ret = jc->j_fdupcurs[j]->c_dup(
-				    jc->j_fdupcurs[j], &jc->j_workcurs[j],
-				    DB_POSITIONI)) != 0)
+				    (ret = __db_c_dup(jc->j_fdupcurs[j],
+				    &jc->j_workcurs[j], DB_POSITION)) != 0)
 					goto err;
 				else
 					jc->j_workcurs[j] = NULL;
@@ -521,7 +609,7 @@ mem_err:			__db_err(dbp->dbenv,
 		 * duplicate duplicates;  store this into jc->j_fdupcurs[i].
 		 */
 		if (SORTED_SET(jc, i) && jc->j_fdupcurs[i] == NULL && (ret =
-		    cp->c_dup(cp, &jc->j_fdupcurs[i], DB_POSITIONI)) != 0)
+		    __db_c_dup(cp, &jc->j_fdupcurs[i], DB_POSITION)) != 0)
 			goto err;
 
 	}
@@ -534,7 +622,7 @@ samekey:	/*
 		 * Get the key we tried and failed to return last time;
 		 * it should be the current datum of all the secondary cursors.
 		 */
-		if ((ret = jc->j_workcurs[0]->c_real_get(jc->j_workcurs[0],
+		if ((ret = __db_c_get(jc->j_workcurs[0],
 		    &jc->j_key, key_n, DB_CURRENT | opmods)) != 0)
 			return (ret);
 		F_CLR(jc, JOIN_RETRY);
@@ -615,7 +703,13 @@ samekey:	/*
 	return (ret);
 }
 
-static int
+/*
+ * __db_join_close --
+ *	DBC->c_close for join cursors.
+ *
+ * PUBLIC: int __db_join_close __P((DBC *));
+ */
+int
 __db_join_close(dbc)
 	DBC *dbc;
 {
@@ -653,11 +747,11 @@ __db_join_close(dbc)
 	 * mucking with.
 	 */
 	for (i = 0; i < jc->j_ncurs; i++) {
-		if (jc->j_workcurs[i] != NULL && (t_ret =
-		    jc->j_workcurs[i]->c_close(jc->j_workcurs[i])) != 0)
+		if (jc->j_workcurs[i] != NULL &&
+		    (t_ret = __db_c_close(jc->j_workcurs[i])) != 0)
 			ret = t_ret;
-		if (jc->j_fdupcurs[i] != NULL && (t_ret =
-		    jc->j_fdupcurs[i]->c_close(jc->j_fdupcurs[i])) != 0)
+		if (jc->j_fdupcurs[i] != NULL &&
+		    (t_ret = __db_c_close(jc->j_fdupcurs[i])) != 0)
 			ret = t_ret;
 	}
 
@@ -711,7 +805,7 @@ __db_join_getnext(dbc, key, data, exhausted, opmods)
 		 */
 		memset(&ldata, 0, sizeof(DBT));
 		F_SET(&ldata, DB_DBT_MALLOC);
-		if ((ret = dbc->c_real_get(dbc,
+		if ((ret = __db_c_get(dbc,
 		    key, &ldata, opmods | DB_CURRENT)) != 0)
 			break;
 		cmp = func(dbp, data, &ldata);
@@ -736,7 +830,7 @@ __db_join_getnext(dbc, key, data, exhausted, opmods)
 		__os_ufree(dbp->dbenv, ldata.data);
 		/* FALLTHROUGH */
 	case 1:
-		ret = dbc->c_real_get(dbc, key, data, opmods | DB_GET_BOTHC);
+		ret = __db_c_get(dbc, key, data, opmods | DB_GET_BOTHC);
 		break;
 	default:
 		ret = EINVAL;
@@ -757,17 +851,14 @@ __db_join_cmp(a, b)
 	DBC *dbca, *dbcb;
 	db_recno_t counta, countb;
 
-	/* In case c_count fails, pretend cursors are equal. */
-	counta = countb = 0;
-
 	dbca = *((DBC * const *)a);
 	dbcb = *((DBC * const *)b);
 
-	if (dbca->c_count(dbca, &counta, 0) != 0 ||
-	    dbcb->c_count(dbcb, &countb, 0) != 0)
+	if (__db_c_count(dbca, &counta) != 0 ||
+	    __db_c_count(dbcb, &countb) != 0)
 		return (0);
 
-	return (counta - countb);
+	return ((long)counta - (long)countb);
 }
 
 /*
@@ -797,7 +888,7 @@ __db_join_primget(dbp, txn, lockerid, key, data, flags)
 	LF_CLR(DB_RMW | DB_DIRTY_READ);
 	DB_ASSERT(flags == 0);
 
-	if ((ret = __db_icursor(dbp,
+	if ((ret = __db_cursor_int(dbp,
 	    txn, dbp->type, PGNO_INVALID, 0, lockerid, &dbc)) != 0)
 		return (ret);
 
@@ -813,10 +904,26 @@ __db_join_primget(dbp, txn, lockerid, key, data, flags)
 	 */
 	SET_RET_MEM(dbc, dbp);
 
-	ret = dbc->c_get(dbc, key, data, DB_SET | rmw);
+	ret = __db_c_get(dbc, key, data, DB_SET | rmw);
 
 	if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
 		ret = t_ret;
 
 	return (ret);
 }
+
+/*
+ * __db_secondary_corrupt --
+ *	Report that a secondary index appears corrupt, as it has a record
+ * that does not correspond to a record in the primary or vice versa.
+ *
+ * PUBLIC: int __db_secondary_corrupt __P((DB *));
+ */
+int
+__db_secondary_corrupt(dbp)
+	DB *dbp;
+{
+	__db_err(dbp->dbenv,
+	    "Secondary index corrupt: not consistent with primary");
+	return (DB_SECONDARY_BAD);
+}
