@@ -1,14 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2002
+ * Copyright (c) 1999-2003
  *	Sleepycat Software.  All rights reserved.
  */
 
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: db_method.c,v 11.78 2002/07/02 19:26:55 sue Exp $";
+static const char revid[] = "$Id: db_method.c,v 11.99 2003/07/08 20:14:17 ubell Exp $";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -27,11 +27,11 @@ static const char revid[] = "$Id: db_method.c,v 11.78 2002/07/02 19:26:55 sue Ex
 #include "dbinc/db_shash.h"
 #include "dbinc/btree.h"
 #include "dbinc/hash.h"
+#include "dbinc/lock.h"
+#include "dbinc/mp.h"
 #include "dbinc/qam.h"
 #include "dbinc/xa.h"
 #include "dbinc_auto/xa_ext.h"
-#include "dbinc/db_shash.h"
-#include "dbinc/lock.h"
 
 #ifdef HAVE_RPC
 #include "dbinc_auto/db_server.h"
@@ -39,26 +39,31 @@ static const char revid[] = "$Id: db_method.c,v 11.78 2002/07/02 19:26:55 sue Ex
 #endif
 
 static int  __db_get_byteswapped __P((DB *, int *));
+static int  __db_get_dbname __P((DB *, const char **, const char **));
+static int  __db_get_env __P((DB *, DB_ENV **));
+static int  __db_get_lorder __P((DB *, int *));
+static int  __db_get_transactional __P((DB *, int *));
 static int  __db_get_type __P((DB *, DBTYPE *dbtype));
 static int  __db_init __P((DB *, u_int32_t));
-static int  __db_key_range
-		__P((DB *, DB_TXN *, DBT *, DB_KEY_RANGE *, u_int32_t));
 static int  __db_set_alloc __P((DB *, void *(*)(size_t),
 		void *(*)(void *, size_t), void (*)(void *)));
 static int  __db_set_append_recno __P((DB *, int (*)(DB *, DBT *, db_recno_t)));
+static int  __db_get_cachesize __P((DB *, u_int32_t *, u_int32_t *, int *));
 static int  __db_set_cachesize __P((DB *, u_int32_t, u_int32_t, int));
-static int  __db_set_cache_priority __P((DB *, DB_CACHE_PRIORITY));
 static int  __db_set_dup_compare
 		__P((DB *, int (*)(DB *, const DBT *, const DBT *)));
+static int  __db_get_encrypt_flags __P((DB *, u_int32_t *));
 static int  __db_set_encrypt __P((DB *, const char *, u_int32_t));
 static int  __db_set_feedback __P((DB *, void (*)(DB *, int, int)));
-static int  __db_set_flags __P((DB *, u_int32_t));
-static int  __db_set_pagesize __P((DB *, u_int32_t));
+static void __db_map_flags __P((DB *, u_int32_t *, u_int32_t *));
+static int  __db_get_flags __P((DB *, u_int32_t *));
+static int  __db_get_pagesize __P((DB *, u_int32_t *));
 static int  __db_set_paniccall __P((DB *, void (*)(DB_ENV *, int)));
 static void __db_set_errcall __P((DB *, void (*)(const char *, char *)));
+static void __db_get_errfile __P((DB *, FILE **));
 static void __db_set_errfile __P((DB *, FILE *));
+static void __db_get_errpfx __P((DB *, const char **));
 static void __db_set_errpfx __P((DB *, const char *));
-static int  __db_stat_fail __P((DB *, void *, u_int32_t));
 static void __dbh_err __P((DB *, int, const char *, ...));
 static void __dbh_errx __P((DB *, const char *, ...));
 
@@ -85,6 +90,8 @@ db_create(dbpp, dbenv, flags)
 	switch (flags) {
 	case 0:
 		break;
+	case DB_REP_CREATE:
+		break;
 	case DB_XA_CREATE:
 		if (dbenv != NULL) {
 			__db_err(dbenv,
@@ -113,25 +120,44 @@ db_create(dbpp, dbenv, flags)
 	else
 #endif
 		ret = __db_init(dbp, flags);
-	if (ret != 0) {
-		__os_free(dbenv, dbp);
-		return (ret);
-	}
+	if (ret != 0)
+		goto err;
 
 	/* If we don't have an environment yet, allocate a local one. */
 	if (dbenv == NULL) {
-		if ((ret = db_env_create(&dbenv, 0)) != 0) {
-			__os_free(dbenv, dbp);
-			return (ret);
-		}
+		if ((ret = db_env_create(&dbenv, 0)) != 0)
+			goto err;
 		F_SET(dbenv, DB_ENV_DBLOCAL);
 	}
-	++dbenv->db_ref;
-
 	dbp->dbenv = dbenv;
+	MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
+	++dbenv->db_ref;
+	MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
+
+	/*
+	 * Set the replication timestamp; it's 0 if we're not in a replicated
+	 * environment.
+	 */
+	dbp->timestamp =
+	    (F_ISSET(dbenv, DB_ENV_DBLOCAL) || !REP_ON(dbenv)) ? 0 :
+	    ((DB_REP *)dbenv->rep_handle)->region->timestamp;
+
+	/* Open a backing DB_MPOOLFILE handle in the memory pool. */
+	if ((ret = __memp_fcreate(dbenv, &dbp->mpf)) != 0)
+		goto err;
+
+	dbp->type = DB_UNKNOWN;
 
 	*dbpp = dbp;
 	return (0);
+
+err:	if (dbp->mpf != NULL)
+		(void)__memp_fclose(dbp->mpf, 0);
+	if (dbenv != NULL && F_ISSET(dbenv, DB_ENV_DBLOCAL))
+		(void)__dbenv_close(dbenv, 0);
+	__os_free(dbenv, dbp);
+	*dbpp = NULL;
+	return (ret);
 }
 
 /*
@@ -156,42 +182,52 @@ __db_init(dbp, flags)
 	FLD_SET(dbp->am_ok,
 	    DB_OK_BTREE | DB_OK_HASH | DB_OK_QUEUE | DB_OK_RECNO);
 
-	dbp->associate = __db_associate;
-	dbp->close = __db_close;
-	dbp->cursor = __db_cursor;
-	dbp->del = __db_delete;
+	dbp->associate = __db_associate_pp;
+	dbp->close = __db_close_pp;
+	dbp->cursor = __db_cursor_pp;
+	dbp->del = __db_del_pp;
 	dbp->err = __dbh_err;
 	dbp->errx = __dbh_errx;
-	dbp->fd = __db_fd;
-	dbp->get = __db_get;
+	dbp->fd = __db_fd_pp;
+	dbp->get = __db_get_pp;
 	dbp->get_byteswapped = __db_get_byteswapped;
+	dbp->get_dbname = __db_get_dbname;
+	dbp->get_env = __db_get_env;
+	dbp->get_open_flags = __db_get_open_flags;
+	dbp->get_transactional = __db_get_transactional;
 	dbp->get_type = __db_get_type;
-	dbp->join = __db_join;
-	dbp->key_range = __db_key_range;
-	dbp->open = __db_open;
-	dbp->pget = __db_pget;
-	dbp->put = __db_put;
-	dbp->remove = __db_remove;
-	dbp->rename = __db_rename;
-	dbp->truncate = __db_truncate;
+	dbp->join = __db_join_pp;
+	dbp->key_range = __db_key_range_pp;
+	dbp->open = __db_open_pp;
+	dbp->pget = __db_pget_pp;
+	dbp->put = __db_put_pp;
+	dbp->remove = __db_remove_pp;
+	dbp->rename = __db_rename_pp;
+	dbp->truncate = __db_truncate_pp;
 	dbp->set_alloc = __db_set_alloc;
 	dbp->set_append_recno = __db_set_append_recno;
+	dbp->get_cachesize = __db_get_cachesize;
 	dbp->set_cachesize = __db_set_cachesize;
-	dbp->set_cache_priority = __db_set_cache_priority;
 	dbp->set_dup_compare = __db_set_dup_compare;
+	dbp->get_encrypt_flags = __db_get_encrypt_flags;
 	dbp->set_encrypt = __db_set_encrypt;
 	dbp->set_errcall = __db_set_errcall;
+	dbp->get_errfile = __db_get_errfile;
 	dbp->set_errfile = __db_set_errfile;
+	dbp->get_errpfx = __db_get_errpfx;
 	dbp->set_errpfx = __db_set_errpfx;
 	dbp->set_feedback = __db_set_feedback;
+	dbp->get_flags = __db_get_flags;
 	dbp->set_flags = __db_set_flags;
+	dbp->get_lorder = __db_get_lorder;
 	dbp->set_lorder = __db_set_lorder;
+	dbp->get_pagesize = __db_get_pagesize;
 	dbp->set_pagesize = __db_set_pagesize;
 	dbp->set_paniccall = __db_set_paniccall;
-	dbp->stat = __db_stat_fail;
-	dbp->sync = __db_sync;
-	dbp->upgrade = __db_upgrade;
-	dbp->verify = __db_verify;
+	dbp->stat = __db_stat_pp;
+	dbp->sync = __db_sync_pp;
+	dbp->upgrade = __db_upgrade_pp;
+	dbp->verify = __db_verify_pp;
 
 					/* Access method specific. */
 	if ((ret = __bam_db_create(dbp)) != 0)
@@ -208,6 +244,9 @@ __db_init(dbp, flags)
 	if (LF_ISSET(DB_XA_CREATE) && (ret = __db_xa_create(dbp)) != 0)
 		return (ret);
 
+	if (LF_ISSET(DB_REP_CREATE))
+		F_SET(dbp, DB_AM_REPLICATION);
+
 	return (0);
 }
 
@@ -245,7 +284,7 @@ __dbh_am_chk(dbp, flags)
  *	Error message, including the standard error string.
  */
 static void
-#ifdef __STDC__
+#ifdef STDC_HEADERS
 __dbh_err(DB *dbp, int error, const char *fmt, ...)
 #else
 __dbh_err(dbp, error, fmt, va_alist)
@@ -263,7 +302,7 @@ __dbh_err(dbp, error, fmt, va_alist)
  *	Error message.
  */
 static void
-#ifdef __STDC__
+#ifdef STDC_HEADERS
 __dbh_errx(DB *dbp, const char *fmt, ...)
 #else
 __dbh_errx(dbp, fmt, va_alist)
@@ -284,12 +323,58 @@ __db_get_byteswapped(dbp, isswapped)
 	DB *dbp;
 	int *isswapped;
 {
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "get_byteswapped");
+	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->get_byteswapped");
 
 	*isswapped = F_ISSET(dbp, DB_AM_SWAP) ? 1 : 0;
 	return (0);
 }
 
+/*
+ * __db_get_dbname --
+ *	Get the name of the database as passed to DB->open.
+ */
+static int
+__db_get_dbname(dbp, fnamep, dnamep)
+	DB *dbp;
+	const char **fnamep, **dnamep;
+{
+	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->get_dbname");
+
+	if (fnamep != NULL)
+		*fnamep = dbp->fname;
+	if (dnamep != NULL)
+		*dnamep = dbp->dname;
+	return (0);
+}
+
+/*
+ * __db_get_env --
+ *	Get the DB_ENV handle that was passed to db_create.
+ */
+static int
+__db_get_env(dbp, dbenvp)
+	DB *dbp;
+	DB_ENV **dbenvp;
+{
+	*dbenvp = dbp->dbenv;
+	return (0);
+}
+
+/*
+ * get_transactional --
+ *	Get whether this database was created in a transaction.
+ */
+static int
+__db_get_transactional(dbp, istxnp)
+	DB *dbp;
+	int *istxnp;
+{
+	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->get_transactional");
+
+	*istxnp = F_ISSET(dbp, DB_AM_TXN) ? 1 : 0;
+	return (0);
+}
+
 /*
  * __db_get_type --
  *	Return type of underlying database.
@@ -299,35 +384,12 @@ __db_get_type(dbp, dbtype)
 	DB *dbp;
 	DBTYPE *dbtype;
 {
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "get_type");
+	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->get_type");
 
 	*dbtype = dbp->type;
 	return (0);
 }
 
-/*
- * __db_key_range --
- *	Return proportion of keys above and below given key.
- */
-static int
-__db_key_range(dbp, txn, key, kr, flags)
-	DB *dbp;
-	DB_TXN *txn;
-	DBT *key;
-	DB_KEY_RANGE *kr;
-	u_int32_t flags;
-{
-	COMPQUIET(txn, NULL);
-	COMPQUIET(key, NULL);
-	COMPQUIET(kr, NULL);
-	COMPQUIET(flags, 0);
-
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "key_range");
-	DB_ILLEGAL_METHOD(dbp, DB_OK_BTREE);
-
-	return (EINVAL);
-}
-
 /*
  * __db_set_append_recno --
  *	Set record number append routine.
@@ -337,7 +399,7 @@ __db_set_append_recno(dbp, func)
 	DB *dbp;
 	int (*func) __P((DB *, DBT *, db_recno_t));
 {
-	DB_ILLEGAL_AFTER_OPEN(dbp, "set_append_recno");
+	DB_ILLEGAL_AFTER_OPEN(dbp, "DB->set_append_recno");
 	DB_ILLEGAL_METHOD(dbp, DB_OK_QUEUE | DB_OK_RECNO);
 
 	dbp->db_append_recno = func;
@@ -345,6 +407,22 @@ __db_set_append_recno(dbp, func)
 	return (0);
 }
 
+/*
+ * __db_get_cachesize --
+ *	Get underlying cache size.
+ */
+static int
+__db_get_cachesize(dbp, cache_gbytesp, cache_bytesp, ncachep)
+	DB *dbp;
+	u_int32_t *cache_gbytesp, *cache_bytesp;
+	int *ncachep;
+{
+	DB_ILLEGAL_IN_ENV(dbp, "DB->get_cachesize");
+
+	return (__memp_get_cachesize(dbp->dbenv,
+	    cache_gbytesp, cache_bytesp, ncachep));
+}
+
 /*
  * __db_set_cachesize --
  *	Set underlying cache size.
@@ -355,33 +433,13 @@ __db_set_cachesize(dbp, cache_gbytes, cache_bytes, ncache)
 	u_int32_t cache_gbytes, cache_bytes;
 	int ncache;
 {
-	DB_ILLEGAL_IN_ENV(dbp, "set_cachesize");
-	DB_ILLEGAL_AFTER_OPEN(dbp, "set_cachesize");
+	DB_ILLEGAL_IN_ENV(dbp, "DB->set_cachesize");
+	DB_ILLEGAL_AFTER_OPEN(dbp, "DB->set_cachesize");
 
-	return (dbp->dbenv->set_cachesize(
+	return (__memp_set_cachesize(
 	    dbp->dbenv, cache_gbytes, cache_bytes, ncache));
 }
 
-/*
- * __db_set_cache_priority --
- *	Set cache priority for pages from this file.
- */
-static int
-__db_set_cache_priority(dbp, priority)
-	DB *dbp;
-	DB_CACHE_PRIORITY priority;
-{
-	/*
-	 * If an underlying DB_MPOOLFILE exists, call it.  Otherwise, save
-	 * the information away until DB->open is called.
-	 */
-	if (dbp->mpf == NULL) {
-		dbp->priority = priority;
-		return (0);
-	}
-	return (dbp->mpf->set_priority(dbp->mpf, priority));
-}
-
 /*
  * __db_set_dup_compare --
  *	Set duplicate comparison routine.
@@ -393,10 +451,10 @@ __db_set_dup_compare(dbp, func)
 {
 	int ret;
 
-	DB_ILLEGAL_AFTER_OPEN(dbp, "dup_compare");
+	DB_ILLEGAL_AFTER_OPEN(dbp, "DB->dup_compare");
 	DB_ILLEGAL_METHOD(dbp, DB_OK_BTREE | DB_OK_HASH);
 
-	if ((ret = dbp->set_flags(dbp, DB_DUPSORT)) != 0)
+	if ((ret = __db_set_flags(dbp, DB_DUPSORT)) != 0)
 		return (ret);
 
 	dbp->dup_compare = func;
@@ -404,6 +462,19 @@ __db_set_dup_compare(dbp, func)
 	return (0);
 }
 
+/*
+ * __db_get_encrypt_flags --
+ */
+static int
+__db_get_encrypt_flags(dbp, flagsp)
+	DB *dbp;
+	u_int32_t *flagsp;
+{
+	DB_ILLEGAL_IN_ENV(dbp, "DB->get_encrypt_flags");
+
+	return (__dbenv_get_encrypt_flags(dbp->dbenv, flagsp));
+}
+
 /*
  * __db_set_encrypt --
  *	Set database passwd.
@@ -417,10 +488,10 @@ __db_set_encrypt(dbp, passwd, flags)
 	DB_CIPHER *db_cipher;
 	int ret;
 
-	DB_ILLEGAL_IN_ENV(dbp, "set_encrypt");
-	DB_ILLEGAL_AFTER_OPEN(dbp, "set_encrypt");
+	DB_ILLEGAL_IN_ENV(dbp, "DB->set_encrypt");
+	DB_ILLEGAL_AFTER_OPEN(dbp, "DB->set_encrypt");
 
-	if ((ret = dbp->dbenv->set_encrypt(dbp->dbenv, passwd, flags)) != 0)
+	if ((ret = __dbenv_set_encrypt(dbp->dbenv, passwd, flags)) != 0)
 		return (ret);
 
 	/*
@@ -432,7 +503,7 @@ __db_set_encrypt(dbp, passwd, flags)
 	    (ret = db_cipher->init(dbp->dbenv, db_cipher)) != 0)
 		return (ret);
 
-	return (dbp->set_flags(dbp, DB_ENCRYPT));
+	return (__db_set_flags(dbp, DB_ENCRYPT));
 }
 
 static void
@@ -440,7 +511,15 @@ __db_set_errcall(dbp, errcall)
 	DB *dbp;
 	void (*errcall) __P((const char *, char *));
 {
-	dbp->dbenv->set_errcall(dbp->dbenv, errcall);
+	__dbenv_set_errcall(dbp->dbenv, errcall);
+}
+
+static void
+__db_get_errfile(dbp, errfilep)
+	DB *dbp;
+	FILE **errfilep;
+{
+	__dbenv_get_errfile(dbp->dbenv, errfilep);
 }
 
 static void
@@ -448,7 +527,15 @@ __db_set_errfile(dbp, errfile)
 	DB *dbp;
 	FILE *errfile;
 {
-	dbp->dbenv->set_errfile(dbp->dbenv, errfile);
+	__dbenv_set_errfile(dbp->dbenv, errfile);
+}
+
+static void
+__db_get_errpfx(dbp, errpfxp)
+	DB *dbp;
+	const char **errpfxp;
+{
+	__dbenv_get_errpfx(dbp->dbenv, errpfxp);
 }
 
 static void
@@ -456,7 +543,7 @@ __db_set_errpfx(dbp, errpfx)
 	DB *dbp;
 	const char *errpfx;
 {
-	dbp->dbenv->set_errpfx(dbp->dbenv, errpfx);
+	__dbenv_set_errpfx(dbp->dbenv, errpfx);
 }
 
 static int
@@ -468,42 +555,127 @@ __db_set_feedback(dbp, feedback)
 	return (0);
 }
 
+/*
+ * __db_map_flags --
+ *	Maps between public and internal flag values.
+ *      This function doesn't check for validity, so it can't fail.
+ */
+static void
+__db_map_flags(dbp, inflagsp, outflagsp)
+	DB *dbp;
+	u_int32_t *inflagsp, *outflagsp;
+{
+	COMPQUIET(dbp, NULL);
+
+	if (FLD_ISSET(*inflagsp, DB_CHKSUM)) {
+		FLD_SET(*outflagsp, DB_AM_CHKSUM);
+		FLD_CLR(*inflagsp, DB_CHKSUM);
+	}
+	if (FLD_ISSET(*inflagsp, DB_ENCRYPT)) {
+		FLD_SET(*outflagsp, DB_AM_ENCRYPT | DB_AM_CHKSUM);
+		FLD_CLR(*inflagsp, DB_ENCRYPT);
+	}
+	if (FLD_ISSET(*inflagsp, DB_TXN_NOT_DURABLE)) {
+		FLD_SET(*outflagsp, DB_AM_NOT_DURABLE);
+		FLD_CLR(*inflagsp, DB_TXN_NOT_DURABLE);
+	}
+}
+
 static int
+__db_get_flags(dbp, flagsp)
+	DB *dbp;
+	u_int32_t *flagsp;
+{
+	static const u_int32_t db_flags[] = {
+		DB_CHKSUM,
+		DB_DUP,
+		DB_DUPSORT,
+		DB_ENCRYPT,
+		DB_RECNUM,
+		DB_RENUMBER,
+		DB_REVSPLITOFF,
+		DB_SNAPSHOT,
+		DB_TXN_NOT_DURABLE,
+		0
+	};
+	u_int32_t f, flags, mapped_flag;
+	int i;
+
+	flags = 0;
+	for (i = 0; (f = db_flags[i]) != 0; i++) {
+		mapped_flag = 0;
+		__db_map_flags(dbp, &f, &mapped_flag);
+		__bam_map_flags(dbp, &f, &mapped_flag);
+		__ram_map_flags(dbp, &f, &mapped_flag);
+		DB_ASSERT(f == 0);
+		if (F_ISSET(dbp, mapped_flag) == mapped_flag)
+			LF_SET(db_flags[i]);
+	}
+
+	*flagsp = flags;
+	return (0);
+}
+
+/*
+ * __db_set_flags --
+ *	DB->set_flags.
+ *
+ * PUBLIC: int  __db_set_flags __P((DB *, u_int32_t));
+ */
+int
 __db_set_flags(dbp, flags)
 	DB *dbp;
 	u_int32_t flags;
 {
+	DB_ENV *dbenv;
 	int ret;
 
-	/*
-	 * !!!
-	 * The hash access method only takes two flags: DB_DUP and DB_DUPSORT.
-	 * The Btree access method uses them for the same purposes, and so we
-	 * resolve them there.
-	 *
-	 * The queue access method takes no flags.
-	 */
-	if (LF_ISSET(DB_ENCRYPT)) {
-		if (!CRYPTO_ON(dbp->dbenv)) {
-			__db_err(dbp->dbenv,
+	dbenv = dbp->dbenv;
+
+	if (LF_ISSET(DB_ENCRYPT) && !CRYPTO_ON(dbenv)) {
+		__db_err(dbenv,
 		    "Database environment not configured for encryption");
-			return (EINVAL);
-		}
-		F_SET(dbp, DB_AM_ENCRYPT);
-		F_SET(dbp, DB_AM_CHKSUM);
-		LF_CLR(DB_ENCRYPT);
-	}
-	if (LF_ISSET(DB_CHKSUM_SHA1)) {
-		F_SET(dbp, DB_AM_CHKSUM);
-		LF_CLR(DB_CHKSUM_SHA1);
+		return (EINVAL);
 	}
+	if (LF_ISSET(DB_TXN_NOT_DURABLE))
+		ENV_REQUIRES_CONFIG(dbenv,
+		    dbenv->tx_handle, "DB_NOT_DURABLE", DB_INIT_TXN);
+
+	__db_map_flags(dbp, &flags, &dbp->flags);
 
 	if ((ret = __bam_set_flags(dbp, &flags)) != 0)
 		return (ret);
 	if ((ret = __ram_set_flags(dbp, &flags)) != 0)
 		return (ret);
 
-	return (flags == 0 ? 0 : __db_ferr(dbp->dbenv, "DB->set_flags", 0));
+	return (flags == 0 ? 0 : __db_ferr(dbenv, "DB->set_flags", 0));
+}
+
+/*
+ * __db_get_lorder --
+ *	Get whether lorder is swapped or not.
+ */
+static int
+__db_get_lorder(dbp, db_lorderp)
+	DB *dbp;
+	int *db_lorderp;
+{
+	int ret;
+
+	/* Flag if the specified byte order requires swapping. */
+	switch (ret = __db_byteorder(dbp->dbenv, 1234)) {
+	case 0:
+		*db_lorderp = F_ISSET(dbp, DB_AM_SWAP) ? 4321 : 1234;
+		break;
+	case DB_SWAPBYTES:
+		*db_lorderp = F_ISSET(dbp, DB_AM_SWAP) ? 1234 : 4321;
+		break;
+	default:
+		return (ret);
+		/* NOTREACHED */
+	}
+
+	return (0);
 }
 
 /*
@@ -519,7 +691,7 @@ __db_set_lorder(dbp, db_lorder)
 {
 	int ret;
 
-	DB_ILLEGAL_AFTER_OPEN(dbp, "set_lorder");
+	DB_ILLEGAL_AFTER_OPEN(dbp, "DB->set_lorder");
 
 	/* Flag if the specified byte order requires swapping. */
 	switch (ret = __db_byteorder(dbp->dbenv, db_lorder)) {
@@ -543,19 +715,33 @@ __db_set_alloc(dbp, mal_func, real_func, free_func)
 	void *(*real_func) __P((void *, size_t));
 	void (*free_func) __P((void *));
 {
-	DB_ILLEGAL_IN_ENV(dbp, "set_alloc");
-	DB_ILLEGAL_AFTER_OPEN(dbp, "set_alloc");
+	DB_ILLEGAL_IN_ENV(dbp, "DB->set_alloc");
+	DB_ILLEGAL_AFTER_OPEN(dbp, "DB->set_alloc");
 
-	return (dbp->dbenv->set_alloc(dbp->dbenv,
-	    mal_func, real_func, free_func));
+	return (__dbenv_set_alloc(dbp->dbenv, mal_func, real_func, free_func));
 }
 
 static int
+__db_get_pagesize(dbp, db_pagesizep)
+	DB *dbp;
+	u_int32_t *db_pagesizep;
+{
+	*db_pagesizep = dbp->pgsize;
+	return (0);
+}
+
+/*
+ * __db_set_pagesize --
+ *	DB->set_pagesize
+ *
+ * PUBLIC: int  __db_set_pagesize __P((DB *, u_int32_t));
+ */
+int
 __db_set_pagesize(dbp, db_pagesize)
 	DB *dbp;
 	u_int32_t db_pagesize;
 {
-	DB_ILLEGAL_AFTER_OPEN(dbp, "set_pagesize");
+	DB_ILLEGAL_AFTER_OPEN(dbp, "DB->set_pagesize");
 
 	if (db_pagesize < DB_MIN_PGSIZE) {
 		__db_err(dbp->dbenv, "page sizes may not be smaller than %lu",
@@ -592,27 +778,7 @@ __db_set_paniccall(dbp, paniccall)
 	DB *dbp;
 	void (*paniccall) __P((DB_ENV *, int));
 {
-	return (dbp->dbenv->set_paniccall(dbp->dbenv, paniccall));
-}
-
-static int
-__db_stat_fail(dbp, sp, flags)
-	DB *dbp;
-	void *sp;
-	u_int32_t flags;
-{
-	COMPQUIET(sp, NULL);
-	COMPQUIET(flags, 0);
-
-	/*
-	 * DB->stat isn't initialized until the actual DB->open call,
-	 * but we don't want to core dump.
-	 */
-	PANIC_CHECK(dbp->dbenv);
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "DB->stat");
-
-	/* NOTREACHED */
-	return (EINVAL);
+	return (__dbenv_set_paniccall(dbp->dbenv, paniccall));
 }
 
 #ifdef HAVE_RPC
@@ -642,9 +808,12 @@ __dbcl_init(dbp, dbenv, flags)
 	dbp->fd = __dbcl_db_fd;
 	dbp->get = __dbcl_db_get;
 	dbp->get_byteswapped = __db_get_byteswapped;
+	dbp->get_transactional = __db_get_transactional;
 	dbp->get_type = __db_get_type;
 	dbp->join = __dbcl_db_join;
 	dbp->key_range = __dbcl_db_key_range;
+	dbp->get_dbname = __dbcl_db_get_name;
+	dbp->get_open_flags = __dbcl_db_get_open_flags;
 	dbp->open = __dbcl_db_open_wrap;
 	dbp->pget = __dbcl_db_pget;
 	dbp->put = __dbcl_db_put;
@@ -652,16 +821,22 @@ __dbcl_init(dbp, dbenv, flags)
 	dbp->rename = __dbcl_db_rename;
 	dbp->set_alloc = __dbcl_db_alloc;
 	dbp->set_append_recno = __dbcl_db_set_append_recno;
+	dbp->get_cachesize = __dbcl_db_get_cachesize;
 	dbp->set_cachesize = __dbcl_db_cachesize;
-	dbp->set_cache_priority = __dbcl_db_cache_priority;
 	dbp->set_dup_compare = __dbcl_db_dup_compare;
+	dbp->get_encrypt_flags = __dbcl_db_get_encrypt_flags;
 	dbp->set_encrypt = __dbcl_db_encrypt;
 	dbp->set_errcall = __db_set_errcall;
+	dbp->get_errfile = __db_get_errfile;
 	dbp->set_errfile = __db_set_errfile;
+	dbp->get_errpfx = __db_get_errpfx;
 	dbp->set_errpfx = __db_set_errpfx;
 	dbp->set_feedback = __dbcl_db_feedback;
+	dbp->get_flags = __dbcl_db_get_flags;
 	dbp->set_flags = __dbcl_db_flags;
+	dbp->get_lorder = __dbcl_db_get_lorder;
 	dbp->set_lorder = __dbcl_db_lorder;
+	dbp->get_pagesize = __dbcl_db_get_pagesize;
 	dbp->set_pagesize = __dbcl_db_pagesize;
 	dbp->set_paniccall = __dbcl_db_panic;
 	dbp->stat = __dbcl_db_stat;
@@ -675,15 +850,23 @@ __dbcl_init(dbp, dbenv, flags)
 	 */
 	dbp->set_bt_compare = __dbcl_db_bt_compare;
 	dbp->set_bt_maxkey = __dbcl_db_bt_maxkey;
+	dbp->get_bt_minkey = __dbcl_db_get_bt_minkey;
 	dbp->set_bt_minkey = __dbcl_db_bt_minkey;
 	dbp->set_bt_prefix = __dbcl_db_bt_prefix;
+	dbp->get_h_ffactor = __dbcl_db_get_h_ffactor;
 	dbp->set_h_ffactor = __dbcl_db_h_ffactor;
 	dbp->set_h_hash = __dbcl_db_h_hash;
+	dbp->get_h_nelem = __dbcl_db_get_h_nelem;
 	dbp->set_h_nelem = __dbcl_db_h_nelem;
+	dbp->get_q_extentsize = __dbcl_db_get_extentsize;
 	dbp->set_q_extentsize = __dbcl_db_extentsize;
+	dbp->get_re_delim = __dbcl_db_get_re_delim;
 	dbp->set_re_delim = __dbcl_db_re_delim;
+	dbp->get_re_len = __dbcl_db_get_re_len;
 	dbp->set_re_len = __dbcl_db_re_len;
+	dbp->get_re_pad = __dbcl_db_get_re_pad;
 	dbp->set_re_pad = __dbcl_db_re_pad;
+	dbp->get_re_source = __dbcl_db_get_re_source;
 	dbp->set_re_source = __dbcl_db_re_source;
 
 	return (__dbcl_db_create(dbp, dbenv, flags));
