@@ -1,14 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2002
+ * Copyright (c) 1996-2003
  *	Sleepycat Software.  All rights reserved.
  */
 
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: db_reclaim.c,v 11.28 2002/08/06 06:11:17 bostic Exp $";
+static const char revid[] = "$Id: db_reclaim.c,v 11.37 2003/06/30 17:19:47 bostic Exp $";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -20,7 +20,7 @@ static const char revid[] = "$Id: db_reclaim.c,v 11.28 2002/08/06 06:11:17 bosti
 #include "dbinc/db_page.h"
 #include "dbinc/db_shash.h"
 #include "dbinc/btree.h"
-#include "dbinc/lock.h"
+#include "dbinc/mp.h"
 
 /*
  * __db_traverse_big
@@ -48,12 +48,12 @@ __db_traverse_big(dbp, pgno, callback, cookie)
 
 	do {
 		did_put = 0;
-		if ((ret = mpf->get(mpf, &pgno, 0, &p)) != 0)
+		if ((ret = __memp_fget(mpf, &pgno, 0, &p)) != 0)
 			return (ret);
 		pgno = NEXT_PGNO(p);
 		if ((ret = callback(dbp, p, cookie, &did_put)) == 0 &&
 		    !did_put)
-			ret = mpf->put(mpf, p, 0);
+			ret = __memp_fput(mpf, p, 0);
 	} while (ret == 0 && pgno != PGNO_INVALID);
 
 	return (ret);
@@ -103,12 +103,8 @@ __db_truncate_callback(dbp, p, cookie, putp)
 	void *cookie;
 	int *putp;
 {
-	DBMETA *meta;
-	DBT ldbt;
-	DB_LOCK metalock;
 	DB_MPOOLFILE *mpf;
 	db_indx_t indx, len, off, tlen, top;
-	db_pgno_t pgno;
 	db_trunc_param *param;
 	u_int8_t *hk, type;
 	int ret;
@@ -147,7 +143,12 @@ __db_truncate_callback(dbp, p, cookie, putp)
 			*putp = 0;
 		break;
 	case P_LRECNO:
-		param->count += top;
+		for (indx = 0; indx < top; indx += O_INDX) {
+			type = GET_BKEYDATA(dbp, p, indx)->type;
+			if (!B_DISSET(type))
+				++param->count;
+		}
+
 		if (((BTREE *)dbp->bt_internal)->bt_root == PGNO(p)) {
 			type = P_LRECNO;
 			goto reinit;
@@ -180,6 +181,9 @@ __db_truncate_callback(dbp, p, cookie, putp)
 					    HKEYDATA_DATA(hk)
 					    + off, sizeof(db_indx_t));
 				}
+				break;
+			default:
+				return (__db_pgfmt(dbp->dbenv, p->pgno));
 			}
 		}
 		/* Don't free the head of the bucket. */
@@ -188,41 +192,9 @@ __db_truncate_callback(dbp, p, cookie, putp)
 
 reinit:			*putp = 0;
 			if (DBC_LOGGING(param->dbc)) {
-				pgno = PGNO_BASE_MD;
-				if ((ret = __db_lget(param->dbc, LCK_ALWAYS,
-				    pgno, DB_LOCK_WRITE, 0, &metalock)) != 0)
+				if ((ret = __db_free(param->dbc, p)) != 0)
 					return (ret);
-				if ((ret = mpf->get(mpf,
-				    &pgno, 0, (PAGE **)&meta)) != 0) {
-					goto err;
-				}
-				memset(&ldbt, 0, sizeof(ldbt));
-				ldbt.data = p;
-				ldbt.size = P_OVERHEAD(dbp);
-				if ((ret = __db_pg_free_log(dbp,
-				    param->dbc->txn, &LSN(meta), 0,
-				    p->pgno, &LSN(meta),
-				    PGNO_BASE_MD, &ldbt, meta->free)) != 0)
-					goto err;
-				LSN(p) = LSN(meta);
-
-				if ((ret =
-				    __db_pg_alloc_log(dbp,
-				    param->dbc->txn, &LSN(meta), 0,
-				    &LSN(meta), PGNO_BASE_MD,
-				    &p->lsn, p->pgno, type, meta->free)) != 0) {
-err:					(void)mpf->put(mpf, (PAGE *)meta, 0);
-					(void)__TLPUT(param->dbc, metalock);
-					return (ret);
-				}
-				LSN(p) = LSN(meta);
-
-				if ((ret = mpf->put(mpf,
-				    (PAGE *)meta, DB_MPOOL_DIRTY)) != 0) {
-					(void)__TLPUT(param->dbc, metalock);
-					return (ret);
-				}
-				if ((ret = __TLPUT(param->dbc, metalock)) != 0)
+				if ((ret = __db_new(param->dbc, type, &p)) != 0)
 					return (ret);
 			} else
 				LSN_NOT_LOGGED(LSN(p));
@@ -239,7 +211,7 @@ err:					(void)mpf->put(mpf, (PAGE *)meta, 0);
 		if ((ret = __db_free(param->dbc, p)) != 0)
 			return (ret);
 	} else {
-		if ((ret = mpf->put(mpf, p, DB_MPOOL_DIRTY)) != 0)
+		if ((ret = __memp_fput(mpf, p, DB_MPOOL_DIRTY)) != 0)
 			return (ret);
 		*putp = 1;
 	}
