@@ -1,20 +1,19 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2002
+ * Copyright (c) 1996-2003
  *	Sleepycat Software.  All rights reserved.
  */
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: log_get.c,v 11.81 2002/08/14 20:09:27 bostic Exp $";
+static const char revid[] = "$Id: log_get.c,v 11.98 2003/09/13 19:20:38 bostic Exp $";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
 #include <sys/types.h>
 
 #include <string.h>
-#include <unistd.h>
 #endif
 
 #include "db_int.h"
@@ -26,10 +25,10 @@ static const char revid[] = "$Id: log_get.c,v 11.81 2002/08/14 20:09:27 bostic E
 
 typedef enum { L_ALREADY, L_ACQUIRED, L_NONE } RLOCK;
 
-static int __log_c_close __P((DB_LOGC *, u_int32_t));
-static int __log_c_get __P((DB_LOGC *, DB_LSN *, DBT *, u_int32_t));
+static int __log_c_close_pp __P((DB_LOGC *, u_int32_t));
+static int __log_c_get_pp __P((DB_LOGC *, DB_LSN *, DBT *, u_int32_t));
 static int __log_c_get_int __P((DB_LOGC *, DB_LSN *, DBT *, u_int32_t));
-static int __log_c_hdrchk __P((DB_LOGC *, HDR *, int *));
+static int __log_c_hdrchk __P((DB_LOGC *, DB_LSN *, HDR *, int *));
 static int __log_c_incursor __P((DB_LOGC *, DB_LSN *, HDR *, u_int8_t **));
 static int __log_c_inregion __P((DB_LOGC *,
 	       DB_LSN *, RLOCK *, DB_LSN *, HDR *, u_int8_t **));
@@ -38,70 +37,88 @@ static int __log_c_io __P((DB_LOGC *,
 static int __log_c_ondisk __P((DB_LOGC *,
 	       DB_LSN *, DB_LSN *, int, HDR *, u_int8_t **, int *));
 static int __log_c_set_maxrec __P((DB_LOGC *, char *));
-static int __log_c_shortread __P((DB_LOGC *, int));
+static int __log_c_shortread __P((DB_LOGC *, DB_LSN *, int));
 
 /*
- * __log_cursor --
- *	Create a log cursor.
+ * __log_cursor_pp --
+ *	DB_ENV->log_cursor
  *
- * PUBLIC: int __log_cursor __P((DB_ENV *, DB_LOGC **, u_int32_t));
+ * PUBLIC: int __log_cursor_pp __P((DB_ENV *, DB_LOGC **, u_int32_t));
  */
 int
-__log_cursor(dbenv, logcp, flags)
+__log_cursor_pp(dbenv, logcp, flags)
 	DB_ENV *dbenv;
 	DB_LOGC **logcp;
 	u_int32_t flags;
 {
-	DB_LOGC *logc;
-	int ret;
+	int rep_check, ret;
 
 	PANIC_CHECK(dbenv);
 	ENV_REQUIRES_CONFIG(dbenv,
 	    dbenv->lg_handle, "DB_ENV->log_cursor", DB_INIT_LOG);
 
-	*logcp = NULL;
-
 	/* Validate arguments. */
 	if ((ret = __db_fchk(dbenv, "DB_ENV->log_cursor", flags, 0)) != 0)
 		return (ret);
 
-	/* Allocate memory for the cursor. */
-	if ((ret = __os_calloc(dbenv, 1, sizeof(DB_LOGC), &logc)) != 0)
-		goto err;
-	if ((ret = __os_calloc(dbenv, 1, sizeof(DB_FH), &logc->c_fh)) != 0)
-		goto err;
-
-	logc->bp_size = DB_LOGC_BUF_SIZE;
-	if ((ret = __os_malloc(dbenv, logc->bp_size, &logc->bp)) != 0)
-		goto err;
-
-	logc->dbenv = dbenv;
-	logc->close = __log_c_close;
-	logc->get = __log_c_get;
-
-	*logcp = logc;
-	return (0);
-
-err:	if (logc != NULL) {
-		if (logc->c_fh != NULL)
-			__os_free(dbenv, logc->c_fh);
-		__os_free(dbenv, logc);
-	}
-
+	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
+	if (rep_check)
+		__env_rep_enter(dbenv);
+	ret = __log_cursor(dbenv, logcp);
+	if (rep_check)
+		__env_rep_exit(dbenv);
 	return (ret);
 }
 
 /*
- * __log_c_close --
- *	Close a log cursor.
+ * __log_cursor --
+ *	Create a log cursor.
+ *
+ * PUBLIC: int __log_cursor __P((DB_ENV *, DB_LOGC **));
+ */
+int
+__log_cursor(dbenv, logcp)
+	DB_ENV *dbenv;
+	DB_LOGC **logcp;
+{
+	DB_LOGC *logc;
+	int ret;
+
+	*logcp = NULL;
+
+	/* Allocate memory for the cursor. */
+	if ((ret = __os_calloc(dbenv, 1, sizeof(DB_LOGC), &logc)) != 0)
+		return (ret);
+
+	logc->bp_size = DB_LOGC_BUF_SIZE;
+	/*
+	 * Set this to something positive.
+	 */
+	logc->bp_maxrec = MEGABYTE;
+	if ((ret = __os_malloc(dbenv, logc->bp_size, &logc->bp)) != 0) {
+		__os_free(dbenv, logc);
+		return (ret);
+	}
+
+	logc->dbenv = dbenv;
+	logc->close = __log_c_close_pp;
+	logc->get = __log_c_get_pp;
+
+	*logcp = logc;
+	return (0);
+}
+
+/*
+ * __log_c_close_pp --
+ *	DB_LOGC->close pre/post processing.
  */
 static int
-__log_c_close(logc, flags)
+__log_c_close_pp(logc, flags)
 	DB_LOGC *logc;
 	u_int32_t flags;
 {
 	DB_ENV *dbenv;
-	int ret;
+	int rep_check, ret;
 
 	dbenv = logc->dbenv;
 
@@ -109,33 +126,56 @@ __log_c_close(logc, flags)
 	if ((ret = __db_fchk(dbenv, "DB_LOGC->close", flags, 0)) != 0)
 		return (ret);
 
-	if (F_ISSET(logc->c_fh, DB_FH_VALID))
-		(void)__os_closehandle(dbenv, logc->c_fh);
+	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
+	if (rep_check)
+		__env_rep_enter(dbenv);
+	ret = __log_c_close(logc);
+	if (rep_check)
+		__env_rep_exit(dbenv);
+	return (ret);
+}
+
+/*
+ * __log_c_close --
+ *	DB_LOGC->close.
+ *
+ * PUBLIC: int __log_c_close __P((DB_LOGC *));
+ */
+int
+__log_c_close(logc)
+	DB_LOGC *logc;
+{
+	DB_ENV *dbenv;
+
+	dbenv = logc->dbenv;
+
+	if (logc->c_fhp != NULL) {
+		(void)__os_closehandle(dbenv, logc->c_fhp);
+		logc->c_fhp = NULL;
+	}
 
 	if (logc->c_dbt.data != NULL)
 		__os_free(dbenv, logc->c_dbt.data);
 
 	__os_free(dbenv, logc->bp);
-	__os_free(dbenv, logc->c_fh);
 	__os_free(dbenv, logc);
 
 	return (0);
 }
 
 /*
- * __log_c_get --
- *	Get a log record.
+ * __log_c_get_pp --
+ *	DB_LOGC->get pre/post processing.
  */
 static int
-__log_c_get(logc, alsn, dbt, flags)
+__log_c_get_pp(logc, alsn, dbt, flags)
 	DB_LOGC *logc;
 	DB_LSN *alsn;
 	DBT *dbt;
 	u_int32_t flags;
 {
 	DB_ENV *dbenv;
-	DB_LSN saved_lsn;
-	int ret;
+	int rep_check, ret;
 
 	dbenv = logc->dbenv;
 
@@ -151,7 +191,8 @@ __log_c_get(logc, alsn, dbt, flags)
 		break;
 	case DB_SET:
 		if (IS_ZERO_LSN(*alsn)) {
-			__db_err(dbenv, "DB_LOGC->get: invalid LSN");
+			__db_err(dbenv, "DB_LOGC->get: invalid LSN: %lu/%lu",
+			    (u_long)alsn->file, (u_long)alsn->offset);
 			return (EINVAL);
 		}
 		break;
@@ -159,6 +200,34 @@ __log_c_get(logc, alsn, dbt, flags)
 		return (__db_ferr(dbenv, "DB_LOGC->get", 1));
 	}
 
+	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
+	if (rep_check)
+		__env_rep_enter(dbenv);
+	ret = __log_c_get(logc, alsn, dbt, flags);
+	if (rep_check)
+		__env_rep_exit(dbenv);
+	return (ret);
+}
+
+/*
+ * __log_c_get --
+ *	DB_LOGC->get.
+ *
+ * PUBLIC: int __log_c_get __P((DB_LOGC *, DB_LSN *, DBT *, u_int32_t));
+ */
+int
+__log_c_get(logc, alsn, dbt, flags)
+	DB_LOGC *logc;
+	DB_LSN *alsn;
+	DBT *dbt;
+	u_int32_t flags;
+{
+	DB_ENV *dbenv;
+	DB_LSN saved_lsn;
+	int ret;
+
+	dbenv = logc->dbenv;
+
 	/*
 	 * On error, we take care not to overwrite the caller's LSN.  This
 	 * is because callers looking for the end of the log loop using the
@@ -297,9 +366,8 @@ __log_c_get_int(logc, alsn, dbt, flags)
 		if (!IS_ZERO_LSN(nlsn)) {
 			/* If at start-of-file, move to the previous file. */
 			if (nlsn.offset == 0) {
-				if (nlsn.file == 1 ||
-				    __log_valid(dblp,
-					nlsn.file - 1, 0, &status) != 0) {
+				if (nlsn.file == 1 || __log_valid(dblp,
+				    nlsn.file - 1, 0, NULL, 0, &status) != 0) {
 					ret = DB_NOTFOUND;
 					goto err;
 				}
@@ -429,7 +497,8 @@ cksum:	/*
 				ret = EIO;
 		} else if (ret == -1) {
 			__db_err(dbenv,
-		    "DB_LOGC->get: log record checksum mismatch");
+		    "DB_LOGC->get: log record LSN %lu/%lu: checksum mismatch",
+			    (u_long)nlsn.file, (u_long)nlsn.offset);
 			__db_err(dbenv,
 		    "DB_LOGC->get: catastrophic recovery may be required");
 			ret = __db_panic(dbenv, DB_RUNRECOVERY);
@@ -515,6 +584,7 @@ __log_c_incursor(logc, lsn, hdr, pp)
 	u_int8_t **pp;
 {
 	u_int8_t *p;
+	int	eof;
 
 	*pp = NULL;
 
@@ -547,9 +617,9 @@ __log_c_incursor(logc, lsn, hdr, pp)
 	 */
 	p = logc->bp + (lsn->offset - logc->bp_lsn.offset);
 	memcpy(hdr, p, hdr->size);
-	if (__log_c_hdrchk(logc, hdr, NULL))
+	if (__log_c_hdrchk(logc, lsn, hdr, &eof))
 		return (DB_NOTFOUND);
-	if (logc->bp_lsn.offset + logc->bp_rlen <= lsn->offset + hdr->len)
+	if (eof || logc->bp_lsn.offset + logc->bp_rlen < lsn->offset + hdr->len)
 		return (0);
 
 	*pp = p;				/* Success. */
@@ -644,7 +714,13 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 
 	/*
 	 * If the requested LSN is greater than the region buffer's first
-	 * byte, we know the entire record is in the buffer.
+	 * byte, we know the entire record is in the buffer on a good LSN.
+	 *
+	 * If we're given a bad LSN, the "entire" record might
+	 * not be in our buffer in order to fail at the chksum.
+	 * __log_c_hdrchk made sure our dest buffer fits, via
+	 * bp_maxrec, but we also need to make sure we don't run off
+	 * the end of this buffer, the src.
 	 *
 	 * If the header check fails for any reason, it must be because the
 	 * LSN is bogus.  Fail hard.
@@ -652,7 +728,9 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 	if (lsn->offset > lp->f_lsn.offset) {
 		p = dblp->bufp + (lsn->offset - lp->w_off);
 		memcpy(hdr, p, hdr->size);
-		if (__log_c_hdrchk(logc, hdr, NULL))
+		if (__log_c_hdrchk(logc, lsn, hdr, NULL))
+			return (DB_NOTFOUND);
+		if (lsn->offset + hdr->len > lp->w_off + lp->buffer_size)
 			return (DB_NOTFOUND);
 		if (logc->bp_size <= hdr->len) {
 			len = ALIGN(hdr->len * 2, 128);
@@ -729,7 +807,7 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 		    logc, lsn->file, lsn->offset, p, &nr, NULL)) != 0)
 			return (ret);
 		if (nr < b_disk)
-			return (__log_c_shortread(logc, 0));
+			return (__log_c_shortread(logc, lsn, 0));
 	}
 
 	/* Copy the header information into the caller's structure. */
@@ -773,14 +851,14 @@ __log_c_ondisk(logc, lsn, last_lsn, flags, hdr, pp, eofp)
 	}
 
 	/* Check the HDR. */
-	if ((ret = __log_c_hdrchk(logc, hdr, eofp)) != 0)
+	if ((ret = __log_c_hdrchk(logc, lsn, hdr, eofp)) != 0)
 		return (ret);
 	if (*eofp)
 		return (0);
 
 	/* Otherwise, we should have gotten the bytes we wanted. */
 	if (nr < hdr->size)
-		return (__log_c_shortread(logc, 0));
+		return (__log_c_shortread(logc, lsn, 1));
 
 	/*
 	 * Regardless of how we return, the previous contents of the cursor's
@@ -831,7 +909,7 @@ __log_c_ondisk(logc, lsn, last_lsn, flags, hdr, pp, eofp)
 	 * record we're reading.
 	 */
 	if (nr < (lsn->offset + hdr->len) - offset)
-		return (__log_c_shortread(logc, 1));
+		return (__log_c_shortread(logc, lsn, 1));
 
 	/* Set up the return information. */
 	logc->bp_rlen = (u_int32_t)nr;
@@ -874,8 +952,9 @@ __log_c_ondisk(logc, lsn, last_lsn, flags, hdr, pp, eofp)
  * the middle of a log record, there's no way to tell.
  */
 static int
-__log_c_hdrchk(logc, hdr, eofp)
+__log_c_hdrchk(logc, lsn, hdr, eofp)
 	DB_LOGC *logc;
+	DB_LSN *lsn;
 	HDR *hdr;
 	int *eofp;
 {
@@ -884,9 +963,24 @@ __log_c_hdrchk(logc, hdr, eofp)
 
 	dbenv = logc->dbenv;
 
-	/* Sanity check the log record's size. */
+	/*
+	 * Check EOF before we do any other processing.
+	 */
+	if (eofp != NULL) {
+		if (hdr->prev == 0 && hdr->chksum[0] == 0 && hdr->len == 0) {
+			*eofp = 1;
+			return (0);
+		}
+		*eofp = 0;
+	}
+
+	/*
+	 * Sanity check the log record's size.
+	 * We must check it after "virtual" EOF above.
+	 */
 	if (hdr->len <= hdr->size)
 		goto err;
+
 	/*
 	 * If the cursor's max-record value isn't yet set, it means we aren't
 	 * reading these records from a log file and no check is necessary.
@@ -904,18 +998,12 @@ __log_c_hdrchk(logc, hdr, eofp)
 		if (logc->bp_maxrec != 0 && hdr->len > logc->bp_maxrec)
 			goto err;
 	}
-
-	if (eofp != NULL) {
-		if (hdr->prev == 0 && hdr->chksum[0] == 0 && hdr->len == 0) {
-			*eofp = 1;
-			return (0);
-		}
-		*eofp = 0;
-	}
 	return (0);
 
 err:	if (!F_ISSET(logc, DB_LOG_SILENT_ERR))
-		__db_err(dbenv, "DB_LOGC->get: invalid log record header");
+		__db_err(dbenv,
+		    "DB_LOGC->get: LSN %lu/%lu: invalid log record header",
+		    (u_long)lsn->file, (u_long)lsn->offset);
 	return (EIO);
 }
 
@@ -943,12 +1031,15 @@ __log_c_io(logc, fnum, offset, p, nrp, eofp)
 	 * If we've switched files, discard the current file handle and acquire
 	 * a new one.
 	 */
-	if (F_ISSET(logc->c_fh, DB_FH_VALID) && logc->bp_lsn.file != fnum)
-		if ((ret = __os_closehandle(dbenv, logc->c_fh)) != 0)
+	if (logc->c_fhp != NULL && logc->bp_lsn.file != fnum) {
+		ret = __os_closehandle(dbenv, logc->c_fhp);
+		logc->c_fhp = NULL;
+		if (ret != 0)
 			return (ret);
-	if (!F_ISSET(logc->c_fh, DB_FH_VALID)) {
+	}
+	if (logc->c_fhp == NULL) {
 		if ((ret = __log_name(dblp, fnum,
-		    &np, logc->c_fh, DB_OSO_RDONLY | DB_OSO_SEQ)) != 0) {
+		    &np, &logc->c_fhp, DB_OSO_RDONLY | DB_OSO_SEQ)) != 0) {
 			/*
 			 * If we're allowed to return EOF, assume that's the
 			 * problem, set the EOF status flag and return 0.
@@ -974,18 +1065,20 @@ __log_c_io(logc, fnum, offset, p, nrp, eofp)
 
 	/* Seek to the record's offset. */
 	if ((ret = __os_seek(dbenv,
-	    logc->c_fh, 0, 0, offset, 0, DB_OS_SEEK_SET)) != 0) {
+	    logc->c_fhp, 0, 0, offset, 0, DB_OS_SEEK_SET)) != 0) {
 		if (!F_ISSET(logc, DB_LOG_SILENT_ERR))
 			__db_err(dbenv,
-			    "DB_LOGC->get: seek: %s", db_strerror(ret));
+			    "DB_LOGC->get: LSN: %lu/%lu: seek: %s",
+			    (u_long)fnum, (u_long)offset, db_strerror(ret));
 		return (ret);
 	}
 
 	/* Read the data. */
-	if ((ret = __os_read(dbenv, logc->c_fh, p, *nrp, nrp)) != 0) {
+	if ((ret = __os_read(dbenv, logc->c_fhp, p, *nrp, nrp)) != 0) {
 		if (!F_ISSET(logc, DB_LOG_SILENT_ERR))
 			__db_err(dbenv,
-			    "DB_LOGC->get: read: %s", db_strerror(ret));
+			    "DB_LOGC->get: LSN: %lu/%lu: read: %s",
+			    (u_long)fnum, (u_long)offset, db_strerror(ret));
 		return (ret);
 	}
 
@@ -997,12 +1090,14 @@ __log_c_io(logc, fnum, offset, p, nrp, eofp)
  *	Read was short -- return a consistent error message and error.
  */
 static int
-__log_c_shortread(logc, silent)
+__log_c_shortread(logc, lsn, check_silent)
 	DB_LOGC *logc;
-	int silent;
+	DB_LSN *lsn;
+	int check_silent;
 {
-	if (!silent || !F_ISSET(logc, DB_LOG_SILENT_ERR))
-		__db_err(logc->dbenv, "DB_LOGC->get: short read");
+	if (!check_silent || !F_ISSET(logc, DB_LOG_SILENT_ERR))
+		__db_err(logc->dbenv, "DB_LOGC->get: LSN: %lu/%lu: short read",
+		    (u_long)lsn->file, (u_long)lsn->offset);
 	return (EIO);
 }
 
@@ -1033,11 +1128,13 @@ __log_c_set_maxrec(logc, np)
 	 * of the file but that's hard -- we may have to decrypt it, checksum
 	 * it and so on.  Stat the file instead.
 	 */
-	if ((ret =
-	    __os_ioinfo(dbenv, np, logc->c_fh, &mbytes, &bytes, NULL)) != 0)
-		return (ret);
-
-	logc->bp_maxrec = mbytes * MEGABYTE + bytes;
+	if (logc->c_fhp != NULL) {
+		if ((ret = __os_ioinfo(dbenv, np, logc->c_fhp,
+		    &mbytes, &bytes, NULL)) != 0)
+			return (ret);
+		if (logc->bp_maxrec < (mbytes * MEGABYTE + bytes))
+			logc->bp_maxrec = mbytes * MEGABYTE + bytes;
+	}
 
 	/*
 	 * If reading from the log file currently being written, we could get
@@ -1052,7 +1149,8 @@ __log_c_set_maxrec(logc, np)
 	 * changed, we don't need a lock on it.
 	 */
 	lp = dblp->reginfo.primary;
-	logc->bp_maxrec += lp->buffer_size;
+	if (logc->bp_maxrec < lp->buffer_size)
+		logc->bp_maxrec = lp->buffer_size;
 
 	return (0);
 }
