@@ -440,6 +440,7 @@ alloc:		/*
 		c_mp->stat.st_pages--;
 		alloc_bhp = NULL;
 		R_UNLOCK(dbenv, &dbmp->reginfo[n_cache]);
+		MUTEX_LOCK(dbenv, &hp->hash_mutex);
 
 		/*
 		 * We can't use the page we found in the pool if DB_MPOOL_NEW
@@ -454,9 +455,6 @@ alloc:		/*
 			b_incr = 0;
 			goto alloc;
 		}
-
-		/* We can use the page -- get the bucket lock. */
-		MUTEX_LOCK(dbenv, &hp->hash_mutex);
 		break;
 	case SECOND_MISS:
 		/*
@@ -555,11 +553,8 @@ alloc:		/*
 	 */
 	if (state != SECOND_MISS && bhp->ref == 1) {
 		bhp->priority = UINT32_T_MAX;
-		if (SH_TAILQ_FIRST(&hp->hash_bucket, __bh) !=
-		     SH_TAILQ_LAST(&hp->hash_bucket, hq, __bh)) {
-			SH_TAILQ_REMOVE(&hp->hash_bucket, bhp, hq, __bh);
-			SH_TAILQ_INSERT_TAIL(&hp->hash_bucket, bhp, hq);
-		}
+		SH_TAILQ_REMOVE(&hp->hash_bucket, bhp, hq, __bh);
+		SH_TAILQ_INSERT_TAIL(&hp->hash_bucket, bhp, hq);
 		hp->hash_priority =
 		    SH_TAILQ_FIRST(&hp->hash_bucket, __bh)->priority;
 	}
