@@ -1,14 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2002
+ * Copyright (c) 1997-2003
  *	Sleepycat Software.  All rights reserved.
  */
 
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: os_spin.c,v 11.11 2002/07/12 18:56:56 bostic Exp $";
+static const char revid[] = "$Id: os_spin.c,v 11.14 2003/06/18 14:21:20 bostic Exp $";
 #endif /* not lint */
 
 #include "db_int.h"
@@ -17,7 +17,7 @@ static const char revid[] = "$Id: os_spin.c,v 11.11 2002/07/12 18:56:56 bostic E
  * __os_spin --
  *	Return the number of default spins before blocking.
  */
-int
+void
 __os_spin(dbenv)
 	DB_ENV *dbenv;
 {
@@ -28,7 +28,7 @@ __os_spin(dbenv)
 	 * out, return it.
 	 */
 	if (dbenv->tas_spins != 0)
-		return (dbenv->tas_spins);
+		return;
 
 	/* Get the number of processors */
 	GetSystemInfo(&SystemInfo);
@@ -41,7 +41,6 @@ __os_spin(dbenv)
 		 dbenv->tas_spins = 50 * SystemInfo.dwNumberOfProcessors;
 	else
 		 dbenv->tas_spins = 1;
-	return (dbenv->tas_spins);
 }
 
 /*
@@ -55,5 +54,5 @@ __os_yield(dbenv, usecs)
 {
 	if (DB_GLOBAL(j_yield) != NULL && DB_GLOBAL(j_yield)() == 0)
 		return;
-	__os_sleep(dbenv, 0, usecs);
+	(void)__os_sleep(dbenv, 0, usecs);
 }
