@@ -1,13 +1,13 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2001-2002
+ * Copyright (c) 2001-2003
  *	Sleepycat Software.  All rights reserved.
  */
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: rep_method.c,v 1.78 2002/09/10 12:58:07 bostic Exp $";
+static const char revid[] = "$Id: rep_method.c,v 1.134 2003/11/13 15:41:51 sue Exp $";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -19,14 +19,12 @@ static const char revid[] = "$Id: rep_method.c,v 1.78 2002/09/10 12:58:07 bostic
 
 #include <stdlib.h>
 #include <string.h>
-#include <unistd.h>
 #endif
 
 #include "db_int.h"
 #include "dbinc/db_page.h"
-#include "dbinc/db_am.h"
+#include "dbinc/btree.h"
 #include "dbinc/log.h"
-#include "dbinc/rep.h"
 #include "dbinc/txn.h"
 
 #ifdef HAVE_RPC
@@ -38,13 +36,16 @@ static int __rep_abort_prepared __P((DB_ENV *));
 static int __rep_bt_cmp __P((DB *, const DBT *, const DBT *));
 static int __rep_client_dbinit __P((DB_ENV *, int));
 static int __rep_elect __P((DB_ENV *, int, int, u_int32_t, int *));
-static int __rep_elect_init __P((DB_ENV *, DB_LSN *, int, int, int, int *));
+static int __rep_elect_init
+	       __P((DB_ENV *, DB_LSN *, int, int, int *, u_int32_t *));
 static int __rep_flush __P((DB_ENV *));
 static int __rep_restore_prepared __P((DB_ENV *));
+static int __rep_get_limit __P((DB_ENV *, u_int32_t *, u_int32_t *));
 static int __rep_set_limit __P((DB_ENV *, u_int32_t, u_int32_t));
 static int __rep_set_request __P((DB_ENV *, u_int32_t, u_int32_t));
 static int __rep_set_rep_transport __P((DB_ENV *, int,
-    int (*)(DB_ENV *, const DBT *, const DBT *, int, u_int32_t)));
+    int (*)(DB_ENV *, const DBT *, const DBT *, const DB_LSN *,
+    int, u_int32_t)));
 static int __rep_start __P((DB_ENV *, DBT *, u_int32_t));
 static int __rep_stat __P((DB_ENV *, DB_REP_STAT **, u_int32_t));
 static int __rep_wait __P((DB_ENV *, u_int32_t, int *, u_int32_t));
@@ -59,18 +60,14 @@ int
 __rep_dbenv_create(dbenv)
 	DB_ENV *dbenv;
 {
-	DB_REP *db_rep;
-	int ret;
-
 #ifdef HAVE_RPC
 	if (F_ISSET(dbenv, DB_ENV_RPCCLIENT)) {
-		COMPQUIET(db_rep, NULL);
-		COMPQUIET(ret, 0);
 		dbenv->rep_elect = __dbcl_rep_elect;
 		dbenv->rep_flush = __dbcl_rep_flush;
 		dbenv->rep_process_message = __dbcl_rep_process_message;
 		dbenv->rep_start = __dbcl_rep_start;
 		dbenv->rep_stat = __dbcl_rep_stat;
+		dbenv->get_rep_limit = __dbcl_rep_get_limit;
 		dbenv->set_rep_limit = __dbcl_rep_set_limit;
 		dbenv->set_rep_request = __dbcl_rep_set_request;
 		dbenv->set_rep_transport = __dbcl_rep_set_rep_transport;
@@ -83,32 +80,60 @@ __rep_dbenv_create(dbenv)
 		dbenv->rep_process_message = __rep_process_message;
 		dbenv->rep_start = __rep_start;
 		dbenv->rep_stat = __rep_stat;
+		dbenv->get_rep_limit = __rep_get_limit;
 		dbenv->set_rep_limit = __rep_set_limit;
 		dbenv->set_rep_request = __rep_set_request;
 		dbenv->set_rep_transport = __rep_set_rep_transport;
-		/*
-		 * !!!
-		 * Our caller has not yet had the opportunity to reset the panic
-		 * state or turn off mutex locking, and so we can neither check
-		 * the panic state or acquire a mutex in the DB_ENV create path.
-		 */
-
-		if ((ret = __os_calloc(dbenv, 1, sizeof(DB_REP), &db_rep)) != 0)
-			return (ret);
-		dbenv->rep_handle = db_rep;
-
-		/* Initialize the per-process replication structure. */
-		db_rep->rep_send = NULL;
 	}
 
 	return (0);
 }
 
+/*
+ * __rep_open --
+ *	Replication-specific initialization of the DB_ENV structure.
+ *
+ * PUBLIC: int __rep_open __P((DB_ENV *));
+ */
+int
+__rep_open(dbenv)
+	DB_ENV *dbenv;
+{
+	DB_REP *db_rep;
+	int ret;
+
+	if ((ret = __os_calloc(dbenv, 1, sizeof(DB_REP), &db_rep)) != 0)
+		return (ret);
+	dbenv->rep_handle = db_rep;
+	ret = __rep_region_init(dbenv);
+	return (ret);
+}
+
 /*
  * __rep_start --
  *	Become a master or client, and start sending messages to participate
  * in the replication environment.  Must be called after the environment
  * is open.
+ *
+ * We must protect rep_start, which may change the world, with the rest
+ * of the DB library.  Each API interface will count itself as it enters
+ * the library.  Rep_start checks the following:
+ *
+ * rep->msg_th - this is the count of threads currently in rep_process_message
+ * rep->start_th - this is set if a thread is in rep_start.
+ * rep->handle_cnt - number of threads actively using a dbp in library.
+ * rep->txn_cnt - number of active txns.
+ * REP_F_READY - Replication flag that indicates that we wish to run
+ * recovery, and want to prohibit new transactions from entering and cause
+ * existing ones to return immediately (with a DB_LOCK_DEADLOCK error).
+ *
+ * There is also the rep->timestamp which is updated whenever significant
+ * events (i.e., new masters, log rollback, etc).  Upon creation, a handle
+ * is associated with the current timestamp.  Each time a handle enters the
+ * library it must check if the handle timestamp is the same as the one
+ * stored in the replication region.  This prevents the use of handles on
+ * clients that reference non-existent files whose creation was backed out
+ * during a synchronizing recovery.
  */
 static int
 __rep_start(dbenv, dbt, flags)
@@ -120,11 +145,12 @@ __rep_start(dbenv, dbt, flags)
 	DB_LSN lsn;
 	DB_REP *db_rep;
 	REP *rep;
-	int announce, init_db, redo_prepared, ret;
+	u_int32_t repflags;
+	int announce, init_db, redo_prepared, ret, sleep_cnt, t_ret, was_client;
 
 	PANIC_CHECK(dbenv);
-	ENV_ILLEGAL_BEFORE_OPEN(dbenv, "rep_start");
-	ENV_REQUIRES_CONFIG(dbenv, dbenv->tx_handle, "rep_stat", DB_INIT_TXN);
+	ENV_ILLEGAL_BEFORE_OPEN(dbenv, "DB_ENV->rep_start");
+	ENV_REQUIRES_CONFIG(dbenv, dbenv->rep_handle, "rep_start", DB_INIT_REP);
 
 	db_rep = dbenv->rep_handle;
 	rep = db_rep->region;
@@ -149,24 +175,59 @@ __rep_start(dbenv, dbt, flags)
 		return (ret);
 
 	/* We need a transport function. */
-	if (db_rep->rep_send == NULL) {
+	if (dbenv->rep_send == NULL) {
 		__db_err(dbenv,
     "DB_ENV->set_rep_transport must be called before DB_ENV->rep_start");
 		return (EINVAL);
 	}
-	
-	/* We'd better not have any logged files open if we are a client. */
-	if (LF_ISSET(DB_REP_CLIENT) && (ret = __dbreg_nofiles(dbenv)) != 0) {
-		__db_err(dbenv, "DB_ENV->rep_start called with open files");
+
+	/*
+	 * If we are about to become (or stay) a master.  Let's flush the log
+	 * to close any potential holes that might happen when upgrading from
+	 * client to master status.
+	 */
+	if (LF_ISSET(DB_REP_MASTER) && (ret = __log_flush(dbenv, NULL)) != 0)
 		return (ret);
+
+	MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
+	/*
+	 * We only need one thread to start-up replication, so if
+	 * there is another thread in rep_start, we'll let it finish
+	 * its work and have this thread simply return.
+	 */
+	if (rep->start_th != 0) {
+		/*
+		 * There is already someone in rep_start.  Return.
+		 */
+#ifdef DIAGNOSTIC
+		if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION))
+			__db_err(dbenv, "Thread already in rep_start");
+#endif
+		goto err;
+	} else
+		rep->start_th = 1;
+
+	/*
+	 * We are about to delete the replication database.  We need to
+	 * make sure that no one else is in the midst of processing a
+	 * replication message.
+	 */
+	for (sleep_cnt = 0; rep->msg_th != 0;) {
+		if (++sleep_cnt % 60 == 0)
+			__db_err(dbenv,
+	"DB_ENV->rep_start waiting %d minutes for replication message thread",
+			    sleep_cnt / 60);
+		MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
+		(void)__os_sleep(dbenv, 1, 0);
+		MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
 	}
 
-	MUTEX_LOCK(dbenv, db_rep->mutexp);
 	if (rep->eid == DB_EID_INVALID)
 		rep->eid = dbenv->rep_eid;
 
 	if (LF_ISSET(DB_REP_MASTER)) {
-		if (F_ISSET(dbenv, DB_ENV_REP_CLIENT)) {
+		was_client = F_ISSET(rep, REP_F_UPGRADE);
+		if (was_client) {
 			/*
 			 * If we're upgrading from having been a client,
 			 * preclose, so that we close our temporary database.
@@ -180,72 +241,92 @@ __rep_start(dbenv, dbt, flags)
 			 * that opened them becomes a master again.
 			 */
 			if ((ret = __rep_preclose(dbenv, 0)) != 0)
-				return (ret);
-
-			/*
-			 * Now write a __txn_recycle record so that
-			 * clients don't get confused with our txnids
-			 * and txnids of previous masters.
-			 */
-			F_CLR(dbenv, DB_ENV_REP_CLIENT);
-			if ((ret = __txn_reset(dbenv)) != 0)
-				return (ret);
+				goto errunlock;
 		}
 
 		redo_prepared = 0;
 		if (!F_ISSET(rep, REP_F_MASTER)) {
 			/* Master is not yet set. */
-			if (F_ISSET(rep, REP_ISCLIENT)) {
-				F_CLR(rep, REP_ISCLIENT);
-				rep->gen = ++rep->w_gen;
+			if (was_client) {
+				if (rep->w_gen > rep->recover_gen)
+					rep->gen = ++rep->w_gen;
+				else if (rep->gen > rep->recover_gen)
+					rep->gen++;
+				else
+					rep->gen = rep->recover_gen + 1;
+				/*
+				 * There could have been any number of failed
+				 * elections, so jump the gen if we need to now.
+				 */
+				if (rep->egen > rep->gen)
+					rep->gen = rep->egen;
 				redo_prepared = 1;
 			} else if (rep->gen == 0)
-				rep->gen = 1;
+				rep->gen = rep->recover_gen + 1;
+			if (F_ISSET(rep, REP_F_MASTERELECT)) {
+				__rep_elect_done(dbenv, rep);
+				F_CLR(rep, REP_F_MASTERELECT);
+			}
+			if (rep->egen <= rep->gen)
+				rep->egen = rep->gen + 1;
+#ifdef DIAGNOSTIC
+			if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION))
+				__db_err(dbenv, "New master gen %lu, egen %lu",
+				    (u_long)rep->gen, (u_long)rep->egen);
+#endif
 		}
-
-		F_SET(rep, REP_F_MASTER);
-		F_SET(dbenv, DB_ENV_REP_MASTER);
-		MUTEX_UNLOCK(dbenv, db_rep->mutexp);
+		rep->master_id = rep->eid;
+		/*
+		 * Note, setting flags below implicitly clears out
+		 * REP_F_NOARCHIVE, REP_F_INIT and REP_F_READY.
+		 */
+		rep->flags = REP_F_MASTER;
+		MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
 		dblp = (DB_LOG *)dbenv->lg_handle;
 		R_LOCK(dbenv, &dblp->reginfo);
 		lsn = ((LOG *)dblp->reginfo.primary)->lsn;
 		R_UNLOCK(dbenv, &dblp->reginfo);
 
 		/*
-		 * Send the NEWMASTER message, then restore prepared txns
-		 * if and only if we just upgraded from being a client.
+		 * Send the NEWMASTER message first so that clients know
+		 * subsequent messages are coming from the right master.
+		 * We need to perform all actions below no master what
+		 * regarding errors.
 		 */
-		if ((ret = __rep_send_message(dbenv,
-		    DB_EID_BROADCAST, REP_NEWMASTER, &lsn, NULL, 0)) == 0 &&
-		    redo_prepared)
-			ret = __rep_restore_prepared(dbenv);
+		(void)__rep_send_message(dbenv,
+		    DB_EID_BROADCAST, REP_NEWMASTER, &lsn, NULL, 0);
+		ret = 0;
+		if (was_client)
+			ret = __txn_reset(dbenv);
+		/*
+		 * Take a transaction checkpoint so that our new generation
+		 * number get written to the log.
+		 */
+		if ((t_ret = __txn_checkpoint(dbenv, 0, 0, DB_FORCE)) != 0 &&
+		    ret == 0)
+			ret = t_ret;
+		if (redo_prepared &&
+		    (t_ret = __rep_restore_prepared(dbenv)) != 0 && ret == 0)
+			ret = t_ret;
 	} else {
-		F_CLR(dbenv, DB_ENV_REP_MASTER);
-		F_SET(dbenv, DB_ENV_REP_CLIENT);
-		if (LF_ISSET(DB_REP_LOGSONLY))
-			F_SET(dbenv, DB_ENV_REP_LOGSONLY);
-
-		announce = !F_ISSET(rep, REP_ISCLIENT) ||
-		    rep->master_id == DB_EID_INVALID;
 		init_db = 0;
-		if (!F_ISSET(rep, REP_ISCLIENT)) {
-			F_CLR(rep, REP_F_MASTER);
-			if (LF_ISSET(DB_REP_LOGSONLY))
-				F_SET(rep, REP_F_LOGSONLY);
-			else
-				F_SET(rep, REP_F_UPGRADE);
+		was_client = F_ISSET(rep, REP_ISCLIENT);
+		announce = !was_client || rep->master_id == DB_EID_INVALID;
 
-			/*
-			 * We initialize the client's generation number to 0.
-			 * Upon startup, it looks for a master and updates the
-			 * generation number as necessary, exactly as it does
-			 * during normal operation and a master failure.
-			 */
-			rep->gen = 0;
+		/* Zero out everything except recovery and tally flags. */
+		repflags = F_ISSET(rep, REP_F_NOARCHIVE |
+		    REP_F_READY | REP_F_RECOVER | REP_F_TALLY);
+		if (LF_ISSET(DB_REP_LOGSONLY))
+			FLD_SET(repflags, REP_F_LOGSONLY);
+		else
+			FLD_SET(repflags, REP_F_UPGRADE);
+
+		rep->flags = repflags;
+		if (!was_client) {
 			rep->master_id = DB_EID_INVALID;
 			init_db = 1;
 		}
-		MUTEX_UNLOCK(dbenv, db_rep->mutexp);
+		MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
 
 		/*
 		 * Abort any prepared transactions that were restored
@@ -256,10 +337,10 @@ __rep_start(dbenv, dbt, flags)
 		 * records come in.  Aborts will simply be ignored.
 		 */
 		if ((ret = __rep_abort_prepared(dbenv)) != 0)
-			return (ret);
+			goto errlock;
 
 		if ((ret = __rep_client_dbinit(dbenv, init_db)) != 0)
-			return (ret);
+			goto errlock;
 
 		/*
 		 * If this client created a newly replicated environment,
@@ -270,9 +351,22 @@ __rep_start(dbenv, dbt, flags)
 		 * simply join in.
 		 */
 		if (announce)
-			ret = __rep_send_message(dbenv,
+			(void)__rep_send_message(dbenv,
 			    DB_EID_BROADCAST, REP_NEWCLIENT, NULL, dbt, 0);
 	}
+
+	/*
+	 * We have separate labels for errors.  If we're returning an
+	 * error before we've set start_th, we use 'err'.  If
+	 * we are erroring while holding the rep_mutex, then we use
+	 * 'errunlock' label.  If we're erroring without holding the rep
+	 * mutex we must use 'errlock'.
+	 */
+errlock:
+	MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
+errunlock:
+	rep->start_th = 0;
+err:	MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
 	return (ret);
 }
 
@@ -308,34 +402,36 @@ __rep_client_dbinit(dbenv, startup)
 	MUTEX_LOCK(dbenv, db_rep->db_mutexp);
 
 	if (startup) {
-		if ((ret = db_create(&dbp, dbenv, 0)) != 0)
+		if ((ret = db_create(&dbp, dbenv, DB_REP_CREATE)) != 0)
 			goto err;
 		/*
 		 * Ignore errors, because if the file doesn't exist, this
 		 * is perfectly OK.
 		 */
-		(void)dbp->remove(dbp, REPDBNAME, NULL, 0);
+		(void)__db_remove(dbp, NULL, REPDBNAME, NULL, DB_FORCE);
 	}
 
-	if ((ret = db_create(&dbp, dbenv, 0)) != 0)
+	if ((ret = db_create(&dbp, dbenv, DB_REP_CREATE)) != 0)
 		goto err;
-	if ((ret = dbp->set_bt_compare(dbp, __rep_bt_cmp)) != 0)
+	if ((ret = __bam_set_bt_compare(dbp, __rep_bt_cmp)) != 0)
 		goto err;
 
 	/* Allow writes to this database on a client. */
 	F_SET(dbp, DB_AM_CL_WRITER);
 
-	flags = (F_ISSET(dbenv, DB_ENV_THREAD) ? DB_THREAD : 0) |
-	    (startup ? DB_CREATE : 0);
-	if ((ret = dbp->open(dbp, NULL,
-	    "__db.rep.db", NULL, DB_BTREE, flags, 0)) != 0)
+	flags = DB_NO_AUTO_COMMIT |
+	    (startup ? DB_CREATE : 0) |
+	    (F_ISSET(dbenv, DB_ENV_THREAD) ? DB_THREAD : 0);
+
+	if ((ret = __db_open(dbp, NULL,
+	    REPDBNAME, NULL, DB_BTREE, flags, 0, PGNO_BASE_MD)) != 0)
 		goto err;
 
 	db_rep->rep_db = dbp;
 
 	if (0) {
 err:		if (dbp != NULL &&
-		    (t_ret = dbp->close(dbp, DB_NOSYNC)) != 0 && ret == 0)
+		    (t_ret = __db_close(dbp, NULL, DB_NOSYNC)) != 0 && ret == 0)
 			ret = t_ret;
 		db_rep->rep_db = NULL;
 	}
@@ -416,12 +512,12 @@ __rep_abort_prepared(dbenv)
 	if (do_aborts) {
 		op = DB_FIRST;
 		do {
-			if ((ret = dbenv->txn_recover(dbenv,
+			if ((ret = __txn_recover(dbenv,
 			    prep, PREPLISTSIZE, &count, op)) != 0)
 				return (ret);
 			for (i = 0; i < count; i++) {
 				p = &prep[i];
-				if ((ret = p->txn->abort(p->txn)) != 0)
+				if ((ret = __txn_abort(p->txn)) != 0)
 					return (ret);
 			}
 			op = DB_NEXT;
@@ -462,7 +558,7 @@ __rep_restore_prepared(dbenv)
 	ZERO_LSN(ckp_lsn);
 	ZERO_LSN(lsn);
 
-	if ((ret = dbenv->log_cursor(dbenv, &logc, 0)) != 0)
+	if ((ret = __log_cursor(dbenv, &logc)) != 0)
 		return (ret);
 
 	/*
@@ -478,7 +574,7 @@ __rep_restore_prepared(dbenv)
 	 */
 	memset(&rec, 0, sizeof(DBT));
 	if ((ret = __txn_getckp(dbenv, &lsn)) == 0) {
-		if ((ret = logc->get(logc, &lsn, &rec, DB_SET)) != 0)  {
+		if ((ret = __log_c_get(logc, &lsn, &rec, DB_SET)) != 0)  {
 			__db_err(dbenv,
 			    "Checkpoint record at LSN [%lu][%lu] not found",
 			    (u_long)lsn.file, (u_long)lsn.offset);
@@ -495,13 +591,13 @@ __rep_restore_prepared(dbenv)
 		ckp_lsn = ckp_args->ckp_lsn;
 		__os_free(dbenv, ckp_args);
 
-		if ((ret = logc->get(logc, &ckp_lsn, &rec, DB_SET)) != 0) {
+		if ((ret = __log_c_get(logc, &ckp_lsn, &rec, DB_SET)) != 0) {
 			__db_err(dbenv,
 			    "Checkpoint LSN record [%lu][%lu] not found",
 			    (u_long)ckp_lsn.file, (u_long)ckp_lsn.offset);
 			goto err;
 		}
-	} else if ((ret = logc->get(logc, &lsn, &rec, DB_FIRST)) != 0) {
+	} else if ((ret = __log_c_get(logc, &lsn, &rec, DB_FIRST)) != 0) {
 		if (ret == DB_NOTFOUND) {
 			/* An empty log means no PBNYC txns. */
 			ret = 0;
@@ -524,7 +620,7 @@ __rep_restore_prepared(dbenv)
 		    (u_int8_t *)rec.data + sizeof(u_int32_t), sizeof(low_txn));
 		if (low_txn != 0)
 			break;
-	} while ((ret = logc->get(logc, &lsn, &rec, DB_NEXT)) == 0);
+	} while ((ret = __log_c_get(logc, &lsn, &rec, DB_NEXT)) == 0);
 
 	/* If there are no txns, there are no PBNYC txns. */
 	if (ret == DB_NOTFOUND) {
@@ -534,7 +630,7 @@ __rep_restore_prepared(dbenv)
 		goto err;
 
 	/* Now, the high txnid. */
-	if ((ret = logc->get(logc, &lsn, &rec, DB_LAST)) != 0) {
+	if ((ret = __log_c_get(logc, &lsn, &rec, DB_LAST)) != 0) {
 		/*
 		 * Note that DB_NOTFOUND is unacceptable here because we
 		 * had to have looked at some log record to get this far.
@@ -548,7 +644,7 @@ __rep_restore_prepared(dbenv)
 		    (u_int8_t *)rec.data + sizeof(u_int32_t), sizeof(hi_txn));
 		if (hi_txn != 0)
 			break;
-	} while ((ret = logc->get(logc, &lsn, &rec, DB_PREV)) == 0);
+	} while ((ret = __log_c_get(logc, &lsn, &rec, DB_PREV)) == 0);
 	if (ret == DB_NOTFOUND) {
 		ret = 0;
 		goto done;
@@ -570,9 +666,9 @@ __rep_restore_prepared(dbenv)
 	 * Since all PBNYC txns still held locks on the old master and
 	 * were isolated, this should be safe.
 	 */
-	for (ret = logc->get(logc, &lsn, &rec, DB_LAST);
+	for (ret = __log_c_get(logc, &lsn, &rec, DB_LAST);
 	    ret == 0 && log_compare(&lsn, &ckp_lsn) > 0;
-	    ret = logc->get(logc, &lsn, &rec, DB_PREV)) {
+	    ret = __log_c_get(logc, &lsn, &rec, DB_PREV)) {
 		memcpy(&rectype, rec.data, sizeof(rectype));
 		switch (rectype) {
 		case DB___txn_regop:
@@ -587,7 +683,7 @@ __rep_restore_prepared(dbenv)
 
 			ret = __db_txnlist_find(dbenv,
 			    txninfo, regop_args->txnid->txnid);
-			if (ret == DB_NOTFOUND)
+			if (ret == TXN_NOTFOUND)
 				ret = __db_txnlist_add(dbenv, txninfo,
 				    regop_args->txnid->txnid,
 				    regop_args->opcode, &lsn);
@@ -595,19 +691,25 @@ __rep_restore_prepared(dbenv)
 			break;
 		case DB___txn_xa_regop:
 			/*
-			 * It's a prepare.  If we haven't put the
-			 * txn on our list yet, it hasn't been
-			 * resolved, so apply and restore it.
+			 * It's a prepare.  If its not aborted and
+			 * we haven't put the txn on our list yet, it
+			 * hasn't been resolved, so apply and restore it.
 			 */
 			if ((ret = __txn_xa_regop_read(dbenv, rec.data,
 			    &prep_args)) != 0)
 				goto err;
 			ret = __db_txnlist_find(dbenv, txninfo,
 			    prep_args->txnid->txnid);
-			if (ret == DB_NOTFOUND)
-				if ((ret = __rep_process_txn(dbenv, &rec)) == 0)
+			if (ret == TXN_NOTFOUND) {
+				if (prep_args->opcode == TXN_ABORT)
+					ret = __db_txnlist_add(dbenv, txninfo,
+					    prep_args->txnid->txnid,
+					    prep_args->opcode, &lsn);
+				else if ((ret =
+				    __rep_process_txn(dbenv, &rec)) == 0)
 					ret = __txn_restore_txn(dbenv,
 					    &lsn, prep_args);
+			}
 			__os_free(dbenv, prep_args);
 			break;
 		default:
@@ -620,7 +722,7 @@ __rep_restore_prepared(dbenv)
 		ret = 0;
 
 done:
-err:	t_ret = logc->close(logc, 0);
+err:	t_ret = __log_c_close(logc);
 
 	if (txninfo != NULL)
 		__db_txnlist_end(dbenv, txninfo);
@@ -628,6 +730,34 @@ err:	t_ret = logc->close(logc, 0);
 	return (ret == 0 ? t_ret : ret);
 }
 
+static int
+__rep_get_limit(dbenv, gbytesp, bytesp)
+	DB_ENV *dbenv;
+	u_int32_t *gbytesp, *bytesp;
+{
+	DB_REP *db_rep;
+	REP *rep;
+
+	PANIC_CHECK(dbenv);
+	ENV_REQUIRES_CONFIG(dbenv, dbenv->rep_handle, "rep_get_limit",
+	    DB_INIT_REP);
+
+	if (!REP_ON(dbenv)) {
+		__db_err(dbenv,
+    "DB_ENV->get_rep_limit: database environment not properly initialized");
+		return (__db_panic(dbenv, EINVAL));
+	}
+	db_rep = dbenv->rep_handle;
+	rep = db_rep->region;
+
+	if (gbytesp != NULL)
+		*gbytesp = rep->gbytes;
+	if (bytesp != NULL)
+		*bytesp = rep->bytes;
+
+	return (0);
+}
+
 /*
  * __rep_set_limit --
  *	Set a limit on the amount of data that will be sent during a single
@@ -636,28 +766,31 @@ err:	t_ret = logc->close(logc, 0);
 static int
 __rep_set_limit(dbenv, gbytes, bytes)
 	DB_ENV *dbenv;
-	u_int32_t gbytes;
-	u_int32_t bytes;
+	u_int32_t gbytes, bytes;
 {
 	DB_REP *db_rep;
 	REP *rep;
 
 	PANIC_CHECK(dbenv);
+	ENV_ILLEGAL_BEFORE_OPEN(dbenv, "DB_ENV->rep_set_limit");
+	ENV_REQUIRES_CONFIG(dbenv, dbenv->rep_handle, "rep_set_limit",
+	    DB_INIT_REP);
 
-	if ((db_rep = dbenv->rep_handle) == NULL) {
+	if (!REP_ON(dbenv)) {
 		__db_err(dbenv,
     "DB_ENV->set_rep_limit: database environment not properly initialized");
 		return (__db_panic(dbenv, EINVAL));
 	}
+	db_rep = dbenv->rep_handle;
 	rep = db_rep->region;
-	MUTEX_LOCK(dbenv, db_rep->mutexp);
+	MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
 	if (bytes > GIGABYTE) {
 		gbytes += bytes / GIGABYTE;
 		bytes = bytes % GIGABYTE;
 	}
 	rep->gbytes = gbytes;
 	rep->bytes = bytes;
-	MUTEX_UNLOCK(dbenv, db_rep->mutexp);
+	MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
 
 	return (0);
 }
@@ -671,8 +804,7 @@ __rep_set_limit(dbenv, gbytes, bytes)
 static int
 __rep_set_request(dbenv, min, max)
 	DB_ENV *dbenv;
-	u_int32_t min;
-	u_int32_t max;
+	u_int32_t min, max;
 {
 	LOG *lp;
 	DB_LOG *dblp;
@@ -680,24 +812,32 @@ __rep_set_request(dbenv, min, max)
 	REP *rep;
 
 	PANIC_CHECK(dbenv);
+	ENV_ILLEGAL_BEFORE_OPEN(dbenv, "DB_ENV->rep_set_request");
+	ENV_REQUIRES_CONFIG(dbenv, dbenv->rep_handle, "rep_set_request",
+	    DB_INIT_REP);
 
-	if ((db_rep = dbenv->rep_handle) == NULL) {
+	if (!REP_ON(dbenv)) {
 		__db_err(dbenv,
     "DB_ENV->set_rep_request: database environment not properly initialized");
 		return (__db_panic(dbenv, EINVAL));
 	}
+	db_rep = dbenv->rep_handle;
 	rep = db_rep->region;
-	MUTEX_LOCK(dbenv, db_rep->mutexp);
+	/*
+	 * Note we acquire the rep_mutexp or the db_mutexp as needed.
+	 */
+	MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
 	rep->request_gap = min;
 	rep->max_gap = max;
-	MUTEX_UNLOCK(dbenv, db_rep->mutexp);
+	MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
+
+	MUTEX_LOCK(dbenv, db_rep->db_mutexp);
 	dblp = dbenv->lg_handle;
 	if (dblp != NULL && (lp = dblp->reginfo.primary) != NULL) {
-		R_LOCK(dbenv, &dblp->reginfo);
 		lp->wait_recs = 0;
 		lp->rcvd_recs = 0;
-		R_UNLOCK(dbenv, &dblp->reginfo);
 	}
+	MUTEX_UNLOCK(dbenv, db_rep->db_mutexp);
 
 	return (0);
 }
@@ -710,32 +850,22 @@ static int
 __rep_set_rep_transport(dbenv, eid, f_send)
 	DB_ENV *dbenv;
 	int eid;
-	int (*f_send) __P((DB_ENV *, const DBT *, const DBT *, int, u_int32_t));
+	int (*f_send) __P((DB_ENV *, const DBT *, const DBT *, const DB_LSN *,
+	    int, u_int32_t));
 {
-	DB_REP *db_rep;
-
 	PANIC_CHECK(dbenv);
 
-	if ((db_rep = dbenv->rep_handle) == NULL) {
-		__db_err(dbenv,
-    "DB_ENV->set_rep_transport: database environment not properly initialized");
-		return (__db_panic(dbenv, EINVAL));
-	}
-
 	if (f_send == NULL) {
 		__db_err(dbenv,
 	"DB_ENV->set_rep_transport: no send function specified");
 		return (EINVAL);
 	}
-
 	if (eid < 0) {
 		__db_err(dbenv,
 	"DB_ENV->set_rep_transport: eid must be greater than or equal to 0");
 		return (EINVAL);
 	}
-
-	db_rep->rep_send = f_send;
-
+	dbenv->rep_send = f_send;
 	dbenv->rep_eid = eid;
 	return (0);
 }
@@ -756,11 +886,11 @@ __rep_elect(dbenv, nsites, priority, timeout, eidp)
 	DB_LSN lsn;
 	DB_REP *db_rep;
 	REP *rep;
-	int in_progress, ret, send_vote, tiebreaker;
-	u_int32_t pid, sec, usec;
+	int done, in_progress, ret, send_vote, tiebreaker;
+	u_int32_t egen, orig_tally, pid, sec, usec;
 
 	PANIC_CHECK(dbenv);
-	ENV_REQUIRES_CONFIG(dbenv, dbenv->tx_handle, "rep_elect", DB_INIT_TXN);
+	ENV_REQUIRES_CONFIG(dbenv, dbenv->rep_handle, "rep_elect", DB_INIT_REP);
 
 	/* Error checking. */
 	if (nsites <= 0) {
@@ -782,45 +912,95 @@ __rep_elect(dbenv, nsites, priority, timeout, eidp)
 	lsn = ((LOG *)dblp->reginfo.primary)->lsn;
 	R_UNLOCK(dbenv, &dblp->reginfo);
 
+	orig_tally = 0;
+	if ((ret = __rep_elect_init(dbenv,
+	    &lsn, nsites, priority, &in_progress, &orig_tally)) != 0) {
+		if (ret == DB_REP_NEWMASTER) {
+			ret = 0;
+			*eidp = dbenv->rep_eid;
+		}
+		goto err;
+	}
+	/*
+	 * If another thread is in the middle of an election we
+	 * just quietly return and not interfere.
+	 */
+	if (in_progress) {
+		*eidp = dbenv->rep_eid;
+		return (0);
+	}
+	(void)__rep_send_message(dbenv,
+	    DB_EID_BROADCAST, REP_MASTER_REQ, NULL, NULL, 0);
+	ret = __rep_wait(dbenv, timeout/4, eidp, REP_F_EPHASE1);
+	switch (ret) {
+		case 0:
+			/* Check if we found a master. */
+			if (*eidp != DB_EID_INVALID) {
+#ifdef DIAGNOSTIC
+				if (FLD_ISSET(dbenv->verbose,
+				    DB_VERB_REPLICATION))
+					__db_err(dbenv,
+					    "Found master %d", *eidp);
+#endif
+				return (0);
+			}
+			/*
+			 * If we didn't find a master, continue
+			 * the election.
+			 */
+			break;
+		case DB_TIMEOUT:
+#ifdef DIAGNOSTIC
+		if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION))
+			__db_err(dbenv,
+			    "Did not find master.  Sending vote1");
+#endif
+			break;
+		default:
+			goto err;
+	}
 	/* Generate a randomized tiebreaker value. */
+restart:
 	__os_id(&pid);
 	if ((ret = __os_clock(dbenv, &sec, &usec)) != 0)
 		return (ret);
 	tiebreaker = pid ^ sec ^ usec ^ (u_int)rand() ^ P_TO_UINT32(&pid);
 
-	if ((ret = __rep_elect_init(dbenv,
-	    &lsn, nsites, priority, tiebreaker, &in_progress)) != 0) {
-		if (ret == DB_REP_NEWMASTER) {
-			ret = 0;
-			*eidp = dbenv->rep_eid;
-		}
-		return (ret);
-	}
+	MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
+	F_SET(rep, REP_F_EPHASE1 | REP_F_NOARCHIVE);
+	F_CLR(rep, REP_F_TALLY);
+
+	/* Tally our own vote */
+	if (__rep_tally(dbenv, rep, rep->eid, &rep->sites, rep->egen,
+	    rep->tally_off) != 0)
+		goto lockdone;
+	__rep_cmp_vote(dbenv, rep, &rep->eid, &lsn, priority, rep->gen,
+	    tiebreaker);
 
-	if (!in_progress) {
 #ifdef DIAGNOSTIC
-		if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION))
-			__db_err(dbenv, "Beginning an election");
+	if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION))
+		__db_err(dbenv, "Beginning an election");
 #endif
-		if ((ret = __rep_send_message(dbenv,
-		    DB_EID_BROADCAST, REP_ELECT, NULL, NULL, 0)) != 0)
-			goto err;
-		DB_ENV_TEST_RECOVERY(dbenv, DB_TEST_ELECTSEND, ret, NULL);
-	}
 
 	/* Now send vote */
-	if ((ret =
-	    __rep_send_vote(dbenv, &lsn, nsites, priority, tiebreaker)) != 0)
-		goto err;
-	DB_ENV_TEST_RECOVERY(dbenv, DB_TEST_ELECTVOTE1, ret, NULL);
-
+	send_vote = DB_EID_INVALID;
+	egen = rep->egen;
+	MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
+	__rep_send_vote(dbenv, &lsn, nsites, priority, tiebreaker, egen,
+	    DB_EID_BROADCAST, REP_VOTE1);
 	ret = __rep_wait(dbenv, timeout, eidp, REP_F_EPHASE1);
-	DB_ENV_TEST_RECOVERY(dbenv, DB_TEST_ELECTWAIT1, ret, NULL);
 	switch (ret) {
 		case 0:
 			/* Check if election complete or phase complete. */
-			if (*eidp != DB_EID_INVALID)
+			if (*eidp != DB_EID_INVALID) {
+#ifdef DIAGNOSTIC
+				if (FLD_ISSET(dbenv->verbose,
+				    DB_VERB_REPLICATION))
+					__db_err(dbenv,
+					    "Ended election phase 1 %d", ret);
+#endif
 				return (0);
+			}
 			goto phase2;
 		case DB_TIMEOUT:
 			break;
@@ -833,17 +1013,42 @@ __rep_elect(dbenv, nsites, priority, timeout, eidp)
 	 * votes to pick a winner and if so, to send out a vote to
 	 * the winner.
 	 */
-	MUTEX_LOCK(dbenv, db_rep->mutexp);
-	send_vote = DB_EID_INVALID;
+	MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
+	/*
+	 * If our egen changed while we were waiting.  We need to
+	 * essentially reinitialize our election.
+	 */
+	if (egen != rep->egen) {
+		MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
+#ifdef DIAGNOSTIC
+		if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION))
+			__db_err(dbenv, "Egen changed from %lu to %lu",
+			    (u_long)egen, (u_long)rep->egen);
+#endif
+		goto restart;
+	}
 	if (rep->sites > rep->nsites / 2) {
+
 		/* We think we've seen enough to cast a vote. */
 		send_vote = rep->winner;
-		if (rep->winner == rep->eid)
-			rep->votes++;
-		F_CLR(rep, REP_F_EPHASE1);
+		/*
+		 * See if we won.  This will make sure we
+		 * don't count ourselves twice if we're racing
+		 * with incoming votes.
+		 */
+		if (rep->winner == rep->eid) {
+			(void)__rep_tally(dbenv, rep, rep->eid, &rep->votes,
+			    egen, rep->v2tally_off);
+#ifdef DIAGNOSTIC
+			if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION))
+				__db_err(dbenv,
+				    "Counted my vote %d", rep->votes);
+#endif
+		}
 		F_SET(rep, REP_F_EPHASE2);
+		F_CLR(rep, REP_F_EPHASE1);
 	}
-	MUTEX_UNLOCK(dbenv, db_rep->mutexp);
+	MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
 	if (send_vote == DB_EID_INVALID) {
 		/* We do not have enough votes to elect. */
 #ifdef DIAGNOSTIC
@@ -855,39 +1060,74 @@ __rep_elect(dbenv, nsites, priority, timeout, eidp)
 		ret = DB_REP_UNAVAIL;
 		goto err;
 
-	}
+	} else {
+		/*
+		 * We have seen enough vote1's.  Now we need to wait
+		 * for all the vote2's.
+		 */
+		if (send_vote != rep->eid) {
 #ifdef DIAGNOSTIC
-	if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION) &&
-	    send_vote != rep->eid)
-		__db_err(dbenv, "Sending vote");
+			if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION) &&
+			    send_vote != rep->eid)
+				__db_err(dbenv, "Sending vote");
 #endif
+			__rep_send_vote(dbenv, NULL, 0, 0, 0, egen,
+			    send_vote, REP_VOTE2);
 
-	if (send_vote != rep->eid && (ret = __rep_send_message(dbenv,
-	    send_vote, REP_VOTE2, NULL, NULL, 0)) != 0)
-		goto err;
-	DB_ENV_TEST_RECOVERY(dbenv, DB_TEST_ELECTVOTE2, ret, NULL);
-
-phase2:	ret = __rep_wait(dbenv, timeout, eidp, REP_F_EPHASE2);
-	DB_ENV_TEST_RECOVERY(dbenv, DB_TEST_ELECTWAIT2, ret, NULL);
-	switch (ret) {
-		case 0:
-			return (0);
-		case DB_TIMEOUT:
-			ret = DB_REP_UNAVAIL;
-			break;
-		default:
-			goto err;
+		}
+phase2:
+		ret = __rep_wait(dbenv, timeout, eidp, REP_F_EPHASE2);
+#ifdef DIAGNOSTIC
+				if (FLD_ISSET(dbenv->verbose,
+				    DB_VERB_REPLICATION))
+					__db_err(dbenv,
+					    "Ended election phase 2 %d", ret);
+#endif
+		switch (ret) {
+			case 0:
+				return (0);
+			case DB_TIMEOUT:
+				ret = DB_REP_UNAVAIL;
+				break;
+			default:
+				goto err;
+		}
+		MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
+		done = rep->votes > rep->nsites / 2;
+#ifdef DIAGNOSTIC
+		if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION))
+			__db_err(dbenv,
+			    "After phase 2: done %d, votes %d, nsites %d",
+			    done, rep->votes, rep->nsites);
+#endif
+		if (send_vote == rep->eid && done) {
+			__rep_elect_master(dbenv, rep, eidp);
+			ret = 0;
+			goto lockdone;
+		}
+		MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
 	}
 
-DB_TEST_RECOVERY_LABEL
-err:	MUTEX_LOCK(dbenv, db_rep->mutexp);
-	ELECTION_DONE(rep);
-	MUTEX_UNLOCK(dbenv, db_rep->mutexp);
+err:	MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
+lockdone:
+	/*
+	 * If we get here because of a non-election error, then we
+	 * did not tally our vote.  The only non-election error is
+	 * from elect_init where we were unable to grow_sites.  In
+	 * that case we do not want to discard all known election info.
+	 */
+	if (ret == 0 || ret == DB_REP_UNAVAIL)
+		__rep_elect_done(dbenv, rep);
+	else if (orig_tally)
+		F_SET(rep, orig_tally);
 
 #ifdef DIAGNOSTIC
 	if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION))
-		__db_err(dbenv, "Ended election with %d", ret);
+		__db_err(dbenv,
+		    "Ended election with %d, sites %d, egen %lu, flags 0x%lx",
+		    ret, rep->sites, (u_long)rep->egen, (u_long)rep->flags);
 #endif
+	MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
 	return (ret);
 }
 
@@ -897,14 +1137,16 @@ err:	MUTEX_LOCK(dbenv, db_rep->mutexp);
  * already in progress; makes it 0 otherwise.
  */
 static int
-__rep_elect_init(dbenv, lsnp, nsites, priority, tiebreaker, beginp)
+__rep_elect_init(dbenv, lsnp, nsites, priority, beginp, otally)
 	DB_ENV *dbenv;
 	DB_LSN *lsnp;
-	int nsites, priority, tiebreaker, *beginp;
+	int nsites, priority;
+	int *beginp;
+	u_int32_t *otally;
 {
 	DB_REP *db_rep;
 	REP *rep;
-	int ret, *tally;
+	int ret;
 
 	db_rep = dbenv->rep_handle;
 	rep = db_rep->region;
@@ -915,14 +1157,16 @@ __rep_elect_init(dbenv, lsnp, nsites, priority, tiebreaker, beginp)
 	rep->stat.st_elections++;
 
 	/* If we are already a master; simply broadcast that fact and return. */
-	if (F_ISSET(dbenv, DB_ENV_REP_MASTER)) {
+	if (F_ISSET(rep, REP_F_MASTER)) {
 		(void)__rep_send_message(dbenv,
 		    DB_EID_BROADCAST, REP_NEWMASTER, lsnp, NULL, 0);
 		rep->stat.st_elections_won++;
 		return (DB_REP_NEWMASTER);
 	}
 
-	MUTEX_LOCK(dbenv, db_rep->mutexp);
+	MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
+	if (otally != NULL)
+		*otally = F_ISSET(rep, REP_F_TALLY);
 	*beginp = IN_ELECTION(rep);
 	if (!*beginp) {
 		/*
@@ -937,35 +1181,41 @@ __rep_elect_init(dbenv, lsnp, nsites, priority, tiebreaker, beginp)
 		DB_ENV_TEST_RECOVERY(dbenv, DB_TEST_ELECTINIT, ret, NULL);
 		rep->nsites = nsites;
 		rep->priority = priority;
-		rep->votes = 0;
 		rep->master_id = DB_EID_INVALID;
-		F_SET(rep, REP_F_EPHASE1);
-
-		/* We have always heard from ourselves. */
-		rep->sites = 1;
-		tally = R_ADDR((REGINFO *)dbenv->reginfo, rep->tally_off);
-		tally[0] = rep->eid;
-
-		if (priority != 0) {
-			/* Make ourselves the winner to start. */
-			rep->winner = rep->eid;
-			rep->w_priority = priority;
-			rep->w_gen = rep->gen;
-			rep->w_lsn = *lsnp;
-			rep->w_tiebreaker = tiebreaker;
-		} else {
-			rep->winner = DB_EID_INVALID;
-			rep->w_priority = 0;
-			rep->w_gen = 0;
-			ZERO_LSN(rep->w_lsn);
-			rep->w_tiebreaker = 0;
-		}
 	}
 DB_TEST_RECOVERY_LABEL
-err:	MUTEX_UNLOCK(dbenv, db_rep->mutexp);
+err:	MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
 	return (ret);
 }
 
+/*
+ * __rep_elect_master
+ *	Set up for new master from election.  Must be called with
+ *	the db_rep->rep_mutex held.
+ *
+ * PUBLIC: void __rep_elect_master __P((DB_ENV *, REP *, int *));
+ */
+void
+__rep_elect_master(dbenv, rep, eidp)
+	DB_ENV *dbenv;
+	REP *rep;
+	int *eidp;
+{
+	rep->master_id = rep->eid;
+	F_SET(rep, REP_F_MASTERELECT);
+	if (eidp != NULL)
+		*eidp = rep->master_id;
+	rep->stat.st_elections_won++;
+#ifdef DIAGNOSTIC
+	if (FLD_ISSET(dbenv->verbose, DB_VERB_REPLICATION))
+		__db_err(dbenv,
+    "Got enough votes to win; election done; winner is %d, gen %lu",
+		    rep->master_id, (u_long)rep->gen);
+#else
+	COMPQUIET(dbenv, NULL);
+#endif
+}
+
 static int
 __rep_wait(dbenv, timeout, eidp, flags)
 	DB_ENV *dbenv;
@@ -975,7 +1225,7 @@ __rep_wait(dbenv, timeout, eidp, flags)
 {
 	DB_REP *db_rep;
 	REP *rep;
-	int done, ret;
+	int done;
 	u_int32_t sleeptime;
 
 	done = 0;
@@ -991,13 +1241,12 @@ __rep_wait(dbenv, timeout, eidp, flags)
 	if (sleeptime == 0)
 		sleeptime++;
 	while (timeout > 0) {
-		if ((ret = __os_sleep(dbenv, 0, sleeptime)) != 0)
-			return (ret);
-		MUTEX_LOCK(dbenv, db_rep->mutexp);
+		(void)__os_sleep(dbenv, 0, sleeptime);
+		MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
 		done = !F_ISSET(rep, flags) && rep->master_id != DB_EID_INVALID;
 
 		*eidp = rep->master_id;
-		MUTEX_UNLOCK(dbenv, db_rep->mutexp);
+		MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
 
 		if (done)
 			return (0);
@@ -1025,21 +1274,21 @@ __rep_flush(dbenv)
 	int ret, t_ret;
 
 	PANIC_CHECK(dbenv);
-	ENV_REQUIRES_CONFIG(dbenv, dbenv->tx_handle, "rep_stat", DB_INIT_TXN);
+	ENV_REQUIRES_CONFIG(dbenv, dbenv->rep_handle, "rep_flush", DB_INIT_REP);
 
-	if ((ret = dbenv->log_cursor(dbenv, &logc, 0)) != 0)
+	if ((ret = __log_cursor(dbenv, &logc)) != 0)
 		return (ret);
 
 	memset(&rec, 0, sizeof(rec));
 	memset(&lsn, 0, sizeof(lsn));
 
-	if ((ret = logc->get(logc, &lsn, &rec, DB_LAST)) != 0)
+	if ((ret = __log_c_get(logc, &lsn, &rec, DB_LAST)) != 0)
 		goto err;
 
-	ret = __rep_send_message(dbenv,
+	(void)__rep_send_message(dbenv,
 	    DB_EID_BROADCAST, REP_LOG, &lsn, &rec, 0);
 
-err:	if ((t_ret = logc->close(logc, 0)) != 0 && ret == 0)
+err:	if ((t_ret = __log_c_close(logc)) != 0 && ret == 0)
 		ret = t_ret;
 	return (ret);
 }
@@ -1059,11 +1308,11 @@ __rep_stat(dbenv, statp, flags)
 	DB_REP_STAT *stats;
 	LOG *lp;
 	REP *rep;
-	u_int32_t queued;
-	int ret;
+	u_int32_t queued, repflags;
+	int dolock, ret;
 
 	PANIC_CHECK(dbenv);
-	ENV_REQUIRES_CONFIG(dbenv, dbenv->tx_handle, "rep_stat", DB_INIT_TXN);
+	ENV_REQUIRES_CONFIG(dbenv, dbenv->rep_handle, "rep_stat", DB_INIT_REP);
 
 	db_rep = dbenv->rep_handle;
 	rep = db_rep->region;
@@ -1079,11 +1328,23 @@ __rep_stat(dbenv, statp, flags)
 	if ((ret = __os_umalloc(dbenv, sizeof(DB_REP_STAT), &stats)) != 0)
 		return (ret);
 
-	MUTEX_LOCK(dbenv, db_rep->mutexp);
+	/*
+	 * Read without holding the lock.  If we are in client
+	 * recovery, we copy just the stats struct so we won't
+	 * block.  We only copy out those stats that don't
+	 * require acquiring any mutex.
+	 */
+	repflags = rep->flags;
+	if (FLD_ISSET(repflags, REP_F_RECOVER))
+		dolock = 0;
+	else {
+		dolock = 1;
+		MUTEX_LOCK(dbenv, db_rep->rep_mutexp);
+	}
 	memcpy(stats, &rep->stat, sizeof(*stats));
 
 	/* Copy out election stats. */
-	if (IN_ELECTION(rep)) {
+	if (IN_ELECTION_TALLY(rep)) {
 		if (F_ISSET(rep, REP_F_EPHASE1))
 			stats->st_election_status = 1;
 		else if (F_ISSET(rep, REP_F_EPHASE2))
@@ -1120,13 +1381,18 @@ __rep_stat(dbenv, statp, flags)
 		rep->stat.st_log_queued = rep->stat.st_log_queued_total =
 		    rep->stat.st_log_queued_max = queued;
 	}
-	MUTEX_UNLOCK(dbenv, db_rep->mutexp);
+
+	if (dolock) {
+		stats->st_in_recovery = 0;
+		MUTEX_UNLOCK(dbenv, db_rep->rep_mutexp);
+		MUTEX_LOCK(dbenv, db_rep->db_mutexp);
+	} else
+		stats->st_in_recovery = 1;
 
 	/*
 	 * Log-related replication info is stored in the log system and
 	 * protected by the log region lock.
 	 */
-	R_LOCK(dbenv, &dblp->reginfo);
 	if (F_ISSET(rep, REP_ISCLIENT)) {
 		stats->st_next_lsn = lp->ready_lsn;
 		stats->st_waiting_lsn = lp->waiting_lsn;
@@ -1137,7 +1403,8 @@ __rep_stat(dbenv, statp, flags)
 			ZERO_LSN(stats->st_next_lsn);
 		ZERO_LSN(stats->st_waiting_lsn);
 	}
-	R_UNLOCK(dbenv, &dblp->reginfo);
+	if (dolock)
+		MUTEX_UNLOCK(dbenv, db_rep->db_mutexp);
 
 	*statp = stats;
 	return (0);
