@@ -1,11 +1,11 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1999-2002
+# Copyright (c) 1999-2003
 #	Sleepycat Software.  All rights reserved.
 #
-# $Id: sdb005.tcl,v 11.18 2002/07/11 18:53:46 sandstro Exp $
+# $Id: sdb005.tcl,v 11.20 2003/01/08 05:52:59 bostic Exp $
 #
-# TEST	subdb005
+# TEST	sdb005
 # TEST	Tests cursor operations in subdbs
 # TEST		Put/get per key
 # TEST		Verify cursor operations work within subdb
@@ -13,7 +13,7 @@
 # TEST
 #
 # We should test this on all btrees, all hash, and a combination thereof
-proc subdb005 {method {nentries 100} args } {
+proc sdb005 {method {nentries 100} args } {
 	source ./include.tcl
 
 	set args [convert_args $method $args]
