@@ -1,27 +1,32 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 2000-2001
+# Copyright (c) 2000-2003
 #	Sleepycat Software.  All rights reserved.
 #
-# $Id: test084.tcl,v 11.7 2001/01/25 18:23:13 bostic Exp $
+# $Id: test084.tcl,v 11.14 2003/08/28 19:59:15 sandstro Exp $
 #
-# Test 84.
-# Basic sanity test (test001) with large (64K) pages.
-#
-proc test084 { method {nentries 10000} {tnum 84} {pagesize 65536} args} {
+# TEST	test084
+# TEST	Basic sanity test (test001) with large (64K) pages.
+proc test084 { method {nentries 10000} {tnum "084"} {pagesize 65536} args} {
 	source ./include.tcl
 
+	set txnenv 0
 	set eindex [lsearch -exact $args "-env"]
 	#
 	# If we are using an env, then testfile should just be the db name.
 	# Otherwise it is the test directory and the name.
 	if { $eindex == -1 } {
-		set testfile $testdir/test0$tnum-empty.db
+		set testfile $testdir/test$tnum-empty.db
 		set env NULL
 	} else {
-		set testfile test0$tnum-empty.db
+		set testfile test$tnum-empty.db
 		incr eindex
 		set env [lindex $args $eindex]
+		set txnenv [is_txnenv $env]
+		if { $txnenv == 1 } {
+			append args " -auto_commit "
+		}
+		set testdir [get_home $env]
 	}
 
 	set pgindex [lsearch -exact $args "-pagesize"]
@@ -34,7 +39,7 @@ proc test084 { method {nentries 10000} {tnum 84} {pagesize 65536} args} {
 
 	set args "-pagesize $pagesize $args"
 
-	eval {test001 $method $nentries 0 $tnum} $args
+	eval {test001 $method $nentries 0 0 $tnum} $args
 
 	set omethod [convert_method $method]
 	set args [convert_args $method $args]
