@@ -1,9 +1,9 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996-2001
+# Copyright (c) 1996-2003
 #	Sleepycat Software.  All rights reserved.
 #
-# $Id: testutils.tcl,v 11.116 2001/07/12 17:39:37 sue Exp $
+# $Id: testutils.tcl,v 11.182 2003/09/30 17:32:26 sandstro Exp $
 #
 # Test system utilities
 #
@@ -12,12 +12,19 @@
 proc timestamp {{opt ""}} {
 	global __timestamp_start
 
+	set now [clock seconds]
+
+	# -c	accurate to the click, instead of the second.
+	# -r	seconds since the Epoch
+	# -t	current time in the format expected by db_recover -t.
+	# -w	wallclock time
+	# else	wallclock plus elapsed time.
 	if {[string compare $opt "-r"] == 0} {
-		clock seconds
+		return $now
 	} elseif {[string compare $opt "-t"] == 0} {
-		# -t gives us the current time in the format expected by
-		# db_recover -t.
-		return [clock format [clock seconds] -format "%y%m%d%H%M.%S"]
+		return [clock format $now -format "%y%m%d%H%M.%S"]
+	} elseif {[string compare $opt "-w"] == 0} {
+		return [clock format $now -format "%c"]
 	} else {
 		if {[string compare $opt "-c"] == 0} {
 			set printclicks 1
@@ -25,8 +32,6 @@ proc timestamp {{opt ""}} {
 			set printclicks 0
 		}
 
-		set now [clock seconds]
-
 		if {[catch {set start $__timestamp_start}] != 0} {
 			set __timestamp_start $now
 		}
@@ -127,32 +132,68 @@ proc get_file_as_key { db txn flags file} {
 
 # open file and call dump_file to dumpkeys to tempfile
 proc open_and_dump_file {
-    dbname dbenv txn outfile checkfunc dump_func beg cont} {
+    dbname env outfile checkfunc dump_func beg cont } {
+	global encrypt
+	global passwd
 	source ./include.tcl
-	if { $dbenv == "NULL" } {
-		set db [berkdb open -rdonly -unknown $dbname]
-		error_check_good dbopen [is_valid_db $db] TRUE
-	} else {
-		set db [berkdb open -env $dbenv -rdonly -unknown $dbname]
-		error_check_good dbopen [is_valid_db $db] TRUE
+
+	set encarg ""
+	if { $encrypt > 0 && $env == "NULL" } {
+		set encarg "-encryptany $passwd"
 	}
+	set envarg ""
+	set txn ""
+	set txnenv 0
+	if { $env != "NULL" } {
+		append envarg " -env $env "
+		set txnenv [is_txnenv $env]
+		if { $txnenv == 1 } {
+			append envarg " -auto_commit "
+			set t [$env txn]
+			error_check_good txn [is_valid_txn $t $env] TRUE
+			set txn "-txn $t"
+		}
+	}
+	set db [eval {berkdb open} $envarg -rdonly -unknown $encarg $dbname]
+	error_check_good dbopen [is_valid_db $db] TRUE
 	$dump_func $db $txn $outfile $checkfunc $beg $cont
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 	error_check_good db_close [$db close] 0
 }
 
 # open file and call dump_file to dumpkeys to tempfile
 proc open_and_dump_subfile {
-    dbname dbenv txn outfile checkfunc dump_func beg cont subdb} {
+    dbname env outfile checkfunc dump_func beg cont subdb} {
+	global encrypt
+	global passwd
 	source ./include.tcl
 
-	if { $dbenv == "NULL" } {
-		set db [berkdb open -rdonly -unknown $dbname $subdb]
-		error_check_good dbopen [is_valid_db $db] TRUE
-	} else {
-		set db [berkdb open -env $dbenv -rdonly -unknown $dbname $subdb]
-		error_check_good dbopen [is_valid_db $db] TRUE
+	set encarg ""
+	if { $encrypt > 0 && $env == "NULL" } {
+		set encarg "-encryptany $passwd"
 	}
+	set envarg ""
+	set txn ""
+	set txnenv 0
+	if { $env != "NULL" } {
+		append envarg "-env $env"
+		set txnenv [is_txnenv $env]
+		if { $txnenv == 1 } {
+			append envarg " -auto_commit "
+			set t [$env txn]
+			error_check_good txn [is_valid_txn $t $env] TRUE
+			set txn "-txn $t"
+		}
+	}
+	set db [eval {berkdb open -rdonly -unknown} \
+	    $envarg $encarg {$dbname $subdb}]
+	error_check_good dbopen [is_valid_db $db] TRUE
 	$dump_func $db $txn $outfile $checkfunc $beg $cont
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 	error_check_good db_close [$db close] 0
 }
 
@@ -310,11 +351,22 @@ proc is_substr { str sub } {
 	}
 }
 
+proc is_serial { str } {
+	global serial_tests
+
+	foreach test $serial_tests {
+		if { [is_substr $str $test]  == 1 } {
+			return 1
+		}
+	}
+	return 0
+}
+
 proc release_list { l } {
 
 	# Now release all the locks
 	foreach el $l {
-		set ret [$el put]
+		catch { $el put } ret
 		error_check_good lock_put $ret 0
 	}
 }
@@ -592,16 +644,33 @@ proc sentinel_init { } {
 	}
 }
 
-proc watch_procs { {delay 30} {max 3600} {quiet 0} } {
+proc watch_procs { pidlist {delay 30} {max 3600} {quiet 0} } {
 	source ./include.tcl
 
 	set elapsed 0
+
+	# Don't start watching the processes until a sentinel
+	# file has been created for each one.
+	foreach pid $pidlist {
+		while { [file exists $testdir/begin.$pid] == 0 } {
+			tclsleep $delay
+			incr elapsed $delay
+			# If pids haven't been created in one-tenth
+			# of the time allowed for the whole test,
+			# there's a problem.  Report an error and fail.
+			if { $elapsed > [expr {$max / 10}] } {
+				puts "FAIL: begin.pid not created"
+				break
+			}
+		}
+	}
+
 	while { 1 } {
 
 		tclsleep $delay
 		incr elapsed $delay
 
-		# Find the list of processes withoutstanding sentinel
+		# Find the list of processes with outstanding sentinel
 		# files (i.e. a begin.pid and no end.pid).
 		set beginlist {}
 		set endlist {}
@@ -651,15 +720,9 @@ proc watch_procs { {delay 30} {max 3600} {quiet 0} } {
 		if { $elapsed > $max } {
 			# We have exceeded the limit; kill processes
 			# and report an error
-			set rlist {}
 			foreach i $l {
-				set r [catch { exec $KILL $i } result]
-				if { $r == 0 } {
-					lappend rlist $i
-				}
+				tclkill $i
 			}
-			error_check_good "Processes still running" \
-			    [llength $rlist] 0
 		}
 	}
 	if { $quiet == 0 } {
@@ -1002,7 +1065,7 @@ proc filecheck { file txn } {
 		unset check_array
 	}
 
-	open_and_dump_file $file NULL $txn $file.dump dbcheck dump_full_file \
+	open_and_dump_file $file NULL $file.dump dbcheck dump_full_file \
 	    "-first" "-next"
 
 	# Check that everything we checked had all its data
@@ -1033,90 +1096,62 @@ proc filecheck { file txn } {
 
 proc cleanup { dir env { quiet 0 } } {
 	global gen_upgrade
-	global upgrade_dir
-	global upgrade_be
-	global upgrade_method
-	global upgrade_name
+	global is_qnx_test
+	global old_encrypt
+	global passwd
 	source ./include.tcl
 
 	if { $gen_upgrade == 1 } {
-		set vers [berkdb version]
-		set maj [lindex $vers 0]
-		set min [lindex $vers 1]
-
-		if { $upgrade_be == 1 } {
-			set version_dir "$maj.${min}be"
-			set en be
-		} else {
-			set version_dir "$maj.${min}le"
-			set en le
-		}
-
-		set dest $upgrade_dir/$version_dir/$upgrade_method
-		exec mkdir -p $dest
-
-		set dbfiles [glob -nocomplain $dir/*.db]
-		foreach dbfile $dbfiles {
-			set basename [string range $dbfile \
-			    [expr [string length $dir] + 1] end-3]
-
-			set newbasename $upgrade_name-$basename
-
-			# db_dump file
-			error_check_good db_dump($dbfile) \
-			    [catch {exec $util_path/db_dump $dbfile > \
-			    $dir/$newbasename.dump}] 0
-
-			# tcl_dump file
-			upgrade_dump $dbfile \
-			    $dir/$newbasename.tcldump
-
-			# Rename dbfile and any dbq files.
-			file rename $dbfile $dir/$newbasename-$en.db
-			foreach dbq \
-			    [glob -nocomplain $dir/__dbq.$basename.db.*] {
-				set s [string length $dir/__dbq.]
-				set newname [string replace $dbq $s \
-				    [expr [string length $basename] + $s - 1] \
-				    $newbasename-$en]
-				file rename $dbq $newname
-			}
-			set cwd [pwd]
-			cd $dir
-			catch {eval exec tar -cvf $dest/$newbasename.tar \
-			    [glob $newbasename* __dbq.$newbasename-$en.db.*]}
-			catch {exec gzip -9v $dest/$newbasename.tar}
-			cd $cwd
-		}
+		save_upgrade_files $dir
 	}
 
 #	check_handles
 	set remfiles {}
 	set ret [catch { glob $dir/* } result]
 	if { $ret == 0 } {
-		foreach file $result {
+		foreach fileorig $result {
 			#
 			# We:
 			# - Ignore any env-related files, which are
 			# those that have __db.* or log.* if we are
-			# running in an env.
+			# running in an env.  Also ignore files whose
+			# names start with REPDIR_;  these are replication
+			# subdirectories.
 			# - Call 'dbremove' on any databases.
 			# Remove any remaining temp files.
 			#
-			switch -glob -- $file {
+			switch -glob -- $fileorig {
+			*/DIR_* -
 			*/__db.* -
 			*/log.*	{
 				if { $env != "NULL" } {
 					continue
 				} else {
-					lappend remfiles $file
+					if { $is_qnx_test } {
+						catch {berkdb envremove -force \
+						    -home $dir} r
+					}
+					lappend remfiles $fileorig
 				}
 				}
 			*.db	{
 				set envargs ""
+				set encarg ""
+				#
+				# If in an env, it should be open crypto
+				# or not already.
+				#
 				if { $env != "NULL"} {
-					set file [file tail $file]
+					set file [file tail $fileorig]
 					set envargs " -env $env "
+					if { [is_txnenv $env] } {
+						append envargs " -auto_commit "
+					}
+				} else {
+					if { $old_encrypt != 0 } {
+						set encarg "-encryptany $passwd"
+					}
+					set file $fileorig
 				}
 
 				# If a database is left in a corrupt
@@ -1126,17 +1161,39 @@ proc cleanup { dir env { quiet 0 } } {
 				# just forcibly remove the file with a warning
 				# message.
 				set ret [catch \
-				    {eval {berkdb dbremove} $envargs $file} res]
+				    {eval {berkdb dbremove} $envargs $encarg \
+				    $file} res]
+				# If dbremove failed and we're not in an env,
+				# note that we don't have 100% certainty
+				# about whether the previous run used
+				# encryption.  Try to remove with crypto if
+				# we tried without, and vice versa.
 				if { $ret != 0 } {
-					if { $quiet == 0 } {
-						puts \
+					if { $env == "NULL" && \
+					    $old_encrypt == 0} {
+						set ret [catch \
+				    		    {eval {berkdb dbremove} \
+						    -encryptany $passwd \
+				    		    $file} res]
+					}
+					if { $env == "NULL" && \
+					    $old_encrypt == 1 } {
+						set ret [catch \
+						    {eval {berkdb dbremove} \
+						    $file} res]
+					}
+					if { $ret != 0 } {
+						if { $quiet == 0 } {
+							puts \
 				    "FAIL: dbremove in cleanup failed: $res"
+						}
+						set file $fileorig
+						lappend remfiles $file
 					}
-					lappend remfiles $file
 				}
 				}
 			default	{
-				lappend remfiles $file
+				lappend remfiles $fileorig
 				}
 			}
 		}
@@ -1148,6 +1205,11 @@ proc cleanup { dir env { quiet 0 } } {
 
 proc log_cleanup { dir } {
 	source ./include.tcl
+	global gen_upgrade_log
+
+	if { $gen_upgrade_log == 1 } {
+		save_upgrade_files $dir
+	}
 
 	set files [glob -nocomplain $dir/log.*]
 	if { [llength $files] != 0} {
@@ -1158,9 +1220,15 @@ proc log_cleanup { dir } {
 }
 
 proc env_cleanup { dir } {
+	global old_encrypt
+	global passwd
 	source ./include.tcl
 
-	set stat [catch {berkdb envremove -home $dir} ret]
+	set encarg ""
+	if { $old_encrypt != 0 } {
+		set encarg "-encryptany $passwd"
+	}
+	set stat [catch {eval {berkdb envremove -home} $dir $encarg} ret]
 	#
 	# If something failed and we are left with a region entry
 	# in /dev/shmem that is zero-length, the envremove will
@@ -1181,6 +1249,51 @@ proc env_cleanup { dir } {
 	cleanup $dir NULL
 }
 
+# Start an RPC server.  Don't return to caller until the 
+# server is up.  Wait up to $maxwait seconds.
+proc rpc_server_start { { encrypted 0 } { maxwait 30 } { args "" } } {
+	source ./include.tcl
+	global rpc_svc
+	global passwd
+
+	set encargs ""
+	if { $encrypted == 1 } {
+		set encargs " -P $passwd "
+	}
+
+	if { [string compare $rpc_server "localhost"] == 0 } {
+		set dpid [eval {exec $util_path/$rpc_svc \
+		    -h $rpc_testdir} $args $encargs &]
+	} else {
+		set dpid [eval {exec rsh $rpc_server \
+		    $rpc_path/$rpc_svc -h $rpc_testdir $args} &]
+	}
+
+	# Wait a couple of seconds before we start looking for 
+	# the server.
+	tclsleep 2
+	set home [file tail $rpc_testdir]
+	if { $encrypted == 1 } {
+		set encargs " -encryptaes $passwd "
+	}
+	for { set i 0 } { $i < $maxwait } { incr i } {
+		if {[catch {set env [eval berkdb_env -create \
+		    -home $home -server $rpc_server $encargs]} res]} {
+			# If we have an error, sleep for a second.
+			tclsleep 1
+		} else {
+			# Server is up, clean up and return to caller
+			error_check_good env_close [$env close] 0
+			break
+		}
+		if { $i >= $maxwait } {
+			puts "FAIL: RPC server\
+			    not started after $maxwait seconds"
+		}
+	}
+	return $dpid
+}
+
 proc remote_cleanup { server dir localdir } {
 	set home [file tail $dir]
 	error_check_good cleanup:remove [berkdb envremove -home $home \
@@ -1276,15 +1389,16 @@ proc op_recover { encodedop dir env_cmd dbfile cmd msg } {
 		# Fork off a child to run the cmd
 		# We append the gid, so start here making sure
 		# we don't have old gid's around.
+		set outfile $testdir/childlog
 		fileremove -f $testdir/gidfile
 		set gidf $testdir/gidfile
-		set proclist {}
+		set pidlist {}
 		# puts "$tclsh_path $test_path/recdscript.tcl $testdir/recdout \
 		#    $op $dir $env_cmd $dbfile $gidf $cmd"
 		set p [exec $tclsh_path $test_path/wrap.tcl recdscript.tcl \
 		    $testdir/recdout $op $dir $env_cmd $dbfile $gidf $cmd &]
-		lappend proclist $p
-		watch_procs 1 120 1
+		lappend pidlist $p
+		watch_procs $pidlist 5
 		set f1 [open $testdir/recdout r]
 		set r [read $f1]
 		puts -nonewline $r
@@ -1295,6 +1409,7 @@ proc op_recover { encodedop dir env_cmd dbfile cmd msg } {
 	}
 	op_recover_rec $op $op2 $dir $env_cmd $dbfile $gidf
 }
+
 proc op_recover_prep { op dir env_cmd dbfile gidf cmd } {
 	global log_log_record_types
 	global recd_debug
@@ -1317,14 +1432,15 @@ proc op_recover_prep { op dir env_cmd dbfile gidf cmd } {
 	catch { file copy -force $dir/$dbfile $dir/$dbfile.init } res
 	copy_extent_file $dir $dbfile init
 
+	convert_encrypt $env_cmd
 	set env [eval $env_cmd]
 	error_check_good envopen [is_valid_env $env] TRUE
-	set db [berkdb open -env $env $dbfile]
+
+	set db [berkdb open -auto_commit -env $env $dbfile]
 	error_check_good dbopen [is_valid_db $db] TRUE
 
 	# Dump out file contents for initial case
-	set tflags ""
-	open_and_dump_file $dbfile $env $tflags $init_file nop \
+	open_and_dump_file $dbfile $env $init_file nop \
 	    dump_file_direction "-first" "-next"
 
 	set t [$env txn]
@@ -1379,17 +1495,14 @@ proc op_recover_prep { op dir env_cmd dbfile gidf cmd } {
 	set record_exec_cmd_ret 0
 	set lenient_exec_cmd_ret 0
 
-	# Sync the file so that we can capture a snapshot to test
-	# recovery.
+	# Sync the file so that we can capture a snapshot to test recovery.
 	error_check_good sync:$db [$db sync] 0
 
 	catch { file copy -force $dir/$dbfile $dir/$dbfile.afterop } res
 	copy_extent_file $dir $dbfile afterop
+	open_and_dump_file $dir/$dbfile.afterop NULL \
+		$afterop_file nop dump_file_direction "-first" "-next"
 
-	#set tflags "-txn $t"
-	open_and_dump_file $dir/$dbfile.afterop NULL $tflags \
-		$afterop_file nop dump_file_direction \
-		"-first" "-next"
 	#puts "\t\t\tExecuting txn_$op:$t"
 	if { $op == "prepare" } {
 		set gid [make_gid global:$t]
@@ -1407,13 +1520,13 @@ proc op_recover_prep { op dir env_cmd dbfile gidf cmd } {
 		"prepare" { puts "\t\tCommand executed and prepared." }
 	}
 
-	# Dump out file and save a copy.
+	# Sync the file so that we can capture a snapshot to test recovery.
 	error_check_good sync:$db [$db sync] 0
-	open_and_dump_file $dir/$dbfile NULL $tflags $final_file nop \
-	    dump_file_direction "-first" "-next"
 
 	catch { file copy -force $dir/$dbfile $dir/$dbfile.final } res
 	copy_extent_file $dir $dbfile final
+	open_and_dump_file $dir/$dbfile.final NULL \
+	    $final_file nop dump_file_direction "-first" "-next"
 
 	# If this is an abort or prepare-abort, it should match the
 	#   original file.
@@ -1468,6 +1581,8 @@ proc op_recover_rec { op op2 dir env_cmd dbfile gidf} {
 	global recd_debug
 	global recd_id
 	global recd_op
+	global encrypt
+	global passwd
 	source ./include.tcl
 
 	#puts "op_recover_rec: $op $op2 $dir $env_cmd $dbfile $gidf"
@@ -1482,19 +1597,23 @@ proc op_recover_rec { op op2 dir env_cmd dbfile gidf} {
 	}
 
 	berkdb debug_check
-	puts -nonewline "\t\tOp_recover_rec: Running recovery ... "
+	puts -nonewline "\t\top_recover_rec: Running recovery ... "
 	flush stdout
 
-	set stat [catch {exec $util_path/db_recover -h $dir -e -c} result]
+	set recargs "-h $dir -c "
+	if { $encrypt > 0 } {
+		append recargs " -P $passwd "
+	}
+	set stat [catch {eval exec $util_path/db_recover -e $recargs} result]
 	if { $stat == 1 } {
 		error "FAIL: Recovery error: $result."
 	}
 	puts -nonewline "complete ... "
 
 	#
-	# We cannot run db_recover here because that will
-	# open an env, run recovery, then close it, which will
-	# abort the outstand txns.  We want to do it ourselves.
+	# We cannot run db_recover here because that will open an env, run
+	# recovery, then close it, which will abort the outstanding txns.
+	# We want to do it ourselves.
 	#
 	set env [eval $env_cmd]
 	error_check_good dbenv [is_valid_widget $env env] TRUE
@@ -1533,8 +1652,7 @@ proc op_recover_rec { op op2 dir env_cmd dbfile gidf} {
 		}
 	}
 
-	set tflags ""
-	open_and_dump_file $dir/$dbfile NULL $tflags $final_file nop \
+	open_and_dump_file $dir/$dbfile NULL $final_file nop \
 	    dump_file_direction "-first" "-next"
 	if { $op == "commit" || $op2 == "commit" } {
 		filesort $afterop_file $afterop_file.sort
@@ -1562,11 +1680,10 @@ proc op_recover_rec { op op2 dir env_cmd dbfile gidf} {
 	}
 
 	berkdb debug_check
-	puts -nonewline \
-	    "\t\tRunning recovery on pre-op database ... "
+	puts -nonewline "\t\tRunning recovery on pre-op database ... "
 	flush stdout
 
-	set stat [catch {exec $util_path/db_recover -h $dir -c} result]
+	set stat [catch {eval exec $util_path/db_recover $recargs} result]
 	if { $stat == 1 } {
 		error "FAIL: Recovery error: $result."
 	}
@@ -1578,7 +1695,7 @@ proc op_recover_rec { op op2 dir env_cmd dbfile gidf} {
 
 	set env [eval $env_cmd]
 
-	open_and_dump_file $dir/$dbfile NULL $tflags $final_file nop \
+	open_and_dump_file $dir/$dbfile NULL $final_file nop \
 	    dump_file_direction "-first" "-next"
 	if { $op == "commit" || $op2 == "commit" } {
 		filesort $final_file $final_file.sort
@@ -1676,11 +1793,11 @@ proc minwrites { myenv locker_id obj_id num } {
 
 proc countlocks { myenv locker_id obj_id num } {
 	set locklist ""
-	for { set i 0} {$i < [expr $locker_id * 4]} { incr i } {
+	for { set i 0} {$i < [expr $obj_id * 4]} { incr i } {
 		set r [catch {$myenv lock_get read $locker_id \
-		    [expr $locker_id * 1000 + $i]} l ]
+		    [expr $obj_id * 1000 + $i]} l ]
 		if { $r != 0 } {
-			puts $errorInfo
+			puts $l
 			return ERROR
 		} else {
 			error_check_good lockget:$obj_id [is_substr $l $myenv] 1
@@ -1689,11 +1806,11 @@ proc countlocks { myenv locker_id obj_id num } {
 	}
 
 	# Now acquire a write lock
-	if { $locker_id != 1 } {
+	if { $obj_id != 1 } {
 		set r [catch {$myenv lock_get write $locker_id \
-		    [expr $locker_id * 1000 + 10]} l ]
+		    [expr $obj_id * 1000 + 10]} l ]
 		if { $r != 0 } {
-			puts $errorInfo
+			puts $l
 			return ERROR
 		} else {
 			error_check_good lockget:$obj_id [is_substr $l $myenv] 1
@@ -1721,7 +1838,7 @@ proc ring { myenv locker_id obj_id num } {
 	source ./include.tcl
 
 	if {[catch {$myenv lock_get write $locker_id $obj_id} lock1] != 0} {
-		puts $errorInfo
+		puts $lock1
 		return ERROR
 	} else {
 		error_check_good lockget:$obj_id [is_substr $lock1 $myenv] 1
@@ -1734,7 +1851,12 @@ proc ring { myenv locker_id obj_id num } {
 		if {[string match "*DEADLOCK*" $lock2] == 1} {
 			set ret DEADLOCK
 		} else {
-			set ret ERROR
+			if {[string match "*NOTGRANTED*" $lock2] == 1} {
+				set ret DEADLOCK
+			} else {
+				puts $lock2
+				set ret ERROR
+			}
 		}
 	} else {
 		error_check_good lockget:$obj_id [is_substr $lock2 $myenv] 1
@@ -1763,7 +1885,7 @@ proc clump { myenv locker_id obj_id num } {
 
 	set obj_id 10
 	if {[catch {$myenv lock_get read $locker_id $obj_id} lock1] != 0} {
-		puts $errorInfo
+		puts $lock1
 		return ERROR
 	} else {
 		error_check_good lockget:$obj_id \
@@ -1776,7 +1898,12 @@ proc clump { myenv locker_id obj_id num } {
 		if {[string match "*DEADLOCK*" $lock2] == 1} {
 			set ret DEADLOCK
 		} else {
-			set ret ERROR
+			if {[string match "*NOTGRANTED*" $lock2] == 1} {
+				set ret DEADLOCK
+			} else {
+				puts $lock2
+				set ret ERROR
+			}
 		}
 	} else {
 		error_check_good \
@@ -1792,17 +1919,29 @@ proc clump { myenv locker_id obj_id num } {
 		error_check_good lockput:$lock2 [$lock2 put] 0
 	}
 	return $ret
- }
+}
 
-proc dead_check { t procs dead clean other } {
+proc dead_check { t procs timeout dead clean other } {
 	error_check_good $t:$procs:other $other 0
 	switch $t {
 		ring {
+			# With timeouts the number of deadlocks is unpredictable:
+			# test for at least one deadlock.
+			if { $timeout != 0 && $dead > 1 } {
+				set clean [ expr $clean + $dead - 1]
+				set dead 1
+			}
 			error_check_good $t:$procs:deadlocks $dead 1
 			error_check_good $t:$procs:success $clean \
 			    [expr $procs - 1]
 		}
 		clump {
+			# With timeouts the number of deadlocks is unpredictable:
+			# test for no more than one successful lock.
+			if { $timeout != 0 && $dead == $procs } {
+				set clean 1
+				set dead [expr $procs - 1]
+			}
 			error_check_good $t:$procs:deadlocks $dead \
 			    [expr $procs - 1]
 			error_check_good $t:$procs:success $clean 1
@@ -1876,6 +2015,9 @@ proc reverse { s } {
 	return $res
 }
 
+#
+# This is a internal only proc.  All tests should use 'is_valid_db' etc.
+#
 proc is_valid_widget { w expected } {
 	# First N characters must match "expected"
 	set l [string length $expected]
@@ -1912,6 +2054,10 @@ proc is_valid_lock { lock env } {
 	return [is_valid_widget $lock $env.lock]
 }
 
+proc is_valid_logc { logc env } {
+	return [is_valid_widget $logc $env.logc]
+}
+
 proc is_valid_mpool { mpool env } {
 	return [is_valid_widget $mpool $env.mp]
 }
@@ -1928,11 +2074,20 @@ proc is_valid_mutex { m env } {
 	return [is_valid_widget $m $env.mutex]
 }
 
+proc is_valid_lock {l env} {
+	return [is_valid_widget $l $env.lock]
+}
+
+proc is_valid_locker {l } {
+	return [is_valid_widget $l ""]
+}
+
 proc send_cmd { fd cmd {sleep 2}} {
 	source ./include.tcl
 
-	puts $fd "set v \[$cmd\]"
-	puts $fd "puts \$v"
+	puts $fd "if \[catch {set v \[$cmd\] ; puts \$v} ret\] { \
+		puts \"FAIL: \$ret\" \
+	}"
 	puts $fd "flush stdout"
 	flush $fd
 	berkdb debug_check
@@ -2019,13 +2174,6 @@ proc make_fixed_length {method data {pad 0}} {
 	return $data
 }
 
-proc make_gid {data} {
-	while { [string length $data] < 127 } {
-		set data [format ${data}0]
-	}
-	return $data
-}
-
 proc make_gid {data} {
 	while { [string length $data] < 128 } {
 		set data [format ${data}0]
@@ -2136,6 +2284,32 @@ proc convert_method { method } {
 	}
 }
 
+proc split_encargs { largs encargsp } {
+	global encrypt
+	upvar $encargsp e
+	set eindex [lsearch $largs "-encrypta*"]
+	if { $eindex == -1 } {
+		set e ""
+		set newl $largs
+	} else {
+		set eend [expr $eindex + 1]
+		set e [lrange $largs $eindex $eend]
+		set newl [lreplace $largs $eindex $eend "-encrypt"]
+	}
+	return $newl
+}
+
+proc convert_encrypt { largs } {
+	global encrypt
+	global old_encrypt
+
+	set old_encrypt $encrypt
+	set encrypt 0
+	if { [lsearch $largs "-encrypt*"] != -1 } {
+		set encrypt 1
+	}
+}
+
 # If recno-with-renumbering or btree-with-renumbering is specified, then
 # fix the arguments to specify the DB_RENUMBER/DB_RECNUM option for the
 # -flags argument.
@@ -2155,6 +2329,7 @@ proc convert_args { method {largs ""} } {
 		return -code return
 	}
 
+	convert_encrypt $largs
 	if { $gen_upgrade == 1 && $upgrade_be == 1 } {
 		append largs " -lorder 4321 "
 	} elseif { $gen_upgrade == 1 && $upgrade_be != 1 } {
@@ -2386,6 +2561,7 @@ proc fileremove { args } {
 }
 
 proc findfail { args } {
+	set errstring {}
 	foreach a $args {
 		if { [file exists $a] == 0 } {
 			continue
@@ -2393,13 +2569,12 @@ proc findfail { args } {
 		set f [open $a r]
 		while { [gets $f line] >= 0 } {
 			if { [string first FAIL $line] == 0 } {
-				close $f
-				return 1
+				lappend errstring $a:$line
 			}
 		}
 		close $f
 	}
-	return 0
+	return $errstring
 }
 
 # Sleep for s seconds.
@@ -2411,6 +2586,16 @@ proc tclsleep { s } {
 	after [expr $s * 1000 + 56]
 }
 
+# Kill a process.
+proc tclkill { id } {
+	source ./include.tcl
+
+	while { [ catch {exec $KILL -0 $id} ] == 0 } {
+		catch {exec $KILL -9 $id}
+		tclsleep 5
+	}
+}
+
 # Compare two files, a la diff.  Returns 1 if non-identical, 0 if identical.
 proc filecmp { file_a file_b } {
 	set fda [open $file_a r]
@@ -2437,9 +2622,39 @@ proc filecmp { file_a file_b } {
 	return 0
 }
 
+# Give two SORTED files, one of which is a complete superset of the other,
+# extract out the unique portions of the superset and put them in
+# the given outfile.
+proc fileextract { superset subset outfile } {
+	set sup [open $superset r]
+	set sub [open $subset r]
+	set outf [open $outfile w]
+
+	# The gets can't be in the while condition because we'll
+	# get short-circuit evaluated.
+	set nrp [gets $sup pline]
+	set nrb [gets $sub bline]
+	while { $nrp >= 0 } {
+		if { $nrp != $nrb || [string compare $pline $bline] != 0} {
+			puts $outf $pline
+		} else {
+			set nrb [gets $sub bline]
+		}
+		set nrp [gets $sup pline]
+	}
+
+	close $sup
+	close $sub
+	close $outf
+	return 0
+}
+
 # Verify all .db files in the specified directory.
-proc verify_dir { {directory "./TESTDIR"} \
+proc verify_dir { {directory $testdir} \
     { pref "" } { noredo 0 } { quiet 0 } { nodump 0 } { cachesize 0 } } {
+	global encrypt
+	global passwd
+
 	# If we're doing database verification between tests, we don't
 	# want to do verification twice without an intervening cleanup--some
 	# test was skipped.  Always verify by default (noredo == 0) so
@@ -2450,7 +2665,7 @@ proc verify_dir { {directory "./TESTDIR"} \
 			if { $quiet == 0 } {
 				puts "Skipping verification."
 			}
-			return
+			return 0
 		}
 		set f [open $directory/NOREVERIFY w]
 		close $f
@@ -2458,25 +2673,27 @@ proc verify_dir { {directory "./TESTDIR"} \
 
 	if { [catch {glob $directory/*.db} dbs] != 0 } {
 		# No files matched
-		return
-	}
-	if { [file exists /dev/stderr] == 1 } {
-		set errfilearg "-errfile /dev/stderr "
-	} else {
-		set errfilearg ""
+		return 0
 	}
+	set errfilearg "-errfile /dev/stderr "
 	set errpfxarg {-errpfx "FAIL: verify" }
 	set errarg $errfilearg$errpfxarg
 	set ret 0
 
-	# If we're using a non-default cachesize, open an env.
-	if { $cachesize != 0 } {
-		set env [berkdb env -create \
-		    -private -cachesize [list 0 $cachesize 0]]
-		set earg " -env $env $errarg "
-	} else {
-		set earg $errarg
+	# Open an env, so that we have a large enough cache.  Pick
+	# a fairly generous default if we haven't specified something else.
+
+	if { $cachesize == 0 } {
+		set cachesize [expr 1024 * 1024]
 	}
+	set encarg ""
+	if { $encrypt != 0 } {
+		set encarg "-encryptaes $passwd"
+	}
+
+	set env [eval {berkdb_env -create -private} $encarg \
+	    {-cachesize [list 0 $cachesize 0]}]
+	set earg " -env $env $errarg "
 
 	foreach db $dbs {
 		if { [catch {eval {berkdb dbverify} $earg $db} res] != 0 } {
@@ -2491,8 +2708,7 @@ proc verify_dir { {directory "./TESTDIR"} \
 			}
 		}
 
-		# Skip the dump if it's dangerous to do it (e.g. subdbs).
-		# XXX we should make dumploadtest work for subdbs instead.
+		# Skip the dump if it's dangerous to do it.
 		if { $nodump == 0 } {
 			if { [catch {eval dumploadtest $db} res] != 0 } {
 				puts $res
@@ -2509,63 +2725,124 @@ proc verify_dir { {directory "./TESTDIR"} \
 		}
 	}
 
-	if { $cachesize != 0 } {
-		error_check_good vrfyenv_close [$env close] 0
-	}
+	error_check_good vrfyenv_close [$env close] 0
 
 	return $ret
 }
 
-proc dumploadtest { db } {
-	global util_path
-
-	set newdbname $db-dumpload.db
-
-	# Do a db_dump test.  Dump/load each file.
-	set rval [catch {exec $util_path/db_dump -k $db | \
-	    $util_path/db_load $newdbname} res]
-	error_check_good db_dump($db) $rval 0
-
-	# Now open original and dump/loaded databases.
-	set olddb [eval {berkdb_open -rdonly $db}]
-	set newdb [eval {berkdb_open -rdonly $newdbname}]
-	error_check_good olddb($db) [is_valid_db $olddb] TRUE
-	error_check_good newdb($db) [is_valid_db $newdb] TRUE
+# Is the database handle in $db a master database containing subdbs?
+proc check_for_subdbs { db } {
+	set stat [$db stat]
+	for { set i 0 } { [string length [lindex $stat $i]] > 0 } { incr i } {
+		set elem [lindex $stat $i]
+		if { [string compare [lindex $elem 0] Flags] == 0 } {
+			# This is the list of flags;  look for
+			# "subdatabases".
+			if { [is_substr [lindex $elem 1] subdatabases] } {
+				return 1
+			}
+		}
+	}
+	return 0
+}
 
+proc db_compare { olddb newdb olddbname newdbname } {
 	# Walk through olddb and newdb and make sure their contents
 	# are identical.
 	set oc [$olddb cursor]
 	set nc [$newdb cursor]
-	error_check_good orig_cursor($db) \
+	error_check_good orig_cursor($olddbname) \
 	    [is_valid_cursor $oc $olddb] TRUE
-	error_check_good new_cursor($db) \
+	error_check_good new_cursor($olddbname) \
 	    [is_valid_cursor $nc $newdb] TRUE
 
 	for { set odbt [$oc get -first] } { [llength $odbt] > 0 } \
 	    { set odbt [$oc get -next] } {
 		set ndbt [$nc get -get_both \
 		    [lindex [lindex $odbt 0] 0] [lindex [lindex $odbt 0] 1]]
-		error_check_good db_compare($db) $ndbt $odbt
+		error_check_good db_compare($olddbname/$newdbname) $ndbt $odbt
 	}
 
 	for { set ndbt [$nc get -first] } { [llength $ndbt] > 0 } \
 	    { set ndbt [$nc get -next] } {
 		set odbt [$oc get -get_both \
 		    [lindex [lindex $ndbt 0] 0] [lindex [lindex $ndbt 0] 1]]
-		error_check_good db_compare_back($db) $odbt $ndbt
+		error_check_good db_compare_back($olddbname) $odbt $ndbt
 	}
 
-	error_check_good orig_cursor_close($db) [$oc close] 0
-	error_check_good new_cursor_close($db) [$nc close] 0
-
-	error_check_good orig_db_close($db) [$olddb close] 0
-	error_check_good new_db_close($db) [$newdb close] 0
-
-	berkdb dbremove $newdbname
+	error_check_good orig_cursor_close($olddbname) [$oc close] 0
+	error_check_good new_cursor_close($olddbname) [$nc close] 0
 
 	return 0
 }
 
+proc dumploadtest { db } {
+	global util_path
+	global encrypt
+	global passwd
+
+	set newdbname $db-dumpload.db
+
+	set dbarg ""
+	set utilflag ""
+	if { $encrypt != 0 } {
+		set dbarg "-encryptany $passwd"
+		set utilflag "-P $passwd"
+	}
+
+	# Dump/load the whole file, including all subdbs.
+	set rval [catch {eval {exec $util_path/db_dump} $utilflag -k \
+	    $db | $util_path/db_load $utilflag $newdbname} res]
+	error_check_good db_dump/db_load($db:$res) $rval 0
+
+	# If the old file was empty, there's no new file and we're done.
+	if { [file exists $newdbname] == 0 } {
+		return 0
+	}
+
+	# Open original database.
+	set olddb [eval {berkdb_open -rdonly} $dbarg $db]
+	error_check_good olddb($db) [is_valid_db $olddb] TRUE
+
+	if { [check_for_subdbs $olddb] } {
+		# If $db has subdatabases, compare each one separately.
+		set oc [$olddb cursor]
+		error_check_good orig_cursor($db) \
+    		    [is_valid_cursor $oc $olddb] TRUE
+
+		for { set dbt [$oc get -first] } \
+		    { [llength $dbt] > 0 } \
+		    { set dbt [$oc get -next] } {
+			set subdb [lindex [lindex $dbt 0] 0]
+
+			set oldsubdb \
+			    [eval {berkdb_open -rdonly} $dbarg {$db $subdb}]
+			error_check_good olddb($db) [is_valid_db $oldsubdb] TRUE
+
+			# Open the new database.
+			set newdb \
+			    [eval {berkdb_open -rdonly} $dbarg {$newdbname $subdb}]
+			error_check_good newdb($db) [is_valid_db $newdb] TRUE
+
+			db_compare $oldsubdb $newdb $db $newdbname
+			error_check_good new_db_close($db) [$newdb close] 0
+			error_check_good old_subdb_close($oldsubdb) [$oldsubdb close] 0
+		}
+
+		error_check_good oldcclose [$oc close] 0
+	} else {
+		# Open the new database.
+		set newdb [eval {berkdb_open -rdonly} $dbarg $newdbname]
+		error_check_good newdb($db) [is_valid_db $newdb] TRUE
+
+		db_compare $olddb $newdb $db $newdbname
+		error_check_good new_db_close($db) [$newdb close] 0
+	}
+
+	error_check_good orig_db_close($db) [$olddb close] 0
+	eval berkdb dbremove $dbarg $newdbname
+}
+
 # Generate randomly ordered, guaranteed-unique four-character strings that can
 # be used to differentiate duplicates without creating duplicate duplicates.
 # (test031 & test032) randstring_init is required before the first call to
@@ -2672,16 +2949,16 @@ proc extractflags { args } {
 # Wrapper for berkdb open, used throughout the test suite so that we can
 # set an errfile/errpfx as appropriate.
 proc berkdb_open { args } {
-	global is_envmethod1
+	global is_envmethod
 
-	if { [info exists is_envmethod1] == 0 } {
-		set is_envmethod1 0
+	if { [info exists is_envmethod] == 0 } {
+		set is_envmethod 0
 	}
 
 	set errargs {}
-	if { $is_envmethod1 == 0 && [file exists /dev/stderr] == 1 } {
+	if { $is_envmethod == 0 } {
 		append errargs " -errfile /dev/stderr "
-		append errargs " -errpfx \\F\\A\\I\\L "
+		append errargs " -errpfx \\F\\A\\I\\L"
 	}
 
 	eval {berkdb open} $errargs $args
@@ -2692,6 +2969,29 @@ proc berkdb_open_noerr { args } {
 	eval {berkdb open} $args
 }
 
+# Wrapper for berkdb env, used throughout the test suite so that we can
+# set an errfile/errpfx as appropriate.
+proc berkdb_env { args } {
+	global is_envmethod
+
+	if { [info exists is_envmethod] == 0 } {
+		set is_envmethod 0
+	}
+
+	set errargs {}
+	if { $is_envmethod == 0 } {
+		append errargs " -errfile /dev/stderr "
+		append errargs " -errpfx \\F\\A\\I\\L"
+	}
+
+	eval {berkdb env} $errargs $args
+}
+
+# Version without errpfx/errfile, used when we're expecting a failure.
+proc berkdb_env_noerr { args } {
+	eval {berkdb env} $args
+}
+
 proc check_handles { {outf stdout} } {
 	global ohandles
 
@@ -2785,17 +3085,138 @@ proc get_pagesize { stat } {
 proc get_file_list { {small 0} } {
 	global is_windows_test
 	global is_qnx_test
+	global src_root
 
 	if { $is_qnx_test } {
 		set small 1
 	}
 	if { $small && $is_windows_test } {
-		return [glob ../*/*.c */env*.obj]
+		return [glob $src_root/*/*.c */env*.obj]
 	} elseif { $small } {
-		return [glob ../*/*.c ./env*.o]
+		return [glob $src_root/*/*.c ./env*.o]
 	} elseif { $is_windows_test } {
-		return [glob ../*/*.c */*.obj */libdb??.dll */libdb??d.dll]
+		return \
+		    [glob $src_root/*/*.c */*.obj */libdb??.dll */libdb??d.dll]
 	} else {
-		return [glob ../*/*.c ./*.o ./.libs/libdb-?.?.s?]
+		return [glob $src_root/*/*.c ./*.o ./.libs/libdb-?.?.s?]
+	}
+}
+
+proc is_cdbenv { env } {
+	set sys [$env attributes]
+	if { [lsearch $sys -cdb] != -1 } {
+		return 1
+	} else {
+		return 0
+	}
+}
+
+proc is_lockenv { env } {
+	set sys [$env attributes]
+	if { [lsearch $sys -lock] != -1 } {
+		return 1
+	} else {
+		return 0
+	}
+}
+
+proc is_logenv { env } {
+	set sys [$env attributes]
+	if { [lsearch $sys -log] != -1 } {
+		return 1
+	} else {
+		return 0
+	}
+}
+
+proc is_mpoolenv { env } {
+	set sys [$env attributes]
+	if { [lsearch $sys -mpool] != -1 } {
+		return 1
+	} else {
+		return 0
+	}
+}
+
+proc is_repenv { env } {
+	set sys [$env attributes]
+	if { [lsearch $sys -rep] != -1 } {
+		return 1
+	} else {
+		return 0
+	}
+}
+
+proc is_rpcenv { env } {
+	set sys [$env attributes]
+	if { [lsearch $sys -rpc] != -1 } {
+		return 1
+	} else {
+		return 0
+	}
+}
+
+proc is_secenv { env } {
+	set sys [$env attributes]
+	if { [lsearch $sys -crypto] != -1 } {
+		return 1
+	} else {
+		return 0
+	}
+}
+
+proc is_txnenv { env } {
+	set sys [$env attributes]
+	if { [lsearch $sys -txn] != -1 } {
+		return 1
+	} else {
+		return 0
+	}
+}
+
+proc get_home { env } {
+	set sys [$env attributes]
+	set h [lsearch $sys -home]
+	if { $h == -1 } {
+		return NULL
+	}
+	incr h
+	return [lindex $sys $h]
+}
+
+proc reduce_dups { nent ndp } {
+	upvar $nent nentries
+	upvar $ndp ndups
+
+	# If we are using a txnenv, assume it is using
+	# the default maximum number of locks, cut back
+	# so that we don't run out of locks.  Reduce
+	# by 25% until we fit.
+	#
+	while { [expr $nentries * $ndups] > 5000 } {
+		set nentries [expr ($nentries / 4) * 3]
+		set ndups [expr ($ndups / 4) * 3]
+	}
+}
+
+proc getstats { statlist field } {
+	foreach pair $statlist {
+		set txt [lindex $pair 0]
+		if { [string equal $txt $field] == 1 } {
+			return [lindex $pair 1]
+		}
+	}
+	return -1
+}
+
+proc big_endian { } {
+	global tcl_platform
+	set e $tcl_platform(byteOrder)
+	if { [string compare $e littleEndian] == 0 } {
+		return 0
+	} elseif { [string compare $e bigEndian] == 0 } {
+		return 1
+	} else {
+		error "FAIL: Unknown endianness $e"
 	}
 }
