@@ -1,7 +1,7 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2002
+ * Copyright (c) 1996-2003
  *	Sleepycat Software.  All rights reserved.
  */
 /*
@@ -39,7 +39,7 @@
 #include "db_config.h"
 
 #ifndef lint
-static const char revid[] = "$Id: txn.c,v 11.179 2002/08/29 17:41:17 margo Exp $";
+static const char revid[] = "$Id: txn.c,v 11.219 2003/12/03 14:33:06 bostic Exp $";
 #endif /* not lint */
 
 #ifndef NO_SYSTEM_INCLUDES
@@ -68,17 +68,18 @@ static const char revid[] = "$Id: txn.c,v 11.179 2002/08/29 17:41:17 margo Exp $
 #include "dbinc/hash.h"
 #include "dbinc/lock.h"
 #include "dbinc/log.h"
+#include "dbinc/mp.h"
 #include "dbinc/txn.h"
 
 #define	SET_LOG_FLAGS(dbenv, txnp, lflags)				\
 	do {								\
-		lflags = DB_COMMIT | DB_PERMANENT;			\
+		lflags = DB_LOG_COMMIT | DB_LOG_PERM;			\
 		if (F_ISSET(txnp, TXN_SYNC))				\
 			lflags |= DB_FLUSH;				\
 		else if (!F_ISSET(txnp, TXN_NOSYNC) &&			\
 		    !F_ISSET(dbenv, DB_ENV_TXN_NOSYNC)) {		\
 			if (F_ISSET(dbenv, DB_ENV_TXN_WRITE_NOSYNC))	\
-				lflags |= DB_WRNOSYNC;			\
+				lflags |= DB_LOG_WRNOSYNC;		\
 			else						\
 				lflags |= DB_FLUSH;			\
 		}							\
@@ -96,68 +97,67 @@ typedef enum {
 	TXN_OP_PREPARE
 } txnop_t;
 
+static int  __txn_abort_pp __P((DB_TXN *));
 static int  __txn_begin_int __P((DB_TXN *, int));
+static int  __txn_commit_pp __P((DB_TXN *, u_int32_t));
+static int  __txn_discard_pp __P((DB_TXN *, u_int32_t));
 static int  __txn_end __P((DB_TXN *, int));
 static int  __txn_isvalid __P((const DB_TXN *, TXN_DETAIL **, txnop_t));
-static int  __txn_set_timeout __P(( DB_TXN *, db_timeout_t, u_int32_t));
 static int  __txn_undo __P((DB_TXN *));
+static int  __txn_dispatch_undo __P((DB_ENV *,
+		DB_TXN *, DBT *, DB_LSN *, void *));
 
-#ifndef db_create
 /*
- * txn_abort --
- * txn_begin --
- * txn_commit --
+ * __txn_begin_pp --
+ *	DB_ENV->txn_begin pre/post processing.
  *
- * When we switched to methods in 4.0, we guessed txn_{abort,begin,commit}
- * were the interfaces applications would likely use and not be willing to
- * change, due to the sheer volume of the calls.  Provide wrappers -- we
- * could do txn_abort and txn_commit using macros, but not txn_begin, as
- * the name of the field is txn_begin, we didn't want to modify it.
- *
- * The issue with txn_begin hits us in another way.  If configured with the
- * --with-uniquename option, we use #defines to re-define DB's interfaces
- * to unique names.  We can't do that for these functions because txn_begin
- * is also a field name in the DB_ENV structure, and the #defines we use go
- * at the end of the db.h file -- we get control too late to #define a field
- * name.  So, modify the script that generates the unique names #defines to
- * not generate them for these three functions, and don't include the three
- * functions in libraries built with that configuration option.
- *
- * EXTERN: int txn_abort __P((DB_TXN *));
- * EXTERN: int txn_begin __P((DB_ENV *, DB_TXN *, DB_TXN **, u_int32_t));
- * EXTERN: int txn_commit __P((DB_TXN *, u_int32_t));
+ * PUBLIC: int __txn_begin_pp __P((DB_ENV *, DB_TXN *, DB_TXN **, u_int32_t));
  */
 int
-txn_abort(txnp)
-	DB_TXN *txnp;
-{
-	return (txnp->abort(txnp));
-}
-
-int
-txn_begin(dbenv, parent, txnpp, flags)
+__txn_begin_pp(dbenv, parent, txnpp, flags)
 	DB_ENV *dbenv;
 	DB_TXN *parent, **txnpp;
 	u_int32_t flags;
 {
-	return (dbenv->txn_begin(dbenv, parent, txnpp, flags));
-}
+	int rep_check, ret;
 
-int
-txn_commit(txnp, flags)
-	DB_TXN *txnp;
-	u_int32_t flags;
-{
-	return (txnp->commit(txnp, flags));
+	PANIC_CHECK(dbenv);
+	ENV_REQUIRES_CONFIG(dbenv, dbenv->tx_handle, "txn_begin", DB_INIT_TXN);
+
+	if ((ret = __db_fchk(dbenv,
+	    "txn_begin", flags,
+	    DB_DIRTY_READ | DB_TXN_NOWAIT |
+	    DB_TXN_NOSYNC | DB_TXN_SYNC)) != 0)
+		return (ret);
+	if ((ret = __db_fcchk(dbenv,
+	    "txn_begin", flags, DB_TXN_NOSYNC, DB_TXN_SYNC)) != 0)
+		return (ret);
+
+	if (parent == NULL) {
+		rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
+		if (rep_check)
+			__op_rep_enter(dbenv);
+	} else
+		rep_check = 0;
+	ret = __txn_begin(dbenv, parent, txnpp, flags);
+	/*
+	 * We only decrement the count if the operation fails.
+	 * Otherwise the count will be decremented when the
+	 * txn is resolved by txn_commit, txn_abort, etc.
+	 */
+	if (ret != 0 && rep_check)
+		__op_rep_exit(dbenv);
+
+	return (ret);
 }
-#endif /* !db_create */
 
 /*
  * __txn_begin --
- *	This is a wrapper to the actual begin process.  Normal transaction
- * begin allocates a DB_TXN structure for the caller, while XA transaction
- * begin does not.  Other than that, both call into common __txn_begin_int
- * code.
+ *	DB_ENV->txn_begin.
+ *
+ * This is a wrapper to the actual begin process.  Normal transaction begin
+ * allocates a DB_TXN structure for the caller, while XA transaction begin
+ * does not.  Other than that, both call into common __txn_begin_int code.
  *
  * Internally, we use TXN_DETAIL structures, but the DB_TXN structure
  * provides access to the transaction ID and the offset in the transaction
@@ -176,18 +176,6 @@ __txn_begin(dbenv, parent, txnpp, flags)
 	int ret;
 
 	*txnpp = NULL;
-	PANIC_CHECK(dbenv);
-	ENV_REQUIRES_CONFIG(dbenv, dbenv->tx_handle, "txn_begin", DB_INIT_TXN);
-
-	if ((ret = __db_fchk(dbenv,
-	    "txn_begin", flags,
-	    DB_DIRTY_READ | DB_TXN_NOWAIT |
-	    DB_TXN_NOSYNC | DB_TXN_SYNC)) != 0)
-		return (ret);
-	if ((ret = __db_fcchk(dbenv,
-	    "txn_begin", flags, DB_TXN_NOSYNC, DB_TXN_SYNC)) != 0)
-		return (ret);
-
 	if ((ret = __os_calloc(dbenv, 1, sizeof(DB_TXN), &txn)) != 0)
 		return (ret);
 
@@ -195,6 +183,7 @@ __txn_begin(dbenv, parent, txnpp, flags)
 	txn->parent = parent;
 	TAILQ_INIT(&txn->kids);
 	TAILQ_INIT(&txn->events);
+	STAILQ_INIT(&txn->logs);
 	txn->flags = TXN_MALLOC;
 	if (LF_ISSET(DB_DIRTY_READ))
 		F_SET(txn, TXN_DIRTY_READ);
@@ -256,11 +245,22 @@ __txn_xa_begin(dbenv, txn)
 {
 	PANIC_CHECK(dbenv);
 
-	memset(txn, 0, sizeof(DB_TXN));
-
+	/*
+	 * We need to initialize the transaction structure, but we must
+	 * be careful not to smash the links.  We manually intialize the
+	 * structure.
+	 */
 	txn->mgrp = dbenv->tx_handle;
 	TAILQ_INIT(&txn->kids);
 	TAILQ_INIT(&txn->events);
+	STAILQ_INIT(&txn->logs);
+	txn->parent = NULL;
+	ZERO_LSN(txn->last_lsn);
+	txn->txnid = TXN_INVALID;
+	txn->tid = 0;
+	txn->cursors = 0;
+	memset(&txn->lock_timeout, 0, sizeof(db_timeout_t));
+	memset(&txn->expire, 0, sizeof(db_timeout_t));
 
 	return (__txn_begin_int(txn, 0));
 }
@@ -289,8 +289,8 @@ __txn_compensate_begin(dbenv, txnpp)
 	txn->mgrp = dbenv->tx_handle;
 	TAILQ_INIT(&txn->kids);
 	TAILQ_INIT(&txn->events);
-	txn->flags = TXN_MALLOC;
-	F_SET(txn, TXN_COMPENSATE);
+	STAILQ_INIT(&txn->logs);
+	txn->flags = TXN_COMPENSATE | TXN_MALLOC;
 
 	*txnpp = txn;
 	return (__txn_begin_int(txn, 1));
@@ -330,6 +330,8 @@ __txn_begin_int(txn, internal)
 	 */
 	if (DBENV_LOGGING(dbenv))
 		__log_txn_lsn(dbenv, &begin_lsn, NULL, NULL);
+	else
+		ZERO_LSN(begin_lsn);
 
 	R_LOCK(dbenv, &mgr->reginfo);
 	if (!F_ISSET(txn, TXN_COMPENSATE) && F_ISSET(region, TXN_IN_RECOVERY)) {
@@ -392,15 +394,16 @@ __txn_begin_int(txn, internal)
 		region->stat.st_maxnactive = region->stat.st_nactive;
 
 	td->txnid = id;
-	td->begin_lsn = begin_lsn;
 	ZERO_LSN(td->last_lsn);
-	td->status = TXN_RUNNING;
+	td->begin_lsn = begin_lsn;
 	if (txn->parent != NULL)
 		td->parent = txn->parent->off;
 	else
 		td->parent = INVALID_ROFF;
-
+	td->status = TXN_RUNNING;
 	td->flags = 0;
+	td->xa_status = 0;
+
 	off = R_OFFSET(&mgr->reginfo, td);
 	R_UNLOCK(dbenv, &mgr->reginfo);
 
@@ -408,9 +411,9 @@ __txn_begin_int(txn, internal)
 	txn->txnid = id;
 	txn->off = (u_int32_t)off;
 
-	txn->abort = __txn_abort;
-	txn->commit = __txn_commit;
-	txn->discard = __txn_discard;
+	txn->abort = __txn_abort_pp;
+	txn->commit = __txn_commit_pp;
+	txn->discard = __txn_discard_pp;
 	txn->id = __txn_id;
 	txn->prepare = __txn_prepare;
 	txn->set_timeout = __txn_set_timeout;
@@ -436,6 +439,26 @@ err:	R_UNLOCK(dbenv, &mgr->reginfo);
 	return (ret);
 }
 
+/*
+ * __txn_commit_pp --
+ *	Interface routine to TXN->commit.
+ */
+static int
+__txn_commit_pp(txnp, flags)
+	DB_TXN *txnp;
+	u_int32_t flags;
+{
+	DB_ENV *dbenv;
+	int not_child, ret;
+
+	dbenv = txnp->mgrp->dbenv;
+	not_child = txnp->parent == NULL;
+	ret = __txn_commit(txnp, flags);
+	if (not_child && IS_ENV_REPLICATED(dbenv))
+		__op_rep_exit(dbenv);
+	return (ret);
+}
+
 /*
  * __txn_commit --
  *	Commit a transaction.
@@ -447,6 +470,7 @@ __txn_commit(txnp, flags)
 	DB_TXN *txnp;
 	u_int32_t flags;
 {
+	DBT list_dbt;
 	DB_ENV *dbenv;
 	DB_LOCKREQ request;
 	DB_TXN *kid;
@@ -489,11 +513,78 @@ __txn_commit(txnp, flags)
 	 * Abort should never fail; if it does, we bail out immediately.
 	 */
 	while ((kid = TAILQ_FIRST(&txnp->kids)) != NULL)
-		if ((ret = kid->commit(kid, flags)) != 0)
+		if ((ret = __txn_commit(kid, flags)) != 0)
 			while ((kid = TAILQ_FIRST(&txnp->kids)) != NULL)
-				if ((t_ret = kid->abort(kid)) != 0)
+				if ((t_ret = __txn_abort(kid)) != 0)
 					return (__db_panic(dbenv, t_ret));
 
+	/*
+	 * If there are any log records, write a log record and sync the log,
+	 * else do no log writes.  If the commit is for a child transaction,
+	 * we do not need to commit the child synchronously since it may still
+	 * abort (if its parent aborts), and otherwise its parent or ultimate
+	 * ancestor will write synchronously.
+	 */
+	if (DBENV_LOGGING(dbenv) && (!IS_ZERO_LSN(txnp->last_lsn) ||
+	    STAILQ_FIRST(&txnp->logs) != NULL)) {
+		if (txnp->parent == NULL) {
+			/*
+			 * We are about to free all the read locks for this
+			 * transaction below.  Some of those locks might be
+			 * handle locks which should not be freed, because
+			 * they will be freed when the handle is closed. Check
+			 * the events and preprocess any trades now so we don't
+			 * release the locks below.
+			 */
+			if ((ret =
+			    __txn_doevents(dbenv, txnp, TXN_PREPARE, 1)) != 0)
+				goto err;
+
+			memset(&request, 0, sizeof(request));
+			if (LOCKING_ON(dbenv)) {
+				request.op = DB_LOCK_PUT_READ;
+				if (IS_REP_MASTER(dbenv) &&
+				    !IS_ZERO_LSN(txnp->last_lsn)) {
+					memset(&list_dbt, 0, sizeof(list_dbt));
+					request.obj = &list_dbt;
+				}
+				ret = __lock_vec(dbenv,
+				    txnp->txnid, 0, &request, 1, NULL);
+			}
+
+			if (ret == 0 && !IS_ZERO_LSN(txnp->last_lsn)) {
+				SET_LOG_FLAGS(dbenv, txnp, lflags);
+				ret = __txn_regop_log(dbenv, txnp,
+				    &txnp->last_lsn, lflags, TXN_COMMIT,
+				    (int32_t)time(NULL), request.obj);
+			}
+
+			if (request.obj != NULL && request.obj->data != NULL)
+				__os_free(dbenv, request.obj->data);
+			if (ret != 0)
+				goto err;
+		} else {
+			/* Log the commit in the parent! */
+			if (!IS_ZERO_LSN(txnp->last_lsn) &&
+			    (ret = __txn_child_log(dbenv,
+			    txnp->parent, &txnp->parent->last_lsn,
+			    0, txnp->txnid, &txnp->last_lsn)) != 0) {
+				goto err;
+			}
+			if (STAILQ_FIRST(&txnp->logs) != NULL) {
+				/*
+				 * Put the child first so we back it out first.
+				 * All records are undone in reverse order.
+				 */
+				STAILQ_CONCAT(&txnp->logs, &txnp->parent->logs);
+				txnp->parent->logs = txnp->logs;
+				STAILQ_INIT(&txnp->logs);
+			}
+
+			F_SET(txnp->parent, TXN_CHILDCOMMIT);
+		}
+	}
+
 	/*
 	 * Process any aborted pages from our children.
 	 * We delay putting pages on the free list that are newly
@@ -502,7 +593,8 @@ __txn_commit(txnp, flags)
 	 * these pages which were not on the free list before.
 	 */
 	if (txnp->txn_list != NULL) {
-		t_ret = __db_do_the_limbo(dbenv, NULL, txnp, txnp->txn_list);
+		t_ret = __db_do_the_limbo(dbenv,
+		      NULL, txnp, txnp->txn_list, LIMBO_NORMAL);
 		__db_txnlist_end(dbenv, txnp->txn_list);
 		txnp->txn_list = NULL;
 		if (t_ret != 0 && ret == 0)
@@ -512,48 +604,6 @@ __txn_commit(txnp, flags)
 	if (ret != 0)
 		goto err;
 
-	/*
-	 * If there are any log records, write a log record and sync the log,
-	 * else do no log writes.  If the commit is for a child transaction,
-	 * we do not need to commit the child synchronously since it may still
-	 * abort (if its parent aborts), and otherwise its parent or ultimate
-	 * ancestor will write synchronously.
-	 */
-	if (DBENV_LOGGING(dbenv) && !IS_ZERO_LSN(txnp->last_lsn)) {
-		if (txnp->parent == NULL) {
-			/*
-			 * We are about to free all the read locks
-			 * for this transaction below.  Some of those
-			 * locks might be handle locks which should
-			 * not be freed, because they will be freed
-			 * when the handle is closed.  Check the
-			 * events and preprocess any trades now so
-			 * that we don't release the locks below.
-			 */
-			if ((ret = __txn_doevents(dbenv, txnp, 0, 1)) != 0)
-				goto err;
-			request.op = DB_LOCK_PUT_READ;
-			if (LOCKING_ON(dbenv) && (ret = dbenv->lock_vec(
-			   dbenv, txnp->txnid, 0, &request, 1, NULL)) != 0)
-				goto err;
-
-			SET_LOG_FLAGS(dbenv, txnp, lflags);
-			if ((ret = __txn_regop_log(dbenv,
-			    txnp, &txnp->last_lsn, lflags,
-			    TXN_COMMIT, (int32_t)time(NULL))) != 0)
-				goto err;
-		} else {
-			/* Log the commit in the parent! */
-			if ((ret = __txn_child_log(dbenv,
-			    txnp->parent, &txnp->parent->last_lsn,
-			    0, txnp->txnid, &txnp->last_lsn)) != 0) {
-				goto err;
-			}
-
-			F_SET(txnp->parent, TXN_CHILDCOMMIT);
-		}
-	}
-
 	/* This is OK because __txn_end can only fail with a panic. */
 	return (__txn_end(txnp, 1));
 
@@ -568,12 +618,31 @@ err:	/*
 	 */
 	if (td->status == TXN_PREPARED)
 		return (__db_panic(dbenv, ret));
-		
-	if ((t_ret = txnp->abort(txnp)) != 0)
+
+	if ((t_ret = __txn_abort(txnp)) != 0)
 		ret = t_ret;
 	return (ret);
 }
 
+/*
+ * __txn_abort_pp --
+ *	Interface routine to TXN->abort.
+ */
+static int
+__txn_abort_pp(txnp)
+	DB_TXN *txnp;
+{
+	DB_ENV *dbenv;
+	int not_child, ret;
+
+	dbenv = txnp->mgrp->dbenv;
+	not_child = txnp->parent == NULL;
+	ret = __txn_abort(txnp);
+	if (not_child && IS_ENV_REPLICATED(dbenv))
+		__op_rep_exit(dbenv);
+	return (ret);
+}
+
 /*
  * __txn_abort --
  *	Abort a transaction.
@@ -607,7 +676,7 @@ __txn_abort(txnp)
 	 * up.
 	 */
 	while ((kid = TAILQ_FIRST(&txnp->kids)) != NULL)
-		if ((ret = kid->abort(kid)) != 0)
+		if ((ret = __txn_abort(kid)) != 0)
 			return (ret);
 
 	if (LOCKING_ON(dbenv)) {
@@ -618,7 +687,7 @@ __txn_abort(txnp)
 		 * handle is closed.  Check the events and preprocess any
 		 * trades now so that we don't release the locks below.
 		 */
-		if ((ret = __txn_doevents(dbenv, txnp, 0, 1)) != 0)
+		if ((ret = __txn_doevents(dbenv, txnp, TXN_ABORT, 1)) != 0)
 			return (__db_panic(dbenv, ret));
 
 		/* Turn off timeouts. */
@@ -631,7 +700,8 @@ __txn_abort(txnp)
 			return (__db_panic(dbenv, ret));
 
 		request.op = DB_LOCK_UPGRADE_WRITE;
-		if ((ret = dbenv->lock_vec(
+		request.obj = NULL;
+		if ((ret = __lock_vec(
 		    dbenv, txnp->txnid, 0, &request, 1, NULL)) != 0)
 			return (__db_panic(dbenv, ret));
 	}
@@ -647,13 +717,33 @@ __txn_abort(txnp)
 	SET_LOG_FLAGS(dbenv, txnp, lflags);
 	if (DBENV_LOGGING(dbenv) && td->status == TXN_PREPARED &&
 	    (ret = __txn_regop_log(dbenv, txnp, &txnp->last_lsn,
-	    lflags, TXN_ABORT, (int32_t)time(NULL))) != 0)
+	    lflags, TXN_ABORT, (int32_t)time(NULL), NULL)) != 0)
 		return (__db_panic(dbenv, ret));
 
 	/* __txn_end always panics if it errors, so pass the return along. */
 	return (__txn_end(txnp, 0));
 }
 
+/*
+ * __txn_discard_pp --
+ *	Interface routine to TXN->discard.
+ */
+static int
+__txn_discard_pp(txnp, flags)
+	DB_TXN *txnp;
+	u_int32_t flags;
+{
+	DB_ENV *dbenv;
+	int not_child, ret;
+
+	dbenv = txnp->mgrp->dbenv;
+	not_child = txnp->parent == NULL;
+	ret = __txn_discard(txnp, flags);
+	if (not_child && IS_ENV_REPLICATED(dbenv))
+		__op_rep_exit(dbenv);
+	return (ret);
+}
+
 /*
  * __txn_discard --
  *	Free the per-process resources associated with this txn handle.
@@ -682,7 +772,6 @@ __txn_discard(txnp, flags)
 
 	/* Should be no children. */
 	DB_ASSERT(TAILQ_FIRST(&txnp->kids) == NULL);
-	DB_ASSERT(F_ISSET(td, TXN_RESTORED));
 
 	/* Free the space. */
 	MUTEX_THREAD_LOCK(dbenv, txnp->mgrp->mutexp);
@@ -709,8 +798,9 @@ __txn_prepare(txnp, gid)
 	DB_TXN *txnp;
 	u_int8_t *gid;
 {
-	DBT xid;
+	DBT list_dbt, xid;
 	DB_ENV *dbenv;
+	DB_LOCKREQ request;
 	DB_TXN *kid;
 	TXN_DETAIL *td;
 	u_int32_t lflags;
@@ -725,9 +815,13 @@ __txn_prepare(txnp, gid)
 
 	/* Commit any unresolved children. */
 	while ((kid = TAILQ_FIRST(&txnp->kids)) != NULL)
-		if ((ret = kid->commit(kid, DB_TXN_NOSYNC)) != 0)
+		if ((ret = __txn_commit(kid, DB_TXN_NOSYNC)) != 0)
 			return (ret);
 
+	if (txnp->txn_list != NULL  &&
+	    (ret = __db_do_the_limbo(dbenv,
+	    NULL, txnp, txnp->txn_list, LIMBO_PREPARE)) != 0)
+		return (ret);
 	/*
 	 * In XA, the global transaction ID in the txn_detail structure is
 	 * already set; in a non-XA environment, we must set it here.  XA
@@ -736,6 +830,21 @@ __txn_prepare(txnp, gid)
 	 * of those states, then we are calling prepare directly and we need
 	 * to fill in the td->xid.
 	 */
+	if ((ret = __txn_doevents(dbenv, txnp, TXN_PREPARE, 1)) != 0)
+		return (ret);
+	memset(&request, 0, sizeof(request));
+	if (LOCKING_ON(dbenv)) {
+		request.op = DB_LOCK_PUT_READ;
+		if (IS_REP_MASTER(dbenv) &&
+		    IS_ZERO_LSN(txnp->last_lsn)) {
+			memset(&list_dbt, 0, sizeof(list_dbt));
+			request.obj = &list_dbt;
+		}
+		if ((ret = __lock_vec(dbenv,
+		    txnp->txnid, 0, &request, 1, NULL)) != 0)
+			return (ret);
+
+	}
 	if (DBENV_LOGGING(dbenv)) {
 		memset(&xid, 0, sizeof(xid));
 		if (td->xa_status != TXN_XA_ENDED &&
@@ -746,14 +855,18 @@ __txn_prepare(txnp, gid)
 		xid.size = sizeof(td->xid);
 		xid.data = td->xid;
 
-		SET_LOG_FLAGS(dbenv, txnp, lflags);
+		lflags = DB_LOG_COMMIT | DB_LOG_PERM | DB_FLUSH;
 		if ((ret = __txn_xa_regop_log(dbenv, txnp, &txnp->last_lsn,
 		    lflags, TXN_PREPARE, &xid, td->format, td->gtrid, td->bqual,
-		    &td->begin_lsn)) != 0) {
+		    &td->begin_lsn, request.obj)) != 0) {
 			__db_err(dbenv, "DB_TXN->prepare: log_write failed %s",
 			    db_strerror(ret));
-			return (ret);
 		}
+		if (request.obj != NULL && request.obj->data != NULL)
+			__os_free(dbenv, request.obj->data);
+		if (ret != 0)
+			return (ret);
+		
 	}
 
 	MUTEX_THREAD_LOCK(dbenv, txnp->mgrp->mutexp);
@@ -777,9 +890,11 @@ __txn_id(txnp)
 
 /*
  * __txn_set_timeout --
- *	Set timeout values in the txn structure.
+ *	DB_ENV->set_txn_timeout.
+ *
+ * PUBLIC: int  __txn_set_timeout __P((DB_TXN *, db_timeout_t, u_int32_t));
  */
-static int
+int
 __txn_set_timeout(txnp, timeout, op)
 	DB_TXN *txnp;
 	db_timeout_t timeout;
@@ -840,8 +955,12 @@ __txn_isvalid(txnp, tdp, op)
 		if (txnp->txnid != tp->txnid)
 			return (0);
 
-		/* What we've got had better be a restored transaction. */
-		if (!F_ISSET(tp, TXN_RESTORED)) {
+		/*
+		 * What we've got had better be either a prepared or
+		 * restored transaction.
+		 */
+		if (tp->status != TXN_PREPARED &&
+		    !F_ISSET(tp, TXN_DTL_RESTORED)) {
 			__db_err(mgrp->dbenv, "not a restored transaction");
 			return (__db_panic(mgrp->dbenv, EINVAL));
 		}
@@ -910,6 +1029,7 @@ __txn_end(txnp, is_commit)
 {
 	DB_ENV *dbenv;
 	DB_LOCKREQ request;
+	DB_TXNLOGREC *lr;
 	DB_TXNMGR *mgr;
 	DB_TXNREGION *region;
 	TXN_DETAIL *tp;
@@ -921,30 +1041,34 @@ __txn_end(txnp, is_commit)
 	do_closefiles = 0;
 
 	/* Process commit events. */
-	if ((ret = __txn_doevents(dbenv, txnp, is_commit, 0)) != 0)
+	if ((ret = __txn_doevents(dbenv,
+	    txnp, is_commit ? TXN_COMMIT : TXN_ABORT, 0)) != 0)
 		return (__db_panic(dbenv, ret));
 
-	/* Release the locks. */
-	request.op = txnp->parent == NULL ||
-	    is_commit == 0 ? DB_LOCK_PUT_ALL : DB_LOCK_INHERIT;
-
 	/*
+	 * Release the locks.
+	 *
 	 * __txn_end cannot return an simple error, we MUST return
 	 * success/failure from commit or abort, ignoring any internal
 	 * errors.  So, we panic if something goes wrong.  We can't
 	 * deadlock here because we're not acquiring any new locks,
 	 * so DB_LOCK_DEADLOCK is just as fatal as any other error.
 	 */
-	if (LOCKING_ON(dbenv) && (ret = dbenv->lock_vec(
-	    dbenv, txnp->txnid, DB_LOCK_FREE_LOCKER, &request, 1, NULL)) != 0)
-		return (__db_panic(dbenv, ret));
+	if (LOCKING_ON(dbenv)) {
+		request.op = txnp->parent == NULL ||
+		    is_commit == 0 ? DB_LOCK_PUT_ALL : DB_LOCK_INHERIT;
+		request.obj = NULL;
+		if ((ret = __lock_vec(dbenv,
+		    txnp->txnid, 0, &request, 1, NULL)) != 0)
+			return (__db_panic(dbenv, ret));
+	}
 
 	/* End the transaction. */
 	R_LOCK(dbenv, &mgr->reginfo);
 
 	tp = (TXN_DETAIL *)R_ADDR(&mgr->reginfo, txnp->off);
 	SH_TAILQ_REMOVE(&region->active_txn, tp, links, __txn_detail);
-	if (F_ISSET(tp, TXN_RESTORED)) {
+	if (F_ISSET(tp, TXN_DTL_RESTORED)) {
 		region->stat.st_nrestores--;
 		do_closefiles = region->stat.st_nrestores == 0;
 	}
@@ -970,6 +1094,10 @@ __txn_end(txnp, is_commit)
 		TAILQ_REMOVE(&txnp->parent->kids, txnp, klinks);
 
 	/* Free the space. */
+	while ((lr = STAILQ_FIRST(&txnp->logs)) != NULL) {
+		STAILQ_REMOVE(&txnp->logs, lr, __txn_logrec, links);
+		__os_free(dbenv, lr);
+	}
 	if (F_ISSET(txnp, TXN_MALLOC)) {
 		MUTEX_THREAD_LOCK(dbenv, mgr->mutexp);
 		TAILQ_REMOVE(&mgr->txn_chain, txnp, links);
@@ -983,15 +1111,39 @@ __txn_end(txnp, is_commit)
 		(void)__dbreg_close_files(dbenv);
 		F_CLR((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER);
 		mgr->n_discards = 0;
-		(void)dbenv->txn_checkpoint(dbenv, 0, 0, DB_FORCE);
+		(void)__txn_checkpoint(dbenv, 0, 0, DB_FORCE);
 	}
+
 	return (0);
 }
 
+static int
+__txn_dispatch_undo(dbenv, txnp, rdbt, key_lsn, txnlist)
+	DB_ENV *dbenv;
+	DB_TXN *txnp;
+	DBT *rdbt;
+	DB_LSN *key_lsn;
+	void *txnlist;
+{
+	int ret;
+
+	ret = __db_dispatch(dbenv, dbenv->recover_dtab,
+	    dbenv->recover_dtab_size, rdbt, key_lsn, DB_TXN_ABORT, txnlist);
+	if (F_ISSET(txnp, TXN_CHILDCOMMIT))
+		(void)__db_txnlist_lsnadd(dbenv,
+		    txnlist, key_lsn, 0);
+	if (ret == DB_SURPRISE_KID) {
+		if ((ret = __db_txnlist_lsninit(
+		    dbenv, txnlist, key_lsn)) == 0)
+			F_SET(txnp, TXN_CHILDCOMMIT);
+	}
+
+	return (ret);
+}
+
 /*
  * __txn_undo --
- *	Undo the transaction with id txnid.  Returns 0 on success and
- *	errno on failure.
+ *	Undo the transaction with id txnid.
  */
 static int
 __txn_undo(txnp)
@@ -1002,6 +1154,7 @@ __txn_undo(txnp)
 	DB_LOGC *logc;
 	DB_LSN key_lsn;
 	DB_TXN *ptxn;
+	DB_TXNLOGREC *lr;
 	DB_TXNMGR *mgr;
 	int ret, t_ret;
 	void *txnlist;
@@ -1021,8 +1174,6 @@ __txn_undo(txnp)
 	 */
 	memset(&rdbt, 0, sizeof(rdbt));
 
-	key_lsn = txnp->last_lsn;
-
 	/*
 	 * Allocate a txnlist for children and aborted page allocs.
 	 * We need to associate the list with the maximal parent
@@ -1045,27 +1196,43 @@ __txn_undo(txnp)
 	    (ret = __db_txnlist_lsninit(dbenv, txnlist, &txnp->last_lsn)) != 0)
 		return (ret);
 
-	if ((ret = dbenv->log_cursor(dbenv, &logc, 0)) != 0)
+	/*
+	 * Take log records from the linked list stored in the transaction,
+	 * then from the log.
+	 */
+	for (lr = STAILQ_FIRST(&txnp->logs);
+	    lr != NULL; lr = STAILQ_NEXT(lr, links)) {
+		rdbt.data = lr->data;
+		rdbt.size = 0;
+		LSN_NOT_LOGGED(key_lsn);
+		ret =
+		    __txn_dispatch_undo(dbenv, txnp, &rdbt, &key_lsn, txnlist);
+		if (ret != 0) {
+			__db_err(dbenv,
+			    "DB_TXN->abort: In-memory log undo failed: %s",
+			    db_strerror(ret));
+			goto err;
+		}
+	}
+
+	key_lsn = txnp->last_lsn;
+
+	if (!IS_ZERO_LSN(key_lsn) &&
+	     (ret = __log_cursor(dbenv, &logc)) != 0)
 		goto err;
-	while (ret == 0 && !IS_ZERO_LSN(key_lsn)) {
+
+	while (!IS_ZERO_LSN(key_lsn)) {
 		/*
 		 * The dispatch routine returns the lsn of the record
 		 * before the current one in the key_lsn argument.
 		 */
-		if ((ret = logc->get(logc, &key_lsn, &rdbt, DB_SET)) == 0) {
-			ret = __db_dispatch(dbenv, dbenv->recover_dtab,
-			    dbenv->recover_dtab_size, &rdbt, &key_lsn,
-			    DB_TXN_ABORT, txnlist);
-			if (F_ISSET(txnp, TXN_CHILDCOMMIT))
-				(void)__db_txnlist_lsnadd(dbenv,
-				    txnlist, &key_lsn, 0);
+		if ((ret = __log_c_get(logc, &key_lsn, &rdbt, DB_SET)) == 0) {
+			ret = __txn_dispatch_undo(dbenv,
+			    txnp, &rdbt, &key_lsn, txnlist);
 		}
-		if (ret == DB_SURPRISE_KID) {
-			if ((ret = __db_txnlist_lsninit(
-			    dbenv, txnlist, &key_lsn)) == 0)
-				F_SET(txnp, TXN_CHILDCOMMIT);
-		} else if (ret != 0) {
-			__db_err(txnp->mgrp->dbenv,
+
+		if (ret != 0) {
+			__db_err(dbenv,
 		    "DB_TXN->abort: Log undo failed for LSN: %lu %lu: %s",
 			    (u_long)key_lsn.file, (u_long)key_lsn.offset,
 			    db_strerror(ret));
@@ -1073,9 +1240,9 @@ __txn_undo(txnp)
 		}
 	}
 
-	ret = __db_do_the_limbo(dbenv, ptxn, txnp, txnlist);
+	ret = __db_do_the_limbo(dbenv, ptxn, txnp, txnlist, LIMBO_NORMAL);
 
-err:	if (logc != NULL && (t_ret = logc->close(logc, 0)) != 0 && ret == 0)
+err:	if (logc != NULL && (t_ret = __log_c_close(logc)) != 0 && ret == 0)
 		ret = t_ret;
 
 	if (ptxn == NULL && txnlist != NULL)
@@ -1084,17 +1251,50 @@ err:	if (logc != NULL && (t_ret = logc->close(logc, 0)) != 0 && ret == 0)
 }
 
 /*
- * Transaction checkpoint.
- * If either kbytes or minutes is non-zero, then we only take the checkpoint
- * more than "minutes" minutes have passed since the last checkpoint or if
- * more than "kbytes" of log data have been written since the last checkpoint.
- * When taking a checkpoint, find the oldest active transaction and figure out
- * its first LSN.  This is the lowest LSN we can checkpoint, since any record
- * written after since that point may be involved in a transaction and may
- * therefore need to be undone in the case of an abort.
+ * __txn_checkpoint_pp --
+ *	DB_ENV->txn_checkpoint pre/post processing.
+ *
+ * PUBLIC: int __txn_checkpoint_pp
+ * PUBLIC:     __P((DB_ENV *, u_int32_t, u_int32_t, u_int32_t));
+ */
+int
+__txn_checkpoint_pp(dbenv, kbytes, minutes, flags)
+	DB_ENV *dbenv;
+	u_int32_t kbytes, minutes, flags;
+{
+	int rep_check, ret;
+
+	PANIC_CHECK(dbenv);
+	ENV_REQUIRES_CONFIG(dbenv,
+	    dbenv->tx_handle, "txn_checkpoint", DB_INIT_TXN);
+
+	/*
+	 * On a replication client, all transactions are read-only; therefore,
+	 * a checkpoint is a null-op.
+	 *
+	 * We permit txn_checkpoint, instead of just rendering it illegal,
+	 * so that an application can just let a checkpoint thread continue
+	 * to operate as it gets promoted or demoted between being a
+	 * master and a client.
+	 */
+	if (IS_REP_CLIENT(dbenv))
+		return (0);
+
+	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
+	if (rep_check)
+		__env_rep_enter(dbenv);
+	ret = __txn_checkpoint(dbenv, kbytes, minutes, flags);
+	if (rep_check)
+		__env_rep_exit(dbenv);
+	return (ret);
+}
+
+/*
+ * __txn_checkpoint --
+ *	DB_ENV->txn_checkpoint.
  *
  * PUBLIC: int __txn_checkpoint
- * PUBLIC:     __P((DB_ENV *, u_int32_t, u_int32_t, u_int32_t));
+ * PUBLIC:	__P((DB_ENV *, u_int32_t, u_int32_t, u_int32_t));
  */
 int
 __txn_checkpoint(dbenv, kbytes, minutes, flags)
@@ -1106,24 +1306,24 @@ __txn_checkpoint(dbenv, kbytes, minutes, flags)
 	DB_TXNREGION *region;
 	TXN_DETAIL *txnp;
 	time_t last_ckp_time, now;
-	u_int32_t bytes, mbytes;
+	u_int32_t bytes, gen, mbytes;
 	int ret;
 
-	PANIC_CHECK(dbenv);
-	ENV_REQUIRES_CONFIG(dbenv,
-	    dbenv->tx_handle, "txn_checkpoint", DB_INIT_TXN);
+	ret = gen = 0;
 
 	/*
-	 * On a replication client, all transactions are read-only;  therefore,
-	 * a checkpoint is a null-op.
-	 *
-	 * We permit txn_checkpoint, instead of just rendering it illegal,
-	 * so that an application can just let a checkpoint thread continue
-	 * to operate as it gets promoted or demoted between being a
-	 * master and a client.
+	 * A client will only call through here during recovery,
+	 * so just sync the Mpool and go home.
 	 */
-	if (F_ISSET(dbenv, DB_ENV_REP_CLIENT))
-		return (0);
+	if (IS_REP_CLIENT(dbenv)) {
+		if (MPOOL_ON(dbenv) && (ret = __memp_sync(dbenv, NULL)) != 0) {
+			__db_err(dbenv,
+		    "txn_checkpoint: failed to flush the buffer cache %s",
+			    db_strerror(ret));
+			return (ret);
+		} else
+			return (0);
+	}
 
 	mgr = dbenv->tx_handle;
 	region = mgr->reginfo.primary;
@@ -1140,6 +1340,12 @@ __txn_checkpoint(dbenv, kbytes, minutes, flags)
 		if (bytes == 0 && mbytes == 0)
 			return (0);
 
+		/*
+		 * If either kbytes or minutes is non-zero, then only take the
+		 * checkpoint if more than "minutes" minutes have passed or if
+		 * more than "kbytes" of log data have been written since the
+		 * last checkpoint.
+		 */
 		if (kbytes != 0 &&
 		    mbytes * 1024 + bytes / 1024 >= (u_int32_t)kbytes)
 			goto do_ckp;
@@ -1163,7 +1369,12 @@ __txn_checkpoint(dbenv, kbytes, minutes, flags)
 			return (0);
 	}
 
-do_ckp:	/* Look through the active transactions for the lowest begin LSN. */
+do_ckp:	/*
+	 * Find the oldest active transaction and figure out its "begin" LSN.
+	 * This is the lowest LSN we can checkpoint, since any record written
+	 * after it may be involved in a transaction and may therefore need
+	 * to be undone in the case of an abort.
+	 */
 	R_LOCK(dbenv, &mgr->reginfo);
 	for (txnp = SH_TAILQ_FIRST(&region->active_txn, __txn_detail);
 	    txnp != NULL;
@@ -1173,7 +1384,7 @@ do_ckp:	/* Look through the active transactions for the lowest begin LSN. */
 			ckp_lsn = txnp->begin_lsn;
 	R_UNLOCK(dbenv, &mgr->reginfo);
 
-	if (MPOOL_ON(dbenv) && (ret = dbenv->memp_sync(dbenv, NULL)) != 0) {
+	if (MPOOL_ON(dbenv) && (ret = __memp_sync(dbenv, NULL)) != 0) {
 		__db_err(dbenv,
 		    "txn_checkpoint: failed to flush the buffer cache %s",
 		    db_strerror(ret));
@@ -1189,6 +1400,8 @@ do_ckp:	/* Look through the active transactions for the lowest begin LSN. */
 		R_LOCK(dbenv, &mgr->reginfo);
 		last_ckp = region->last_ckp;
 		R_UNLOCK(dbenv, &mgr->reginfo);
+		if (REP_ON(dbenv))
+			__rep_get_gen(dbenv, &gen);
 
 		/*
 		 * Put out records for the open files before we log
@@ -1199,9 +1412,9 @@ do_ckp:	/* Look through the active transactions for the lowest begin LSN. */
 		 * checkpoint.
 		 */
 		if ((ret = __dbreg_open_files(dbenv)) != 0 ||
-		    (ret = __txn_ckp_log(dbenv,
-		    NULL, &ckp_lsn, DB_PERMANENT | DB_FLUSH, &ckp_lsn,
-		    &last_ckp, (int32_t)time(NULL))) != 0) {
+		    (ret = __txn_ckp_log(dbenv, NULL,
+		    &ckp_lsn, DB_FLUSH | DB_LOG_PERM | DB_LOG_CHKPNT, &ckp_lsn,
+		    &last_ckp, (int32_t)time(NULL), gen)) != 0) {
 			__db_err(dbenv,
 			    "txn_checkpoint: log failed at LSN [%ld %ld] %s",
 			    (long)ckp_lsn.file, (long)ckp_lsn.offset,
@@ -1211,7 +1424,7 @@ do_ckp:	/* Look through the active transactions for the lowest begin LSN. */
 
 		__txn_updateckp(dbenv, &ckp_lsn);
 	}
-	return (0);
+	return (ret);
 }
 
 /*
@@ -1283,7 +1496,7 @@ __txn_force_abort(dbenv, buffer)
 {
 	DB_CIPHER *db_cipher;
 	HDR *hdr;
-	u_int32_t offset, opcode, rec_len, rec_type, sum_len;
+	u_int32_t hdrlen, offset, opcode, rec_len, sum_len;
 	u_int8_t *bp, *key, chksum[DB_MAC_KEY];
 	size_t hdrsize;
 	int ret;
@@ -1292,28 +1505,25 @@ __txn_force_abort(dbenv, buffer)
 
 	/*
 	 * This routine depends on the layout of HDR and the __txn_regop
-	 * record in txn.src.  We are passed the beginning of the commit
-	 * record in the log buffer and overwrite the commit with an abort
-	 * and recalculate the checksum.  We may be passed a txn_xa_regop
-	 * that is, an XA prepare), there's no need to overwrite that one.
+	 * __txn_xa_regop records in txn.src.  We are passed the beginning
+	 * of the commit record in the log buffer and overwrite the
+	 * commit with an abort and recalculate the checksum.
 	 */
+	hdrsize = CRYPTO_ON(dbenv) ? HDR_CRYPTO_SZ : HDR_NORMAL_SZ;
+
 	hdr = (HDR *)buffer;
-	memcpy(&rec_type, hdr, sizeof(rec_type));
-	if (rec_type == DB___txn_xa_regop)
-		return (0);
+	memcpy(&hdrlen, buffer + SSZ(HDR, len), sizeof(hdr->len));
+	rec_len = hdrlen - hdrsize;
 
 	offset = sizeof(u_int32_t) + sizeof(u_int32_t) + sizeof(DB_LSN);
-	rec_len = offset + sizeof(u_int32_t) + sizeof(int32_t);
 	if (CRYPTO_ON(dbenv)) {
 		key = db_cipher->mac_key;
-		hdrsize = HDR_CRYPTO_SZ;
 		sum_len = DB_MAC_KEY;
 		if ((ret = db_cipher->decrypt(dbenv, db_cipher->data,
 		    &hdr->iv[0], buffer + hdrsize, rec_len)) != 0)
 			return (__db_panic(dbenv, ret));
 	} else {
 		key = NULL;
-		hdrsize = HDR_NORMAL_SZ;
 		sum_len = sizeof(u_int32_t);
 	}
 	bp = buffer + hdrsize + offset;
